* Machine Language Graphic Fill
* by Randy Ding  V2.3  6/94, (V2.2  8/93)
* randyd@crl.com
*
* draws on HARDBUFF
* unlimited grob size! (well really 65536x65536)
*
* Takes three objects from stack:
* [1]   X Y %0  ->
*       (%0 is real no. 0, X & Y are user binary no. (hex strings))
*       will solid fill region boardered by ON pixels or edge of grob
*
* [2]   X Y PATTERN ->
*       (three user binarys (hex strings))
*       will pattern fill region boardered by ON pixels or edge of grob
*       pattern is a 64 bit hex string 8x8 bitmap
*       !!! Each of the 16 four bit nibbles have the bit order reversed !!!
*
* Note for pattern fill it will first do a solid fill then mask off
* everything that should not be part of the pattern.  Maybe a future
* version will not have this extra step.  It is still fast, however.

ASSEMBLE

           NIBASC  /HPHP48-E/
POPHXS     EQU     #53F8D

RPL

::
 CK3NOLASTWD
 CK&DISPATCH1
 #BB1                        ( *3:x hex string; 2:y hex string; 1:0 real* )
 ::
  SKIP
  "R. Ding v2.3 6/94"
  %0<> case 2DROP            ( *abort if %0 not on level 1* )
  HXS># SWAP
  HXS># SWAP
  HARDBUFF GROBDIM           ( *x, y, h, w* )
  4PICKOVER #<
  4PICK 4PICK #<
  AND NOTcase 4DROP          ( *abort if x,y not in VDISP range* )
  SWAP                       ( *x, y, w, h* )
  CODE

* ---------------------------------------------
* FILL PROGRAM, draws on current display grob
* takes #x #y #width #height ->
* does not check for bad input values!
* width and height is current display grob size
* max VDISP grob X or Y size is 65536 pixels
* total stack levels = mine is 1, however a rom
*   call may have another, not sure.
* alters all registers
* written by Randy Ding 8/93

* status reg bit assignments
*
upstate    EQU #B            state of pixel x,y-1
dnstate    EQU #A            state of pixel x,y+1
xystate    EQU #9            state of pixel x,y
oldupstate EQU #8            prev upstate
olddnstate EQU #7            prev dnstate
dir        EQU #6            0:left 1:right
upvalid    EQU #5            1: y-1 is in range
dnvalid    EQU #4            1: y+1 is in range

* ram offset constants
* these vars are kept in the beginning of ram allocated with MAKE$N
* they are 5 nibs each execpt for list which is a stack of 8 nib entries
*
x          EQU #0            starting point of current horiz. line
y          EQU #5            "
width      EQU #A            grob width (const)
height     EQU #F            grob height (const)
numlist    EQU #14           # of x,y pairs (8 nib entries) in list below
nibs/line  EQU #19           # nibs in each line of grob (const)
ylineptr   EQU #1E           ptr to beginning of line y in grob
x2         EQU #23           left/right moving x, dir specified in ST reg
stklist    EQU #28           begin stack of x,y pairs, x=4 nibs, y=4 nibs

* ammount of ram to allocate with MAKE$N, adjust so maxnumlist is whole #
*
ramnibs    EQU (#800)+(stklist)

* max number of entries in list, 8 nibs per entry, make it a whole number
*
maxnumlis0 EQU (ramnibs)-(stklist)
maxnumlist EQU (maxnumlis0)/8

* on stack: #x #y #VDISPwidth #VDISPheight ->
* begin initial setup
*   entire  \ /  B(A) holds ->RAM allocated with MAKE$N
*   program / \  D(A) holds ->ROW1 of grob
*
FILL       GOSBVL  =POP#
           R4=A              height
           GOSBVL  =POP#
           R3=A              width
           GOSBVL  =POP#
           R2=A              y
           GOSBVL  =POP#
           R1=A              x
           GOSBVL  =SAVPTR
           LC(5)   ramnibs   allocate some ram, size = ramnibs
           GOSBVL  =MAKE$N
           AD0EX
           B=A     A         keep, B is ->Ram  !!!!!!!!!!
           GOSBVL  =D0->Row1
           CD0EX
           D=C     A         keep, D is ->Row1  !!!!!!!!!!
           C=R1.F  A         copy R1 to x
           A=B     A
           D0=A
           DAT0=C  A
           C=R2.F  A         copy R2 to y
           D0=D0+  5
           DAT0=C  A
           C=R3.F  A         copy R3 to width
           D0=D0+  5
           DAT0=C  A
           C=R4.F  A         copy R4 to height
           D0=D0+  5
           DAT0=C  A
           C=0     A         put #0 in numlist
           D0=D0+  5
           DAT0=C  A
           C=R3.F  A         calculate nibs/line
           SB=0              nibs/line = width div 4
           CSRB.F  A
           CSRB.F  A
           ?SB=0             add 1 to nibs/line if have a remainder
           GOYES   NL1
           C=C+1   A
NL1        ?CBIT=0 0         add 1 to nibs/line if odd (byte align)
           GOYES   NL2
           C=C+1   A
NL2        D0=D0+  5
           DAT0=C  A
           GOTO    MAIN1LOOP

* ------------------------------------
* MAIN LOOP
MAINLOOP   LA(5)   numlist   pop x and y from stklist
           A=A+B   A
           D0=A
           A=DAT0  A
           A=A-1   A         decr numlist
           GONC    NOTDONE   exit program if no entries in stklist
           GOVLNG  =GETPTRLOOP

* location of topmost x,y in stklist = ->RAM + stklist offset +8 * (numlist-1)
NOTDONE    DAT0=A  A         store numlist back one less
           A=A+A   A         mult by 8 nibs per entry
           A=A+A   A
           A=A+A   A
           LC(5)   stklist
           A=A+C   A
           A=A+B   A
           D0=A              -> x,y entry; x is in lower memory
           LA(5)   x
           A=A+B   A
           D1=A              -> pointer to x, y follows
           A=0     A
           P=      3
           A=DAT0  WP        read (16 bits) and store x
           DAT1=A  A
           D0=D0+  4
           D1=D1+  5
           A=DAT0  WP        read (16 bits) and store y
           DAT1=A  A
           P=      0
MAIN1LOOP  ST=1    upstate   entering here skips poping x,y from stklist
           ST=1    dnstate
           ST=0    dir
           LA(5)   x         copy x to x2
           A=A+B   A
           D0=A
           C=DAT0  A
           LA(5)   x2
           A=A+B   A
           D0=A
           DAT0=C  A
           GOSUB   CALCYLINE
           GOSUB   TESTXY
           ?ST=0   xystate   skip horizontal loop if x,y is allready on
           GOYES   HORIZLOOP
           GOTO    MAINLOOP

* --------------------------------------------
* HORIZONTAL LINE LOOP
HORIZLOOP  ?ST=0   upvalid   skip looking up if not valid
           GOYES   HLD
           ?ST=1   upstate   test for 1 to 0 transition of upstate
           GOYES   HLD
           ?ST=0   oldupstate
           GOYES   HLD
           LA(5)   y         push y-1,x2 to stklist
           A=A+B   A
           D0=A
           C=DAT0  A
           C=C-1   A
           GOSUB   PUSHLIST
HLD        ?ST=0   dnvalid   skip looking down if not valid
           GOYES   NEXTX
           ?ST=1   dnstate   test for 1 to 0 transition of dnstate
           GOYES   NEXTX
           ?ST=0   olddnstate
           GOYES   NEXTX
           LA(5)   y         push y+1,x2 to stklist
           A=A+B   A
           D0=A
           C=DAT0  A
           C=C+1   A
           GOSUB   PUSHLIST
NEXTX      ?ST=1   dir
           GOYES   GORIGHT
           LA(5)   x2        going left, decr x2
           A=A+B   A
           D0=A
           A=DAT0  A
           A=A-1   A
           GOC     REVDIR    change to right dir when x2=0
           DAT0=A  A         put back x2-1
           GOSUB   TESTXY
           ?ST=1   xystate   change to right dir when new x2,y is on
           GOYES   REVDIR
           GOTO    HORIZLOOP
REVDIR     LA(5)   x         copy x to x2 and change dir to right (=1)
           A=A+B   A
           D1=A
           LA(5)   x2
           A=A+B   A
           D0=A
           A=DAT1  A
           DAT0=A  A
           ST=1    dir       set right dir flag
           GOSUB   TESTXY    to initialize upstate/dnstate
GORIGHT    LA(5)   x2        going right, incr x2
           A=A+B   A
           D0=A
           A=DAT0  A
           A=A+1   A
           DAT0=A  A         put back x2+1
           LC(5)   width     compare against grob width
           C=C+B   A
           D0=C
           C=DAT0  A
           ?A>=C   A         test for new x2 in range
           GOYES   GOMAIN    go if out of range
           GOSUB   TESTXY    go if new x2,y pixel allready on
           ?ST=1   xystate
           GOYES   GOMAIN
           GOTO    HORIZLOOP
GOMAIN     GOTO    MAINLOOP

* ----------------------------------------------------
* SUBROUTINE PUSHLIST
* takes y value from C(A), pushes it and x2 to stklist
* does numlist housekeeping
* aborts push if numlist will exceed memory limit
* alters:  A, C, D0, D1, R0
PUSHLIST   R0=C.F  A         temp save C in R0
           LA(5)   numlist
           A=A+B   A
           D0=A
           A=DAT0  A
           LC(5)   maxnumlist
           ?A>=C   A         true if stklist is full
           RTNYES
           A=A+1   A
           DAT0=A  A         sto back numlist plus one
           A=A-1   A
           A=A+A   A         (numlist-1) * 8 nibs/entry
           A=A+A   A
           A=A+A   A
           A=A+B   A
           LC(5)   stklist
           A=A+C   A
           D1=A              pointer to top entry in stklist
           C=R0.F  A         recall saved C
           CSL     W         make room for 4 nib x2
           CSL     W
           CSL     W
           CSL     W
           LA(5)   x2        take first 4 nibs of x2
           A=A+B   A
           D0=A
           P=      3
           C=DAT0  WP
           P=      7         store 8 nib entry; x,y pair
           DAT1=C  WP
           P=      0
           RTN


* ------------------------------------------------
* SUBROUTINE CALCYLINE
* calc ylineptr and initialize up/dnvalid flags
* ylineptr = ->grobline1 + y*(nibs/line), do C=A*D
* alters R0, A, C, D0, stack levels 0
CALCYLINE  C=D     A         temp save D in R0
           R0=C.F  A
           LA(5)   nibs/line   put nibs/line into D
           A=A+B   A
           D0=A
           C=DAT0  A
           D=C     A
           LA(5)   y         put y in A
           A=A+B   A
           D0=A
           A=DAT0  A
           ST=1    upvalid   init upvalid/dnvalid flags
           ST=1    dnvalid
           ?A#0    A
           GOYES   YL1
           ST=0    upvalid   when y=0
YL1        LC(5)   height
           C=C+B   A
           D0=C
           C=DAT0  A
           C=C-1   A
           ?A<C    A
           GOYES   YL2
           ST=0    dnvalid   when y=height-1
YL2        P=      #D        count, use max of first 14 bits of nibs/line
           C=0     A         mult y by nibs/line -> C(A)
MULTLOOP   SB=0
           DSRB.F  A
           ?SB=0
           GOYES   NOADD
           C=C+A   A
NOADD      ?D=0    A         quit ahead of time if all bits are zero
           GOYES   MULTDONE
           A=A+A   A
           P=P-1
           GONC    MULTLOOP
MULTDONE   P=      0
           A=R0.F  A         add ->grob to mult result
           C=C+A   A
           LA(5)   ylineptr  store result in ylineptr
           A=A+B   A
           D0=A
           DAT0=C  A
           C=R0.F  A         recall saved D from R0
           D=C     A
           RTN


* -------------------------------------------------------------
* SUBROUTINE TESTXY test states of x2,y; x2,y-1; x2,y+1
* then changes flags xystate, upstate, dnstate
* x,y pixel is turned on after testing states
* reads upvalid, dnvalid flags
* also moves previous up/dnstate flags to oldupstate/olddnstate
* alters A, C, R0, R1, D0, D1, stack levels = 0
TESTXY     LA(5)   x2        put x into C
           A=A+B   A
           D0=A
           C=DAT0  A
           LA(5)   ylineptr  put ptr in A
           A=A+B   A
           D0=A
           A=DAT0  A
           SB=0              div C by 4, put remainder in P
           CSRB.F  A
           ?SB=0
           GOYES   PO1
           P=P+1
PO1        SB=0
           CSRB.F  A
           ?SB=0
           GOYES   PO2
           P=P+1
           P=P+1
PO2        A=A+C   A         add C offset to ylineptr
           D1=A              D1 is ptr to pixel address
           C=0     A         put bit mask in first nib of C
           C=C+1   A         make C=1
POLOOP     ?P=     0
           GOYES   POLDONE
           C=C+C   B         shift C bit left P times
           P=P-1
           GOTO    POLOOP
POLDONE    A=B     A         save B in R0 and put nibs/line in B
           R0=A.F  A
           LA(5)   nibs/line
           A=A+B   A
           D0=A
           A=DAT0  A
           B=A     A
           ST=0    oldupstate  copy up/dnstate to old
           ST=0    olddnstate
           ?ST=0   upstate
           GOYES   PO3
           ST=1    oldupstate
PO3        ?ST=0   dnstate
           GOYES   PO4
           ST=1    olddnstate
PO4        ST=0    xystate
           A=DAT1  P         test x,y
           A=A&C   P
           ?A=0    P
           GOYES   PO5
           ST=1    xystate
PO5        A=DAT1  P         turn on x,y pixel
           A=A!C   P
           DAT1=A  P
           ?ST=0   upvalid   test x,y-1 only if valid
           GOYES   PO7
           ST=0    upstate
           AD1EX
           R1=A.F  A         save current D1 in R1(A)
           A=A-B   A         subtract nibs/line
           AD1EX
           A=DAT1  P
           A=A&C   P
           ?A=0    P
           GOYES   PO6
           ST=1    upstate
PO6        A=R1.F  A         rcl previous D1
           D1=A
PO7        ?ST=0   dnvalid   test x,y+1 only if valid
           GOYES   PODONE
           ST=0    dnstate
           AD1EX
           A=A+B   A         add nibs/line
           AD1EX
           A=DAT1  P
           A=A&C   P
           ?A=0    P
           GOYES   PODONE
           ST=1    dnstate
PODONE     A=R0.F  A         rcl saved B from R0
           B=A     A
           RTN

  ENDCODE
 ;                           ( *end first dispatch* )



* ------------------------------------------------------------------
* begin second dispatch, 3 hex strings
* stack:  x, y, pattern8x8  -->
*
 #BBB
 ::
  UNROT
  HXS># SWAP
  HXS># SWAP
  HARDBUFF GROBDIM           ( *pattern, x, y, h, w* )
  4PICKOVER #<               ( *test x<w* )
  4PICK 4PICK #<             ( *test y<h* )
  AND NOTcase 5DROP          ( *exit if x,y not in range* )
  SWAP                       ( *pattern, x, y, w, h* )
  HARDBUFF TOTEMPOB 6UNROLL  ( *duplicate origional HB grob* )
  2SWAP 4PICK 4PICK          ( *HB, pattern, w, h, x, y, w, h* )
  CODE
           GOTO    FILL      do solid fill  ( x,y,w,h --> )
  ENDCODE
  CODE

* mask HARDBUFF with 8x8pattern, 16 nib hex string
*
DOMASK     GOSBVL  =POP#     do masking  ( pattern,w,h --> )
           R4=A.F  A         height
           GOSBVL  =POP#
           R3=A.F  A         width
           GOSBVL  POPHXS    64 bit pattern in A, does SAVPTR
           R2=A
           GOSBVL  =D0->Row1
           A=R2              keep pattern in A(W)
           C=R3.F  A         bytes/line = width div 8
           SB=0
           CSRB.F  A
           CSRB.F  A
           CSRB.F  A
           ?SB=0             add 1 to bytes/line if have a remainder
           GOYES   BL
           C=C+1   A
BL         C=C-1   A         count thru 0
           GOC     EXIT
           R3=C.F  A         store bytes/line over width in R3
           C=R4.F  A         init vert count in B
           C=C-1   A         count thru 0
           GOC     EXIT
           B=C     A
VLOOP      C=R3.F  A         init byte horiz count in D
           D=C     A
HLOOP      C=DAT0  B         do masking byte for byte
           C=C&A   B
           DAT0=C  B
           D0=D0+  2         next byte
           D=D-1   A         dec horiz byte count
           GONC    HLOOP
           ASRC              rotate pattern for masking next line
           ASRC
           B=B-1   A         vert count
           GONC    VLOOP
EXIT       GOVLNG  =GETPTRLOOP

  ENDCODE
  TRUE                       ( *OR Flag* )
  HARDBUFF ROT
  ZEROZERO GROB+# DROP       ( *orig OR HARDBUFF -> HARDBUFF* )
 ;                           ( *end second dispatch* )
;                            ( *end program* )
