**********************************************************************
* Desc:		Halleck's factorization method for short odd integers
* Stack:	( Z --> Z1 Z2 )
* Notes:
*		N = PQ = (S-A)(S+B)
*
* Idea:		a+=1	: (s-a)(s+b) -= s+b
*		b+=1	: (s-a)(s+b) += s-a
*
*		Pretty fast when factors are close to sqrt(n)
*
**********************************************************************
CODE
		GOSUB	POPZ
		R0=A			R0 = N
		GOSUB	ISQRT		S = SQRT(N)

		B=A	W		P = S
		C=A	W		Q = S
		GOSBVL	=MPY		P*Q
		C=R0
		A=A-C	W		P*Q-N (<=0)
		C=B	W		Q = S

		?A=0	W
		GOYES	halfound
halloop1	C=C+1	W		Q++
		A=A+B	W
		GONC	halloop1
		?A=0	W
		GOYES	halfound
halloop2	B=B-1	W
		A=A-C	W
		GOC	halloop1
		?A#0	W
		GOYES	halloop2
halfound	A=B	W		P
		R4=C	W		Q
		GOSUB	PUSHZ
		A=R4
		GOSUB	PUSHZ
		GOVLNG	=GETPTRLOOP

*---------------------------------------------------------------------
PUSHZ		P=	15
-		?A#0	P
		GOYES	+
		P=P-1
		GONC	-
+		P=P+1
		GOSBVL	=PUSHhxs
		GOVLNG	=SAVPTR
*---------------------------------------------------------------------
POPZ		GOSBVL	=PopASavptr
		D0=A
		D0=D0+	5
		A=0	W
		C=DAT0	A
		D0=D0+	5
		C=C-CON	A,7
		RTNC			CS: Zero
		P=C	0
		A=DAT0	WP
		P=	0
		RTN
*---------------------------------------------------------------------
ISQRT		P=	0
		C=0	W		remainder = 0
		?A=0	W		sqrt(0) = 0
		RTNYES
		LC(2)	32-1
		D=C	A		D[B]=loop counter
		B=0	W		B[W]=result
		C=0	A		C[W]=remainder
-		?A#0	S		Find range faster
		GOYES	+
		ASL	W
		D=D-CON	B,2
		GONC	-
+
ISQRTLP		C=C+C	W		Generate bit 1
		A=A+A	W
		GONC	+
		C=C+1	P
+		C=C+C	W		Generate bit 2
		A=A+A	W
		GONC	+
		C=C+1	P
+		B=B+B	W		result *= 2
		B=B+B	W
		B=B+1	P
		?C<B	W		y < result*2+1
		GOYES	+
		C=C-B	W		y -= result*2+1
		B=B+1	P
+		BSRB			fix back to correct order
		D=D-1	B
		GONC	ISQRTLP
		A=B	W		A[W] = result
		RTN
ENDCODE
**********************************************************************

