	STITLE	ALG48 Special Functions Library
**********************************************************************
*		Library Specifics
**********************************************************************

** ROMID

xROMID 38F ( 911 )
ASSEMBLE
=SPECROMID	EQU #38F
RPL

** CONFIG

ASSEMBLE
=SPECcfg
RPL
:: [#] SPECROMID TOSRRP ;

** MESSAGES
ASSEMBLE
=SPECmsg
RPL

** EXTERNALS

*EXTERNAL xFACTOR
EXTERNAL xGAMMA		( x --> y )
EXTERNAL xBETA		( x1 x2 --> y )
EXTERNAL xERF		( x --> y )
EXTERNAL xERFC		( x --> y )
EXTERNAL xJ.0		( x --> y )
EXTERNAL xJ.1		( x --> y )
EXTERNAL xJ.n		( x n --> y )
EXTERNAL xY.n		( x n --> y )
EXTERNAL xI.n		( x n --> y )
EXTERNAL xK.n		( x n --> y )
EXTERNAL xP.n		( n x --> y )
EXTERNAL xP.nm		( n m x --> y )
EXTERNAL xY.nm		( n m x1 x2 --> y )
EXTERNAL xH.n		( n x --> y )
EXTERNAL xT.n		( n x --> y )
EXTERNAL xU.n		( n x --> y )
EXTERNAL xL.n		( n x --> y )
EXTERNAL xL.nm		( n m x --> y )

*	EXTERNAL ZFactorial	( Z --> Z' | Error )
EXTERNAL Factorial	( # --> Z )
EXTERNAL DoPn
EXTERNAL Pn
EXTERNAL CKPnm
EXTERNAL Pnm
EXTERNAL Hn
EXTERNAL Tn
EXTERNAL Un
EXTERNAL Ln
EXTERNAL Sym:
EXTERNAL RealSym:
EXTERNAL SymReal:
EXTERNAL SymSym:
EXTERNAL Gamma
EXTERNAL Bessel

**********************************************************************
*		Unsupported Entries
**********************************************************************
ASSEMBLE

=xUTPN		EQU #2003A
=scknum1	EQU #545A0
=handleDVZ	EQU #5203A
=syminner&	EQU #5E2F8

=POP2%SPLTAC	EQU #29FF8
=END%		EQU #2AED7
=NoXcpEND%	EQU #2AEDB
=NoXcpEND%SB	EQU #2AEE9
=aLNF		EQU #2B698
=aEXP15F	EQU #2B6AA
=aSIN15F	EQU #2B6E0
=aCOS15F	EQU #2B6E9
=aFCSTRT	EQU #2B716
=aFLOOR		EQU #2B767
=MULT2		EQU #2BBD9
=GETCD1		EQU #2C017
=PUTAB1		EQU #2C066

TST<	EQU 1		TST15 test type equates
TST=	EQU 2
TST>	EQU 4
TST#	EQU 13

RPL

**********************************************************************
*		Programs
**********************************************************************

**********************************************************************
* Name:		FACTOR
* Desc:		Factorial command generalized to infinite precision
* Stack:	( % --> %' | Z )
**********************************************************************
*	ASSEMBLE
*		CON(1)	8
*	RPL
*	xNAME FACTOR
*	::
*	  CK1&Dispatch
*	  real
*	  ::
*		FastLib
*		DUP %0>= NOTcase %FACT		( Regular ! if n<0 )
*		DUPDUP %IP %<> case %FACT	( Regular ! if n not integer )
*		COERCE Factorial		( --> Z )
*		DUP HXS B FFF05A4D8E0 ZNLT?	( Z flag )
*		case Z>R Z>S			( Choose output form )
*	  ;
*	  hxs
*	  ::
*		FastLib
*		H>Z ZFactorial			( --> Z )
*	  ;
*	  str
*	  ::
*		FastLib
*		S>Z ZFactorial Z>S		( --> S )
*	  ;
*	;

*---------------------------------------------------------------------
* Desc:		Factorial of integer Z
* Stack:	( Z --> Z | Error )
*---------------------------------------------------------------------
*	NULLNAME ZFactorial
*	::
*	  DupZIsNeg? case ErrInfRes		( Error for negative Z )
*	  DUP HXS 6 FFFFF0 ZNLT?		( Z flag )
*	  NOTcase ErrInfRes			( Too big Z )
*	  HXS>#					( --> # )
*	  Factorial
*	;

*---------------------------------------------------------------------
* Desc:		Factorial of integer #n
* Stack:	( # --> Z )
* Notes:	The loop gets very slow as n increases.
*		A better algorithm would be recursive.
*---------------------------------------------------------------------
NULLNAME Factorial
::
  DUPTWO #< casedrop HXS 2 10		( 0! = 1! = 1 )
  HXS 2 10 SWAP				( 0! #n )
  #1+_ONE_DO (DO)			( k-1! )
	INDEX@ #>Z QMul			( k! )
  LOOP
;


**********************************************************************
* Name:		xGAMMA
* Desc:		Gamma function
* Stack:	( % --> %' )
*		( sym --> sym' )
**********************************************************************
ASSEMBLE
	CON(3)	000
RPL
xNAME GAMMA
::
  CK1&Dispatch
  real	Gamma
  sym	:: scknum1 Sym: xGAMMA ;
;

**********************************************************************
* Name:		xBETA
* Desc:		Beta function
* Stack:	( % --> %' )
*		( sym --> sym' )
**********************************************************************
ASSEMBLE
	CON(3)	000
RPL
xNAME BETA
::
  CK2&Dispatch
  realreal
  ::
	2DUP %+	Gamma %>%% UNROT	( %%g[x+y] %x %y )
	Gamma %>%% SWAP			( %%g[x+y] %%g[y] %x )
	Gamma %>%%			( %%g[x+y] %%g[y] %%g[x] )
	%%* SWAP %%/ %%>%		( %b[x,y] )
  ;
  realsym	:: nscknum2 RealSym: xBETA ;
  symreal	:: sncknum2 SymReal: xBETA ;
  symsym	:: sscknum2 SymSym:  xBETA ;
;

**********************************************************************
* Name:		xERF
* Desc:		Error function
* Stack:	( % --> %' )
*		( sym --> sym' )
**********************************************************************
ASSEMBLE
	CON(3)	0
RPL
xNAME ERF
::
  CK1&Dispatch
  real
  ::
	DUP%0= ?SEMI			( Erf[0] = 0 )
	DUP %SGN			( sign x )
	%0 % .5 4ROLL %ABS		( sign 0 0.5 |x| )
	'EvalNoCK: xUTPN		( sign x' )
	%-2 %* %1+ %*			( Erf[x] )
  ;
  sym
  :: scknum1 Sym: xERF ;
;

**********************************************************************
* Name:		xERFC
* Desc:		Complementary error function
* Stack:	( % --> %' )
*		( sym --> sym' )
**********************************************************************
ASSEMBLE
	CON(3)	000
RPL
xNAME ERFC
::
  CK1&Dispatch
  real
  ::
	DUP%0= casedrop %1		( Erf[0] = 1 )
	%0 % .5 3PICK %ABS		( x 0 0.5 |x| )
	'EvalNoCK: xUTPN		( x x' )
	%2 %* SWAP %0> ?SEMI
	%2 SWAP %-
  ;
  sym
  :: scknum1 Sym: xERFC ;
;
**********************************************************************
* Name:		xJ.0
* Desc:		Bessel function of the first kind, order 0
* Stack:	( % --> %' )
*		( sym --> sym' )
**********************************************************************
ASSEMBLE
	CON(3)	0
RPL
xNAME J.0
::
  CK1&Dispatch
  real	:: %0 SWAP TrueTrue Bessel ;
  sym	:: scknum1 Sym: xJ.0 ;
;

**********************************************************************
* Name:		xJ.1
* Desc:		Bessel function of the first kind, order 1
* Stack:	( % --> %' )
*		( sym --> sym' )
**********************************************************************
ASSEMBLE
	CON(3)	0
RPL
xNAME J.1
::
  CK1&Dispatch
  real	:: %1 SWAP TrueTrue Bessel ;
  sym	:: scknum1 Sym: xJ.1 ;
;

**********************************************************************
* Name:		xJ.n
* Desc:		Bessel function of the first kind, order n
* Stack:	( %x %n --> % )
*		( x n --> sym )
**********************************************************************
ASSEMBLE
	CON(3)	0
RPL
xNAME J.n
::
  CK2&Dispatch
  realreal	:: TrueTrue Bessel ;
  realsym	:: nscknum2 RealSym: xJ.n ;
  symreal	:: sncknum2 SymReal: xJ.n ;
  symsym	:: sscknum2 SymSym:  xJ.n ;
;

**********************************************************************
* Name:		xY.n
* Desc:		Bessel function of the second kind, order n
* Stack:	( %x %n --> % )
*		( x n --> sym )
**********************************************************************
ASSEMBLE
	CON(3)	0
RPL
xNAME Y.n
::
  CK2&Dispatch
  realreal	:: FalseTrue Bessel ;
  realsym	:: nscknum2 RealSym: xY.n ;
  symreal	:: sncknum2 SymReal: xY.n ;
  symsym	:: sscknum2 SymSym:  xY.n ;
;

**********************************************************************
* Name:		xI.n
* Desc:		Modified Bessel function of the first kind, order n
* Stack:	( %x %n --> % )
*		( x n --> sym )
**********************************************************************
ASSEMBLE
	CON(3)	0
RPL
xNAME I.n
::
  CK2&Dispatch
  realreal	:: TrueFalse Bessel ;
  realsym	:: nscknum2 RealSym: xI.n ;
  symreal	:: sncknum2 SymReal: xI.n ;
  symsym	:: sscknum2 SymSym:  xI.n ;
;

**********************************************************************
* Name:		xK.n
* Desc:		Modified Bessel function of the second kind, order n
* Stack:	( %x %n --> % )
*		( x n --> sym )
**********************************************************************
ASSEMBLE
	CON(3)	0
RPL
xNAME K.n
::
  CK2&Dispatch
  realreal	:: FalseFalse Bessel ;
  realsym	:: nscknum2 RealSym: xK.n ;
  symreal	:: sncknum2 SymReal: xK.n ;
  symsym	:: sscknum2 SymSym:  xK.n ;
;


**********************************************************************
* Name:		xP.n
* Desc:		Legendre polynomial
* Stack:	( %n sym --> sym' )
* Definition:
*		P[0] = 1
*		P[1] = X
*		          (2n+1)XP[n] - nP[n-1]
*		P[n+1] = ---------------------
*			           n+1
**********************************************************************
ASSEMBLE
	CON(3)	0
RPL
xNAME P.n
::
  CK2&Dispatch
  realsym	DoPn
;

*---------------------------------------------------------------------
NULLNAME DoPn
::
  FastLib				( n sym )
  OVER %0>= NcaseTYPEERR		( Check n >= 0 )
  OVER %IP 3PICK %= NcaseTYPEERR	( Check n integer )

  SWAP COERCE				( sym n )
  Pn					( sym Pd Pn )

* Output symbolic with '1/n' factor in front

  SWAP HXS 2 10 4ROLL ONE{}N R2A
  HXS 2 10 ROT
  DupZIsOne? case2DROP
  NULL{} R2A SWAP XMul
;

*---------------------------------------------------------------------
* Desc:		Legendre polynomial
* Stack:	( #n --> Pd Pn )
*		          (2n+1)XP[n] - nP[n-1]
*		P[n+1] = ---------------------
*			           n+1
*---------------------------------------------------------------------
NULLNAME Pn
::
  HXS 2 10 DUP				( #n Pd[0] Pn[0] )
  { HXS 2 10 HXS 1 0 } OVER		( #n Pd[0] Pn[0] Pd[1] Pn[1] )
  5ROLL DUP#0=case 3DROP		( --> Pd[0] Pn[0] )
  DUP#1= casedrop :: 2SWAP 2DROP ;	( --> Pd[1] Pn[1] )

  ONE_DO (DO)			( Pd[n-1] Pn[n-1] Pd[n] Pn[n] )
	2SWAP SWAP		( Pd[n] Pn[n] Pn[n-1] Pd[n-1] )
	INDEX@ #>Z QMul		( Pd[n] Pn[n] Pn[n-1] Pd[n-1]*n )
	SWAP4PICK		( Pd[n] Pn[n] Pd[n-1]*n Pn[n-1] Pd[n] )
	INDEX@ #2* #1+ #>Z QMul	( Pd[n] Pn[n] Pd[n-1]*n Pn[n-1] Pd[n]*2n+1 )
	HXS 1 0 >TCOMP		( Pd[n] Pn[n] Pd[n-1]*n Pn[n-1] Pd[n]*2n+1*X )
	4PICK 2SWAP RSub	( Pd[n] Pn[n] Pd[n+1] Pn[n]' )
	INDEX@ #1+ #>Z QMul	( Pd[n] Pn[n] Pd[n+1] Pn[n+1] )
  LOOP
  2SWAP 2DROP			( Pd[n+1] Pn[n+1] )
  RCanno
;

**********************************************************************
* Name:		xP.nm
* Desc:		Assoc. Legendre Function
* Stack:	( %n %m sym --> sym' )
* Definition:
*		                    m
*		  m          2 m/2 d
*		 P (x) = (1-x )   --- P (x)
*		  n                 m  n
*		                  dx
*
*		  -m         m (n-m)!  m
*		 P (x) = (-1)  ------ P (x)
*		  n            (n+m)!  n
*
* Notes:
*		If sym is of the form 'COS(any)' then
*			'SIN(any)^2 + COS(any)^2 = 1'
*		is used to simplify the result.
*
**********************************************************************
ASSEMBLE
	CON(3)	0
RPL
xNAME P.nm
::
  CK3&Dispatch
  # 11A		:: FastLib CKPnm Pnm ;
;

*---------------------------------------------------------------------
* Desc:		Check n,m arguments for P.nm function
* Stack:	( %n %m sym --> #n #m mpositive? sym )
*---------------------------------------------------------------------
NULLNAME CKPnm
::
  3PICK %0>= NcaseTYPEERR			( Check n>=0 )
  3PICK DUP %IP %= NcaseTYPEERR			( Check n integer )
  OVERDUP %IP %= NcaseTYPEERR			( Check m integer )
* OVER %ABS 4PICK %<= NcaseTYPEERR		( Check ABS[m]<=n )
  ROT COERCE UNROT				( #n %m var )
  SWAPDUP %0>= UNROT				( #n mpos? var %m )
  %ABS COERCE UNROT				( #n #m mpos? var )
;

*---------------------------------------------------------------------
* Desc:		Assoc. Legendre Function
* Stack:	( #n #m mpositive? sym --> sym' )
*---------------------------------------------------------------------
NULLNAME Pnm
::

  4PICK 4PICK				( #n #m mpos? sym #n #m )
  #<case :: 4DROP %0 ;			( |m|>n - result is zero )

   ONE{}N				( #n #m mpos? {var} )

* Calculate Pn

   4PICK Pn				( #n #m mpos? {var} Pd Pn )

* Take m'th derivative

  5PICK DUP#0=ITE
	DROP
	::
	  ROTSWAP
	  ZERO_DO (DO)
		PDer
	  LOOP
	  SWAP RCanno
	;

* Multiply by n,m factor if m < 0	( #n #m mpos? {var} Pd Pn )

  4ROLL ITE
    :: HXS 2 10 SWAP ;
    ::
	5PICK 5PICK #+ Factorial	( #n #m {var} Pd Pn Zn+m! )
	QMul
	5PICK 5PICK #- Factorial	( #n #m {var} Pd Pn' Zn-m! )
	SWAP RCanno

	5PICK ONE #AND #0=?SEMI		( [-1]^m = 1 - done )
	SWAP QNeg SWAP
    ;

*					( #n #m {var} Pd Pd' Pn' )

   ROT PPZ SWAP 4UNROLL
   ROT QMul SWAP RCanno			( #n #m {var} Pd Pd' Pn' )

   NULL{} R2A
   SWAP HXS 2 10 4PICK R2A XMul		( #n #m {var} sym )

* Multiply by the root term

   4ROLLDROP
   3PICK #0=case UNROT2DROP

* Choose poly or trig form
   OVER CARCOMP syminner #1-SWAP
   ' xCOS EQ case
   ::
     putop: xSIN
     DUP #4+PICK DUP#1= ITE
       DROP
       :: UNCOERCE ' x^ ROT #2+ ;
     SYMBN XMul UNROT2DROP
   ;

   NDROP %1 ROT CARCOMP syminner
   ' xSQ ' x- ROT #3+ SYMBN

   ROTDUP #2= casedrop XMul
   #>Z HXS 2 20 RCanno
   NULL{} R2A
   syminner&
   putop: x^ SYMBN
   XMul
;


**********************************************************************
* Name:		xY.nm
* Desc:		Spherical Harmonics
* Stack:	( %n %m sym1 sym2 --> sym' )
* Definition:
*		                        ____________
*		  m                  m |(2n+1)(n-m)!  m            im alpha
*		 Y (theta,alpha)=(-1)  |------------ P (cos theta)e   
*		  n                   \| 4pi (n+m)!   n
*		                   ____________
*		  -m               |(2n+1)(n-m)!  m           -im alpha
*		 Y (theta,alpha)=  |------------ P (cos theta)e   
*		  n               \| 4pi (n+m)!   n
*
**********************************************************************
ASSEMBLE
	CON(3)	0
RPL
xNAME Y.nm
::
  CK4&Dispatch
  #11AA
  ::
	FastLib
	4UNROLL				( sym2 %n %m sym1 )

* Check n,m arguments

	CKPnm				( sym2 #n #m mpos? sym1 )

* Check for special case |m| > n

	4PICK 4PICK #<case
	:: 5DROP %0 ;

* Calculate P.nm(cos theta) part

	syminner putop: xCOS SYMBN	( sym2 #n #m mpos? sym1' )
	4PICK 4PICK TRUE 4ROLL Pnm	( sym2 #n #m mpos? sym )

* Calculate exp part

	3PICK DUP#0=ITE
	  :: DROP 5ROLLDROP ;
	  ::
		UNCOERCE 3PICK ?SKIP %CHS
		DOSYMB xi SEMI XMul
		6ROLL XMul
		syminner putop: xEXP SYMBN
		XMul
	  ;

* Calculate root part			( #n #m mpos? sym' )

	4PICK #2* #1+ #>Z		( #n #m mpos? sym' Z2n+1 )
	5PICK 5PICK #- Factorial	( #n #m mpos? sym' Z2n+1 Zn-m! )
	QMul				( #n #m mpos? sym' Zd )
	5PICK 5PICK #+ Factorial	( #n #m mpos? sym' Zd Zn-m! )
	HXS 2 40 QMul			( #n #m mpos? sym' Zd Zn )
	RCanno				( #n #m mpos? sym' Zd' Zn' )
	Z>R SWAP Z>R SWAP		( #n #m mpos? sym' %d' %n' )
	{ xPI x* x/ xSQRT }
	INNERCOMP #2+ SYMBN		( #n #m mpos? sym' symroot )
	SWAP XMul			( #n #m mpos? sym'' )

* Calculate (-1)^m part

	4ROLLDROP
	SWAP NOTcase SWAPDROP		( sym'' )
	SWAP ONE #AND #0=?SEMI		( sym'' )
	XNeg
  ;
;

**********************************************************************
* Name:		xH.n
* Desc:		Hermite Functions
* Stack:	( %n sym --> sym' )
* Definition:
*		H[0] = 1
*		H[1] = 2X
*		H[n+1] = 2X*H[n]-2n*H[n-1]
**********************************************************************
ASSEMBLE
	CON(3)	0
RPL
xNAME H.n
::
  CK2&Dispatch
  realsym
  ::
	FastLib
	OVER %0>= NcaseTYPEERR		( Check n>=0 )
	OVERDUP %IP %= NcaseTYPEERR	( Check n integer )
	SWAP COERCE			( sym #n )
	Hn				( sym P )
	HXS 2 10 ROT ONE{}N R2A		( sym' )
  ;
;

*---------------------------------------------------------------------
* Desc:		Herminte function
* Stack:	( #n --> P )
*---------------------------------------------------------------------
NULLNAME Hn
::
  HXS 2 10				( #n H[0] )
  { HXS 2 20 HXS 1 0 }			( #n H[0] H[1] )

  ROT DUP#0=case 2DROP			( --> H[0] )
  DUP#1= casedrop SWAPDROP		( --> H[1] )

  ONE_DO (DO)
	DUP HXS 1 0 >TCOMP		( H[n-1] H[n] H[n]*X )
	ROT INDEX@ #>Z QMul		( H[n] H[n]*X H[n-1]*n )
	QSub HXS 2 20 QMul		( H[n] H[n+1] )
  LOOP
  SWAPDROP				( H[n+1] )
;

**********************************************************************
* Name:		xT.n
* Desc:		Tschebyscheff Polynomials
* Stack:	( %n sym --> sym' )
* Definition:
*		T[0] = 1
*		T[1] = X
*		T[n+1] = 2X*T[n]-T[n-1]
**********************************************************************
ASSEMBLE
	CON(3)	0
RPL
xNAME T.n
::
  CK2&Dispatch
  realsym
  ::
	FastLib
	OVER %0>= NcaseTYPEERR		( Check n>=0 )
	OVERDUP %IP %= NcaseTYPEERR	( Check n integer )
	SWAP COERCE			( sym #n )
	Tn				( sym P )
	HXS 2 10 ROT ONE{}N R2A		( sym' )
  ;
;

*---------------------------------------------------------------------
* Desc:		Tschebyscheff polynomial
* Stack:	( #n --> P )
*---------------------------------------------------------------------
NULLNAME Tn
::
   HXS 2 10
   { HXS 2 10 HXS 1 0 }			( #n T[0] T[1] )
   ROT DUP#0=case 2DROP			( --> T[0] )
   DUP#1= casedrop SWAPDROP		( --> T[1] )
   ONE_DO (DO)				( T[n-1] T[n] )
     DUP HXS 2 20 QMul			( T[n-1] T[n] T[n]*2 )
     HXS 1 0 >TCOMP			( T[n-1] T[n] T[n]*2*X )
     ROT QSub				( T[n] T[n+1] )
   LOOP
   SWAPDROP				( T[n+1] )
;

**********************************************************************
* Name:		xU.n
* Desc:		Tschebyscheff Polynomials II
* Stack:	( %n sym --> sym' )
* Definition:
*		U[0] = 1
*		U[1] = 2X
*		U[n+1] = 2X*U[n]-U[n-1]
**********************************************************************
ASSEMBLE
	CON(3)	0
RPL
xNAME U.n
::
  CK2&Dispatch
  realsym
  ::
	FastLib
	OVER %0>= NcaseTYPEERR		( Check n>=0 )
	OVERDUP %IP %= NcaseTYPEERR	( Check n integer )
	SWAP COERCE			( sym #n )
	Un				( sym P )
	HXS 2 10 ROT ONE{}N R2A		( sym' )
  ;
;

*---------------------------------------------------------------------
* Desc:		Tschebyscheff Polynomials II
* Stack:	( #n --> P )
*---------------------------------------------------------------------
NULLNAME Un
::
   HXS 2 10
   { HXS 2 20 HXS 1 0 }			( #n U[0] U[1] )
   ROT DUP#0=case 2DROP			( --> U[0] )
   DUP#1= casedrop SWAPDROP		( --> U[1] )

   ONE_DO (DO)
     DUP HXS 2 20 QMul			( U[n-1] U[n] U[n]*2 )
     HXS 1 0 >TCOMP			( U[n-1] U[n] U[n]*2*X )
     ROT QSub				( U[n] U[n+1] )
   LOOP
   SWAPDROP				( U[n+1] )
;

**********************************************************************
* Name:		xL.n
* Desc:		Laguerre Polynomials
* Stack:	( %n sym --> sym' )
* Definition:
*		L[0] = 1
*		L[1] = -X+1
*		         (2n+1-x)L[n]-nL[n-1]
*		L[n+1] = --------------------
*		                n+1
**********************************************************************
ASSEMBLE
	CON(3)	1
RPL
xNAME L.n
::
  CK2&Dispatch
  realsym
  ::
	FastLib
	OVER %0>= NcaseTYPEERR		( Check n>=0 )
	OVERDUP %IP %= NcaseTYPEERR	( Check n integer )
	SWAP COERCE			( sym #n )
	Ln				( sym Pd Pn )

* Output symolic with '1/n' factor
	SWAP HXS 2 10 4ROLL ONE{}N R2A
	HXS 2 10 ROT
	DupZIsOne? case2DROP
	NULL{} R2A SWAP XMul
  ;
;

*---------------------------------------------------------------------
* Desc:		Laguarre Polynomials
* Stack:	( #n --> Pd Pn )
*---------------------------------------------------------------------
NULLNAME Ln
::
   HXS 2 10 DUP				( #n Pd[0] Pn[0] )
   { HXS 2 1F HXS 2 10 } OVER		( #n Pd[0] Pn[0] Pd[1] Pn[1] )

   5ROLL DUP#0=case 3DROP		( --> Pd[0] Pn[0] )
   DUP#1= casedrop :: 2SWAP 2DROP ;	( --> Pd[1] Pn[1] )

   ONE_DO (DO)			( Pd[n-1] Pn[n-1] Pd[n] Pn[n] )
     OVERDUP			( Pd[n-1] Pn[n-1] Pd[n] Pn[n] Pd[n] Pd[n] )
     INDEX@ #2* #1+ #>Z QMul
     SWAP HXS 1 0 >TCOMP QSub	( Pd[n-1] Pn[n-1] Pd[n] Pn[n] Pd[n]*2n+1-x )
     OVER
     6ROLL INDEX@ #>Z QMul
     6ROLL RSub			( Pd[n] Pn[n] Pd[n+1] Pn[n+1]/n+1 )

     INDEX@ #1+ #>Z QMul	( Pd[n] Pn[n] Pd[n+1] Pn[n+1] )
     
   LOOP
   2SWAP 2DROP			( Pd[n+1] Pn[n+1] )
   RCanno
;

**********************************************************************
* Name:		xL.nm
* Desc:		Accoc. Laguerre Polynomials
* Stack:	( %n %m sym --> sym' )
* Definition:
*		                k
*		  k          k d
*		 L (x) = (-1) ---L[n+k]
*		  n             k
*		              dx
**********************************************************************
ASSEMBLE
	CON(3)	0
RPL
xNAME L.nm
::
  CK3&Dispatch
  #11A
  ::
	FastLib
	OVER %0>= NcaseTYPEERR		( Check n>=0 )
	OVERDUP %IP %<> case SETTYPEERR	( Check n integer )
	3PICK %0>= NcaseTYPEERR		( Check m>=0 )
	3PICK DUP %IP %= NcaseTYPEERR	( Check m integer )

	SWAP COERCE ROT COERCE		( sym #m #n )

*	2DUP#> case :: 3DROP %0 ;	( Special case m>n )

	OVER#+ Ln			( sym #m Pn[n+m] Pd[n+m] )
	3PICK DUP#0=ITE
	:: DROP HXS 2 10 SWAP ;		( sym #m Pn Z1 Pd )
	::
	   ROTSWAP			( sym #m Pd Pn #m )
	   ZERO_DO (DO) PDer LOOP	( sym #m Pd Pn' )
	   PPZ ROT RCanno		( sym #m Pn'' Pd' )
	;
	( sym #m Pn N Pd )
	NULL{} R2A
	UNROTSWAP ONE #AND
	#0=?SKIP QNeg
	HXS 2 10 4ROLL ONE{}N R2A
	XMul
  ;
;

**********************************************************************
* Desc:		Build unary function after scknum1 failure
* Stack:	( Meta --> Meta' )
**********************************************************************
NULLNAME Sym:
::
  1GETLAM ONE SYMBN		( Meta 'fcn' )
  ONE ' xFCNAPPLY		( Meta 'fcn' 1 FCNAPPLY )
  4ROLL #3+			( Meta' )
;
**********************************************************************
* Desc:		Build binary function after nscknum2 failure
* Stack:	( Meta --> Meta' )
**********************************************************************
NULLNAME RealSym:
::
  3GETLAM ONE pshtop&		( %&Meta )
  1GETLAM ONE SYMBN		( Meta' 'fcn' )
  TWO ' xFCNAPPLY		( Meta' 'fcn' 2 FCNAPPLY )
  4ROLL #3+			( Meta' )
;
**********************************************************************
* Desc:		Build binary function after sncknum2 failure
* Stack:	( Meta --> Meta' )
**********************************************************************
NULLNAME SymReal:
::
  3GETLAM SWAP#1+		( Meta&% )
  1GETLAM ONE SYMBN		( Meta' 'fcn' )
  TWO ' xFCNAPPLY		( Meta' 'fcn' 2 FCNAPPLY )
  4ROLL #3+			( Meta' )
;
**********************************************************************
* Desc:		Build binary function after sscknum2 failure
* Stack:	( Meta1 Meta2 --> Meta )
**********************************************************************
NULLNAME SymSym:
::
  top&				( Meta' )
  1GETLAM ONE SYMBN		( Meta' 'fcn' )
  TWO ' xFCNAPPLY		( Meta' 'fcn' 2 FCNAPPLY )
  4ROLL #3+			( Meta' )
;

**********************************************************************
**********************************************************************
* Desc:		Gamma function
* Stack:	( % --> Gamma[%] )
* Notes:	See ROM for the algorithm (SRPL)
**********************************************************************
**********************************************************************
NULLNAME Gamma
CODE

		GOSBVL	=POP1%SPLITA	AB = x+1
		GOSBVL	=RSUB1		AB = x 
		GOSUB	gSTAB0		R01 = x

		C=0	W		Load 260 to CD
		P=	16-3
		LCHEX	026
		D=C	W
		C=0	W
		C=C+1	A
		C=C+1	A
		P=	TST>
		GOSUB	gTST15		CS if x > 260
		GONC	gammanow	CC - continue

*---------------------------------------------------------------------
* x > 260 - Gamma[x+1] overflows
*---------------------------------------------------------------------
		A=0	W		Load %%MAXR to AB
		P=	4
		A=A+1	P
		B=0	W
		B=B+1	S
		BSR	W
		GOTO	gEND%		And handle the exception
*---------------------------------------------------------------------
* x <= 260 - continue
*---------------------------------------------------------------------
gammanow	GOSBVL	=CLRFRC		Clear fractional part
		GOSUB	gRCCD0		CD = x
		P=	TST=
		GOSUB	gTST15		CS if x is integer
		GONC	gammafrac	CC - handle fractional gamma
*---------------------------------------------------------------------
* Gamma[integer]
*---------------------------------------------------------------------
		?A#0	S
		GOYES	gammanegint	Gamma[-integer] = infinite
		HS=0	3
		GOSBVL	=aFCSTRT	Call internal ! for integers
		GOVLNG	=NoXcpEND%SB	Push the result

*---------------------------------------------------------------------
* Gamma[-integer] = +-infinity
*---------------------------------------------------------------------
gammanegint	GOSUB	+
		CON(5)	=DOCOL		The internal code for the case
		CON(5)	=handleDVZ
		CON(5)	=DROP
		CON(5)	=%-MAXREAL
		CON(5)	=SEMI
		CON(5)	=DOCODE
		REL(5)	->gammaend
+		C=RSTK
		A=C	A
		SETHEX
		GOSBVL	=GETPTR
		D1=D1-	5		Unpop %
		D=D+1	A
		PC=(A)			Dispatch to RPL

*---------------------------------------------------------------------
* Gamma[fraction]
*---------------------------------------------------------------------
gammafrac	GOSUB	gRCAB0		AB = x
		C=0	W		Load 260 to CD
		P=	16-3
		LCHEX	26
		D=C	W
		C=0	W
		C=C+1	A
		C=C+1	A
		C=-C-1	S
		P=	TST<
		GOSUB	gTST15		CS if |x| < 260
		GONC	gammasmall

*---------------------------------------------------------------------
* Gamma[x] where x < -260 and not integer
*---------------------------------------------------------------------
		GOSBVL	=CLRFRC		Clear fractional part
		C=0	W		Load 2 to CD
		D=0	W
		D=D+1	S
		D=D+1	S
		DSR	W
		GOSBVL	=aMODF		Modulo 2
		A=0	W		Load +-%%MAXR according to result
		P=	4
		LAHEX	9
		C=0	W
		C=C+1	S
		CSR	W
		CBEX	W
		?C=0	W
		GOYES	+
		A=R0.F	S		sign(x)
		?A=0	S
		GOYES	+		If positive then +%%MAXR
		A=-A-1	S		-%%MAXR
+		GOTO	gEND%		Handle the exception

*---------------------------------------------------------------------
* Gamma[x] where |x| < 260 and not integer
*---------------------------------------------------------------------

gammasmall	C=0	W
		P=	7
		CPEX	14
		D=C	W
		C=0	W		CD = 7
		A=-A-1	S		AB = -x
		GOSUB	gADDF		AB = 7-x
		GOSBVL	=aFLOOR
		?B=0	W
		GOYES	+
		?A=0	S
		GOYES	++
+		GOSUB	gRCAB0		AB = x
		GOSUB	GammaX
		GOTO	gEND%

++		GOSUB	gSTAB2
		GOSUB	CoerceDec
		RSTK=C			loop counter
		GOSUB	gRCAB0		x
		GOSUB	gRCCD2		int(7-x)
		GOSUB	gADDF
		GOSUB	GammaX		gamma(x+int(7-x))
		GOSUB	gSTAB2		Save it
		GOSBVL	=D1=DSKTOP
		D1=D1-	5
		A=DAT1	A
		D1=A
		D1=D1+	5
		A=DAT1	W
		GOSBVL	=SPLITA		x
*		GOSBVL	=RADD1
		GOSUB	gSTAB0		x+1
gammaloop	C=RSTK
		C=C-1	A
		?C=0	A
		GOYES	gammaok
		RSTK=C
		GOSBVL	=RADD1
		GOSUB	gRCCD0
		GOSUB	gSTAB0
		GOSUB	gMULTF
		GOSBVL	=EXAB0
		GOTO	gammaloop
gammaok		GOSBVL	=RCAB2
		GOSUB	gRCCD0
		GOSUB	gDIVF
gEND%		GOVLNG	=NoXcpEND%

*---------------------------------------------------------------------
* Jump table
*---------------------------------------------------------------------

gSTAB0		GOVLNG	=STAB0
gSTAB2		GOVLNG	=STAB2
gRCAB0		GOVLNG	=RCAB0
gRCCD0		GOVLNG	=RCCD0
gRCCD2		GOVLNG	=RCCD2
gXYEX		GOVLNG	=XYEX
gTST15		GOVLNG	=TST15
gADDF		GOVLNG	=ADDF
gMULTF		GOVLNG	=MULTF
gDIVF		GOVLNG	=DIVF

*---------------------------------------------------------------------
* Gamma calculation subroutine
* ROM:
*	1+ DUP				x+1 x+1
*	2DUP *				x+1 x+1 (x+1)^2
*	12 *				x+1 x+1 12(x+1)^2
*	DUP 60 +			x+1 x+1 12(x+1)^2 12(x+1)^2+60
*	SWAP -495.920119017593 /	x+1 x+1 12(x+1)^2+60 y
*	30.3479606073615 +		x+1 x+1 12(x+1)^2+60 y'
*	OVER +				x+1 x+1 12(x+1)^2+60 y''
*	-76.5594818140208 SWAP /	x+1 x+1 y'' z
*	9.33584905660377 +		x+1 x+1 y'' z'
*	OVER +				x+1 x+1 y'' z''
*	-1.21142857142857 SWAP /	x+1 x+1 z'' y'''
*	.4 +				x+1 x+1 z'' y''''
*	+ /				x+1 x'
*	OVER -				x+1 x''
*	.918938533204673 +		x+1 x'''
*	OVER LN				x+1 x''' ln(x+1)
*	ROT .5 -			x''' ln(x+1) x+0.5
*	* + EXP				x''''
*---------------------------------------------------------------------
GammaX		GOSBVL	=RADD1
		GOSUB	gSTAB2		X+1
		C=B	W
		D=C	W
		C=A	W
		GOSUB	gMULTF		(X+1)^2
		C=0	W
		P=	13
		LCHEX	12
		P=	0
		D=C	W
		C=0	W
		C=C+1	A
		GOSUB	gMULTF		12(X+1)^2
		GOSUB	gSTAB0
		C=0	W
		P=	14
		LCHEX	6
		P=	0
		D=C	W
		C=0	W
		C=C+1	A		60
		GOSUB	gADDF
		P=	0
		LCHEX	0495920119017593
		D=C	W
		C=0	W
		C=-C-1	S
		C=C+1	A
		C=C+1	A		-495.92...
		GOSUB	gXYEX
		GOSUB	gDIVF
		P=	0
		LCHEX	0303479606073615
		D=C	W
		C=0	A
		C=C+1	A		30.347...	
		GOSUB	gADDF
		GOSUB	gRCCD0
		GOSUB	gADDF
		P=	0
		LCHEX	0765594818140208
		D=C	W
		C=0	W
		C=-C-1	S
		C=C+1	A		-76.559..
		GOSUB	gXYEX
		GOSUB	gDIVF
		P=	0
		LCHEX	0933584905660377
		D=C	W
		C=0	W		0.9335..
		GOSUB	gADDF
		GOSUB	gRCCD0
		GOSUB	gADDF
		P=	0
		LCHEX	0121142857142857
		D=C	W
		C=0	W
		C=-C-1	S		-1.211..
		GOSUB	gXYEX
		GOSUB	gDIVF
		C=0	W
		P=	4
		CPEX	14
		D=C	W
		C=0	W
		C=C-1	A		.4
		GOSUB	gADDF
		GOSUB	gRCCD0
		GOSUB	gADDF
		GOSUB	gRCCD2
		GOSUB	gXYEX
		GOSUB	gDIVF
		GOSUB	gRCCD2
		C=-C-1	S
		GOSUB	gADDF
		P=	0
		LCHEX	0918938533204673		
		D=C	W
		C=0	W
		C=C-1	A		.9189..
		GOSUB	gADDF
		GOSBVL	=EXAB2
		GOSUB	gSTAB0
		C=0	W
		P=	5
		CPEX	14
		D=C	W
		C=0	W
		C=-C-1	S
		C=C-1	A		-.5
		GOSUB	gADDF
		AR1EX
		ABEX	W
		R4=A
		A=R0
		GOSBVL	=aLNF
		C=R4
		D=C	W
		C=R1
		GOSUB	gMULTF
		GOSUB	gRCCD2
		GOSUB	gADDF
		GOVLNG	=aEXP15F

*---------------------------------------------------------------------
* Convert dec float (integer) to decimal (DEC mode active)
* B[W] --> C[A]
*---------------------------------------------------------------------
CoerceDec	B=0	S
		B=0	A
-		BSLC	
		A=A-1	A
		GONC	-
		BSLC
		C=B	A
		RTN

*---------------------------------------------------------------------
* Code taken from COERCE
* B[W] --> A[A]
*---------------------------------------------------------------------
*	Coerce		SETHEX
*			P=	0
*			?B=0	W
*			GOYES	coerce0
*			C=0	A
*			LC(1)	6
*			?A>C	A
*			GOYES	coercemax
*			C=A	A		exp
*			D=C	A		exp
*			C=0	W
*	coerceloop	C=C+C	W
*			A=C	W
*			C=C+C	W
*			C=C+C	W
*			C=C+A	W
*			BSL	W
*			?B=0	S
*			GOYES	+
*	-		C=C+1	W
*			B=B-1	S
*			?B#0	S
*			GOYES	-
*	+		D=D-1	A
*			GONC	coerceloop
*			A=C	A
*			C=0	A
*			?C=0	W
*			RTNYES
*	coercemax	A=0	A
*			A=A-1	A
*			RTN
*	coerce0		A=0	A
*			RTN
->gammaend
ENDCODE


**********************************************************************
* Desc:		Bessel Function Calculation
* Stack:	( %n %x 1stkind? notmodified? --> % )
* Notes:	See Stegun-Abramovitz
**********************************************************************

NULLNAME Bessel
CODE

sBESNEG	EQU 12		Flag to indicate negative X (TurnOff flag, safe!)

* Flag 11 is used in the subroutines

	ABASE	#100		IRAMBUFF saves
BES_X	ALLOC	21		X (argument)
BES_N	ALLOC	21		N (degree)
BES_Z	ALLOC	21		Z (temporary save)
BES_I	ALLOC	21		I (degree counter)
BES_A	ALLOC	21		A
BES_S	ALLOC	21		S

* Determine function from flags

		GOSBVL	=popflag
		GONC	modified
		GOSBVL	=popflag
		GONC	+
		GOTO	DoJn
+		GOTO	DoYn

modified	GOSBVL	=popflag
		GONC	+
		GOTO	DoIn
+		GOTO	DoKn

*---------------------------------------------------------------------
* Bessel, 1st kind
* Args:	n >= 0, integer
*	x >= 0, real
* 	Jn(0) = finite
*	Jn(inf) = 0
*	Jn(-x ) = (-1)^n Jn(x)
*	J-n(x) = (-1)^n Jn(x)
*	J-n(-x) = Jn(x)
*---------------------------------------------------------------------
DoJn		GOSUB	popbessel
		GOSUB	checkxneg	Flag (-1)^n from x sign
		GOSUB	checknneg	Flag (-1)^n from n sign
		GOSUB	savebessel
		GOSUB	besselN
		?B=0	W
		GOYES	doJ0
		GOSUB	rsub1
		?B#0	W
		GOYES	doJn
doJ1		GOSUB	J1
		GOTO	push%
doJ0		GOSUB	J0
		GOTO	push%
doJn		GOSUB	Jn
		GOTO	push%
*---------------------------------------------------------------------
* Bessel, 2nd kind
* Args:	n >= 0, integer
*	x >= 0, real
* 	Yn(0) = -inf
*	Yn(inf) = 0
*	Yn(-x) = complex
*	Y-n(x) = (-1)^n Yn(x)
*	Y-n(-x) = complex
*---------------------------------------------------------------------
DoYn		GOSUB	popbessel
		GOSUB	checkxpos	Error if x<0
		GOSUB	checknneg	Flag (-1)^n from n sign
		GOSUB	savebessel
		GOSUB	besselX
		?B=0	W
		GOYES	doYn0
		GOSUB	besselN
		?B=0	W
		GOYES	doY0
		GOSUB	rsub1
		?B#0	W
		GOYES	doYn
doY1		GOSUB	Y1
		GOTO	push%
doY0		GOSUB	Y0
		GOTO	push%
doYn0		GOTO	push-inf
doYn		GOSUB	Yn
		GOTO	push%
*---------------------------------------------------------------------
* Modified Bessel, 1st kind
* Args:	n >= 0, integer
*	x >= 0, real
* 	In(0) = finite
*	In(inf) = inf
*	In(-x) = (-1)^n In(x)
*	I-n(x) = In(x)
*	I-n(-x) = (-1)^n In(x)
*---------------------------------------------------------------------
DoIn		GOSUB	popbessel
		GOSUB	checkxneg	Flag (-1)^n from x sign
		GOSUB	checknabs	Take abs N
		GOSUB	savebessel
		GOSUB	besselN
		?B=0	W
		GOYES	doI0
		GOSUB	rsub1
		?B#0	W
		GOYES	doIn
doI1		GOSUB	I1
		GOTO	push%
doI0		GOSUB	I0
		GOTO	push%
doIn		GOSUB	In
		GOTO	push%
*---------------------------------------------------------------------
* Modified Bessel, 2nd kind
* Args:	n >= 0, integer
*	x >= 0, real
* 	Kn(0) = inf
*	Kn(inf) = 0
*	Kn(-x) = complex
*	K-n(x) = Kn(x)
*	K-n(-x) = complex
*---------------------------------------------------------------------
DoKn		GOSUB	popbessel
		GOSUB	checkxpos	Error if x<0
		GOSUB	checknabs	Take abs N
		GOSUB	savebessel
		GOSUB	besselX
		?B=0	W
		GOYES	doKn0
		GOSUB	besselN
		?B=0	W
		GOYES	doK0
		GOSUB	rsub1
		?B#0	W
		GOYES	doKn
doK1		GOSUBL	K1
		GOTO	push%
doK0		GOSUBL	K0
		GOTO	push%
doKn0		GOTO	pushinf
doKn		GOSUBL	Kn
		GOTO	push%

*---------------------------------------------------------------------
* Return Infinite results
*---------------------------------------------------------------------
pushinf		A=0	W
		GOTO	+
push-inf	A=0	W
		A=-A-1	S
+		GOSUB	bsetxm
		P=	3
		GOVLNG	=END%
bsetxm		RTNSXM
*---------------------------------------------------------------------
* Pop and check args
*---------------------------------------------------------------------
popbessel	GOSBVL	=POP2%SPLTAC	Pop N, X
		GOSUB	stab0		Save N
		GOSBVL	=STCD2		Save X

		GOSUB	clrfrc		Check INT(N) == N
		GOSUB	rccd0
		P=	TST=
		GOSUB	tst15		INT(N) == N ?
		P=	0
		RTNC			Yes - no error

besselerr	SETHEX
		LC(5)	#304
		GOVLNG	=GPErrjmpC

*---------------------------------------------------------------------
* Various argument checking utilities
*---------------------------------------------------------------------

* Error if x is negative

checkxpos	ST=0	sBESNEG		Flag no negation
		GOSUB	rcab2		AB = X
		?A=0	S
		RTNYES
		GONC	besselerr

* Take absolute value of n

checknabs	GOSUB	rcab0		AB = N
		A=0	S
		GOTO	stab0

* Flag negation if x<0 and n is odd

checkxneg	ST=0	sBESNEG		Flag no negation
		GOSUB	rcab2		AB = X
		?A=0	S
		RTNYES			No negation if X is positive
		A=0	S
		GOSUB	stab2
		GOSUB	niseven?	CC if odd
		RTNC
setresneg	ST=1	sBESNEG		Flag result negation
		RTN

* Flag negation if n<0 and n is odd. This should be called after checking X

checknneg	GOSUB	rcab0		AB = N
		?A=0	S
		RTNYES			N is positive - done
		A=0	S
		GOSUB	stab0
		GOSUB	niseven?	CC if odd
		RTNC			CS: Even n => no negation
		?ST=0	sBESNEG
		GOYES	setresneg	negpos = neg
		ST=0	sBESNEG		negneg = pos
		RTN

* Return CS if N is even

niseven?	GOSUB	rcab0		AB = N
		P=	14		first nibble of mantissa
-		A=A-1	A		exponent--
		GOC	+
		P=P-1
		GONC	-
+		SB=0			Test lowest bit in mantissa
		BSRB.F	P
		P=	0
		?SB=0
		RTNYES			SB clear - N is even
		RTN			SB set - N is odd

*---------------------------------------------------------------------
* Save args in RX,RY
*---------------------------------------------------------------------
savebessel	D1=(5)	(=IRAM@)-4
		C=DAT1	A
		LCHEX	0100
		R4=C
		D1=C
		GOSUB	rcab2		X
		GOSUB	putab1
		GOSUB	rcab0		N
		GOTO	putab1

*---------------------------------------------------------------------
besselN		A=R4
		D1=A
		D1=(2)	BES_N
		GOTO	getab1
*---------------------------------------------------------------------
besselX		A=R4
		D1=A
		D1=(2)	BES_X
getab1		GOVLNG	=GETAB1
*---------------------------------------------------------------------
besselY		C=R4
		D1=C
		D1=(2)	BES_X
getcd1		GOVLNG	=GETCD1
*---------------------------------------------------------------------
!besselZ	AR4EX
		D1=A
		AR4EX
		D1=(2)	BES_Z
putab1		GOVLNG	=PUTAB1
*---------------------------------------------------------------------
besselZ		A=R4
		D1=A
		D1=(2)	BES_Z
		GOTO	getab1
*---------------------------------------------------------------------
!besselI	AR4EX
		D1=A
		AR4EX
		D1=(2)	BES_I
		GOTO	putab1
*---------------------------------------------------------------------
besselI		A=R4
		D1=A
		D1=(2)	BES_I
		GOTO	getab1
*---------------------------------------------------------------------
!besselA	AR4EX
		D1=A
		AR4EX
		D1=(2)	BES_A
		GOTO	putab1
*---------------------------------------------------------------------
besselA		A=R4
		D1=A
		D1=(2)	BES_A
		GOTO	getab1
*---------------------------------------------------------------------
!besselS	AR4EX
		D1=A
		AR4EX
		D1=(2)	BES_S
		GOTO	putab1
*---------------------------------------------------------------------
besselS		A=R4
		D1=A
		D1=(2)	BES_S
		GOTO	getab1
*---------------------------------------------------------------------
* J0(X)
* Uses: R0 R1 R2 R3 R4[S]
*---------------------------------------------------------------------
J0		GOSUB	besselX
		GOSUB	Load3
		GOSUB	tst<
		GOC	x<3j0
x>=3j0		GOSUB	JY0
		GOTO	J01x>=3Common
x<3j0		GOSUB	divf
		GOSUB	sqf
		GOSUBL	TabJ0x<3
		GOTO	evaluate
*---------------------------------------------------------------------
* J1(X)
* Uses: R0 R1 R2 R3 R4[S]
*---------------------------------------------------------------------
J1		GOSUB	besselX
		GOSUB	Load3
		GOSUB	tst<
		GOC	x<3j1
x>=3j1		GOSUB	JY1
J01x>=3Common	GOSUB	stab2
		GOSUB	xyex
		GOSUB	cosf
		GOSUB	rccd2
		GOSUB	multf
		GOSUB	stab2
		GOSUB	besselX
		GOSUB	sqrf
		GOSUB	rccd2
		GOTO	idivf
x<3j1		GOSUB	divf
		GOSUB	sqf
		GOSUBL	TabJ1x<3
		GOSUB	evaluate
		GOSUB	besselY
		GOTO	multf
*---------------------------------------------------------------------
* Jn(X)
* Uses: R0 R1 R2 R3 R4[S] Tmp Cnt si
*---------------------------------------------------------------------
Jn		GOSUB	besselN
		GOSUB	besselY
		P=	(TST<)!(TST=)
		GOSUB	tst15
		GONC	Jnx<=n
		GOSUB	J0
		GOSUB	!besselZ
		GOSUB	J1
		ST=1	sJYN
		GOTO	YKnCommon
Jnx<=n		GOSUB	besselX
		GOSUB	Load2
		GOSUB	idivf
		GOSUB	stab2	't=2/x'
		GOSUB	besselN
		GOSUB	Load40
		GOSUB	multf	40n
		GOSUB	sqrf
		GOSUB	clrfrc
		GOSUB	xyex
		GOSUB	besselN
		GOSUB	addf	+n
		GOSUB	div2
		GOSUB	clrfrc
		GOSUB	mult2
		GOSUB	!besselI
		A=0	W
		B=0	W	0
		GOSUB	!besselZ
		GOSUB	!besselA
		GOSUB	!besselS
		B=B+1	S
		BSR	W	1
		GOSUB	stab0
sPASS1	EQU 11
		GOSUB	besselI	m
		ST=1	sPASS1
JnLoop		GOSUB	rccd2	t
		GOSUB	multf
		GOSUB	rccd0
		GOSUB	multf
		GOSUB	xyex
		GOSUB	besselZ
		A=-A-1	S
		GOSUB	addf
		GOSBVL	=EXAB0
		GOSUB	!besselZ
		GOSUB	rcab0
		A=0	S
		GOSUB	LoadInHI
		P=	TST>
		GOSUB	tst15
		GONC	+
		GOSUB	rcab0
		GOSUB	LoadInLO
		GOSUB	multf	*d
		GOSUB	stab0
		GOSUB	besselZ
		GOSUB	LoadInLO
		GOSUB	multf	*d
		GOSUB	!besselZ
		GOSUB	besselA
		GOSUB	LoadInLO
		GOSUB	multf	*d
		GOSUB	!besselA
		GOSUB	besselS
		GOSUB	LoadInLO
		GOSUB	multf	*d
		GOSUB	!besselS
+		GOSUB	besselI
		GOSUB	xyex
		GOSUB	besselN
		P=	TST=
		GOSUB	tst15
		GONC	+
		GOSUB	besselZ
		GOSUB	!besselA
+		?ST=1	sPASS1
		GOYES	+
		GOSUB	besselS
		GOSUB	rccd0
		GOSUB	addf
		GOSUB	!besselS
		ST=1	sPASS1
		?ST=1	sPASS1
		GOYES	++
+		ST=0	sPASS1
++		GOSUB	besselI
		GOSUB	rsub1
		GOSUB	!besselI
		?B=0	W
		GOYES	JnDone
		GOTO	JnLoop
JnDone		GOSUB	besselS
		GOSUB	mult2
		GOSUB	rccd0
		C=-C-1	S
		GOSUB	addf
		GOSUB	xyex
		GOSUB	besselA
		GOTO	divf

*---------------------------------------------------------------------
* Y0(X)
* Uses: R0 R1 R2 R3 R4[S]
*---------------------------------------------------------------------
Y0		GOSUB	besselX
		GOSUB	Load3
		GOSUB	tst<
		GOC	x<3y0
x>=3y0		GOSUB	JY0
		GOTO	Y01x>=3Common
x<3y0		GOSUB	besselX
		GOSUB	J0
		GOSUB	stab2
		GOSUB	besselX
		GOSUB	div2
		GOSUB	lnf
		GOSUB	rccd2
		GOSUB	multf
		GOSUB	mult2
		GOSUB	LoadPI
		GOSUB	divf
		GOSUB	stab2
		GOSUB	besselX
		GOSUB	Load3
		GOSUB	divf
		GOSUB	sqf
		GOSUBL	TabY0x<3
		GOSUB	evaluate
		GOSUB	rccd2
		GOTO	addf
*---------------------------------------------------------------------
* Y1(X)
* Uses: R0 R1 R2 R3 R4[S]
*---------------------------------------------------------------------
Y1		GOSUB	besselX
		GOSUB	Load3
		GOSUB	tst<
		GOC	x<3y1
x>=3y1		GOSUB	JY1
Y01x>=3Common	GOSUB	stab2
		GOSUB	xyex
		GOSUB	sinf
		GOSUB	rccd2
		GOSUB	multf
		GOSUB	stab2
		GOSUB	besselX
		GOSBVL	=SQRF
		GOSUB	rccd2
		GOSUB	xyex
		GOTO	divf
x<3y1		GOSUB	besselX
		GOSUB	J1
		GOSUB	stab2
		GOSUB	besselX
		GOSUB	div2
		GOSUB	lnf
		GOSUB	rccd2
		GOSUB	multf
		GOSUB	besselY
		GOSUB	multf
		GOSUB	mult2
		GOSUB	LoadPI
		GOSUB	divf
		GOSUB	stab2
		GOSUB	besselX
		GOSUB	Load3
		GOSUB	divf
		GOSUB	sqf
		GOSUBL	TabY1x<3
		GOSUB	evaluate
		GOSUB	rccd2
		GOSUB	addf
		GOSUB	besselY
		GOTO	divf
*---------------------------------------------------------------------
* Yn(X)
* Uses: R0 R1 R2 R3 R4[S] Tmp Cnt
*---------------------------------------------------------------------

sJYN	EQU 11

Yn		GOSUB	Y0
		GOSUB	!besselZ
		GOSUB	Y1
		ST=1	sJYN
YKnCommon	GOSUB	stab0
		GOSUB	besselX
		GOSUB	Load2
		GOSUB	idivf
		GOSUB	stab2
		A=0	W
		B=0	W
		B=B+1	S
		BSR	W	cnt=1
		GOSUB	!besselI
YnLoop		GOSUB	rccd2
		GOSUB	multf
		GOSUB	rccd0
		GOSUB	multf
		GOSUB	xyex
		GOSUB	besselZ
		?ST=0	sJYN
		GOYES	+
		A=-A-1	S
+		GOSUB	addf
		GOSBVL	=EXAB0
		GOSUB	!besselZ
		GOSUB	besselI
		GOSUB	radd1
		GOSUB	!besselI
		GOSUB	xyex
		GOSUB	besselN
		GOSUB	xyex
		GOSUB	tst<
		GOC	YnLoop
		GOTO	rcab0

*---------------------------------------------------------------------
*	Jump Table
*---------------------------------------------------------------------

push%		?ST=0	sBESNEG		Negate answer if sBESNEG is set
		GOYES	+		eg if x was negative.
		A=-A-1	S		Suitable for Jn and In only
+		GOVLNG	=NoXcpEND%
tst<		P=	TST<
tst15		GOVLNG	=TST15
addf		GOVLNG	=ADDF
sqf		C=B	W
		D=C	W
		C=A	W
multf		GOVLNG	=MULTF
idivf		GOSUB	xyex
divf		GOVLNG	=DIVF
radd1		GOVLNG	=RADD1
rsub1		GOVLNG	=RSUB1
div2		GOVLNG	=DIV2
mult2		GOVLNG	=MULT2
sqrf		GOVLNG	=SQRF
stab0		GOVLNG	=STAB0
stab2		GOVLNG	=STAB2
rcab0		GOVLNG	=RCAB0
rcab2		GOVLNG	=RCAB2
rccd0		GOVLNG	=RCCD0
rccd2		GOVLNG	=RCCD2
xyex		GOVLNG	=XYEX
clrfrc		GOVLNG	=CLRFRC
cosf		ST=0	4
		ST=1	9
		GOVLNG	=aCOS15F
sinf		ST=0	4
		ST=1	9
		GOVLNG	=aSIN15F
lnf		GOVLNG	=aLNF
expf		GOVLNG	=aEXP15F

*---------------------------------------------------------------------
* I0(X)
* Uses: R0 R1 R2 R3 R4[S]
*---------------------------------------------------------------------
I0		GOSUB	besselX
		A=0	S
		GOSUB	Load3.75
		GOSUB	tst<
		GONC	i0>=3.75
i0<3.75		GOSUB	besselX
		GOSUB	divf
		GOSUB	sqf
		GOSUBL	TabI0x<3.75
		GOTO	evaluate	
i0>=3.75	GOSUBL	TabI0x>=3.75
		GOTO	i01Common
*---------------------------------------------------------------------
* I1(X)
* Uses: R0 R1 R2 R3 R4[S]
*---------------------------------------------------------------------
I1		GOSUB	besselX
		A=0	S
		GOSUB	Load3.75
		GOSUB	tst<
		GONC	i1>=3.75
i1<3.75		GOSUB	besselX
		GOSUB	divf
		GOSUB	sqf
		GOSUBL	TabI1x<3.75
		GOSUB	evaluate
		GOSUB	besselY
		GOTO	multf
i1>=3.75	GOSUBL	TabI1x>=3.75
i01Common	GOSUB	Load3.75
		GOSUB	idivf
		GOSUB	evaluate
		GOSUB	stab2
		GOSUB	besselX
		A=0	S
		GOSUB	expf
		GOSUB	rccd2
		GOSUB	multf
		GOSUB	stab2
		GOSUB	besselX
		A=0	S
		GOSUB	sqrf
		GOSUB	rccd2
		GOTO	idivf
*---------------------------------------------------------------------
* In(x)
* Uses: R0 R1 R2 R3 R4[S] Tmp Cnt
*---------------------------------------------------------------------
In		GOSUB	besselX
		GOSUB	Load2
		GOSUB	idivf
		GOSUB	stab2	't=2/x'

		GOSUB	besselN
		GOSUB	Load40
		GOSUB	multf	40n
		GOSUB	sqrf
		GOSUB	clrfrc
		GOSUB	xyex
		GOSUB	besselN
		GOSUB	addf	+n
		GOSUB	mult2
		GOSUB	!besselI
		A=0	W
		B=0	W	0
		GOSUB	!besselZ
		GOSUB	!besselA
		B=B+1	S
		BSR	W	1
		GOSUB	stab0
InLoop		GOSUB	besselI	j
		GOSUB	rccd2	t
		GOSUB	multf
		GOSUB	rccd0
		GOSUB	multf
		GOSUB	xyex
		GOSUB	besselZ
		GOSUB	addf
		GOSBVL	=EXAB0
		GOSUB	!besselZ
		GOSUB	rcab0
		A=0	S
		GOSUB	LoadInHI
		P=	TST>
		GOSUB	tst15
		GONC	+
		GOSUB	rcab0
		GOSUB	LoadInLO
		GOSUB	multf	*d
		GOSUB	stab0
		GOSUB	besselZ
		GOSUB	LoadInLO
		GOSUB	multf	*d
		GOSUB	!besselZ
		GOSUB	besselA
		GOSUB	LoadInLO
		GOSUB	multf	*d
		GOSUB	!besselA
+		GOSUB	besselI
		GOSUB	xyex
		GOSUB	besselN
		P=	TST=
		GOSUB	tst15
		GONC	+
		GOSUB	besselZ
		GOSUB	!besselA
+		GOSUB	besselI
		GOSUB	rsub1
		GOSUB	!besselI
		?B=0	W
		GOYES	InDone
		GOTO	InLoop

InDone		GOSUB	besselA
		GOSUB	rccd0
		GOSUB	divf
		GOSUB	!besselZ
		GOSUB	I0
		GOSUB	xyex
		GOSUB	besselZ
		GOTO	multf

* Load .0000000001
LoadInLO	D=0	W
		D=D+1	S
		DSR	W
		C=0	W
		C=C-1	A	99999
		P=	0
		C=0	P	99990
		RTN
* Load 10000000000
LoadInHI	D=0	W
		D=D+1	S
		DSR	W
		C=0	W
		P=	0
		LCHEX	10
		RTN
*---------------------------------------------------------------------
* K0(X)
* Uses: R0 R1 R2 R3 R4[S] Tmp Cnt
*---------------------------------------------------------------------
K0		GOSUB	besselX
		A=0	S
		GOSUB	Load2
		GOSUB	tst<
		GONC	k0x>=2
k0x<2		GOSUB	I0
		GOSUB	stab2
		GOSUB	besselX
		GOSUB	div2
		GOSUB	lnf
		A=-A-1	S
		GOSUB	rccd2
		GOSUB	multf
		GOSUB	stab2
		GOSUB	besselX
		GOSUB	div2
		GOSUB	sqf
		GOSUBL	TabK0x<2
		GOSUB	evaluate
		GOSUB	rccd2
		GOTO	addf
k0x>=2		GOSUBL	TabK0x>=2
		GOTO	K01Common
*---------------------------------------------------------------------
* K1(X)
* Uses: R0 R1 R2 R3 R4[S]
*---------------------------------------------------------------------
K1		GOSUB	besselX
		A=0	S
		GOSUB	Load2
		GOSUB	tst<
		GONC	k1x>=2
k1x<2		GOSUB	I1
		GOSUB	stab2
		GOSUB	besselX
		GOSUB	div2
		GOSUB	lnf
		GOSUB	rccd2
		GOSUB	multf
		GOSUB	stab2
		GOSUB	besselX
		GOSUB	div2
		GOSUB	sqf
		GOSUBL	TabK1x<2
		GOSUB	evaluate
		GOSUB	besselY
		GOSUB	divf
		GOSUB	rccd2
		GOTO	addf

k1x>=2		GOSUBL	TabK1x>=2
K01Common	GOSUB	Load2
		GOSUB	idivf
		GOSUB	evaluate
		GOSUB	stab2
		GOSUB	besselX
		A=-A-1	S
		GOSUB	expf
		GOSUB	rccd2
		GOSUB	multf
		GOSUB	stab2
		GOSUB	besselX
		GOSUB	sqrf
		GOSUB	rccd2
		GOTO	idivf
*---------------------------------------------------------------------
* Kn(X)
* Uses: R0 R1 R2 R3 R4[S] Tmp
*---------------------------------------------------------------------
Kn		GOSUB	K0
		GOSUB	!besselZ
		GOSUB	K1
		ST=0	sJYN
		GOTO	YKnCommon

*---------------------------------------------------------------------
Load2		C=0	W
		D=0	W
		D=D+1	S
		D=D+1	S
		DSR	W
		RTN
*---------------------------------------------------------------------
Load3		C=0	W
		D=0	W
		D=D+1	S
		D=D+1	S
		D=D+1	S
		DSR	W
		RTN
*---------------------------------------------------------------------
Load3.75	C=0	W
		P=	16-4
		LCHEX	0375
		P=	0
		D=C	W
		C=0	W
		RTN
*---------------------------------------------------------------------
LoadPI		P=	0
		LCHEX	0314159265358979
		D=C	W
		C=0	W
		RTN
*---------------------------------------------------------------------
Load40		C=0	W
		P=	16-2
		LCHEX	04
		P=	0
		D=C	W
		C=C+1	A	40
		RTN
*---------------------------------------------------------------------
* Help for J0 & Y0
* Uses: R0 R1 R2 R3 R4[S]
*---------------------------------------------------------------------
JY0		GOSUBL	besselX
		GOSUB	Load3
		GOSUB	idivf
		GOSUBL	TabJY00
		GOSUB	evaluate
		GOSUBL	besselY
		GOSUB	addf
		GOSUB	stab2
		GOSUBL	TabJY01
		GOSUB	evaluate0
		GOTO	rccd2
*---------------------------------------------------------------------
* Help for J1 & Y1
* Uses: R0 R1 R2 R3 R4[S]
*---------------------------------------------------------------------
JY1		GOSUBL	besselX
		GOSUB	Load3
		GOSUB	idivf
		GOSUBL	TabJY10
		GOSUB	evaluate
		GOSUBL	besselY
		GOSUB	addf
		GOSUB	stab2
		GOSUBL	TabJY11
		GOSUB	evaluate0
		GOTO	rccd2
*---------------------------------------------------------------------
* Evaluate polynomial in D0 at X
* Uses: R0 R1 R4[S]
*---------------------------------------------------------------------
evaluate	GOSUB	stab0
evaluate0	C=DAT1	S		Pop count
		D1=D1+	1
		R4=C.F	S		R4[S] = count
		GOSUB	poptable	CD = coefficient
-		GOSUB	rcab0		AB = X
		GOSUB	multf		AB = coeff[i]*X
		GOSUB	poptable	CD = coeff[i+1]
		GOSUB	addf		AB = coeff[i]*X+coeff[i+1]
		C=R4.F	S
		C=C-1	S		count--
		RTNC
		R4=C.F	S
		GOSUB	xyex		CD = coeff[i]*X+coeff[i+1]
		GONC	-


* Special GETCD1 substitute so we can use packed tables
* The savings for the tables are 116(4+6) = 1160 nibbles
* Note: All exponents must be <= 0 so that the sign extension works.

poptable	D1=D1-	16-11
		C=DAT1	W
		P=	6-1
		C=0	WP
		P=	0
		D=C	W		D[W] = mantissa
		C=D	S		C[S] = sign
		D=0	S		D[S] = 0
		D1=D1+	16-11
*		C=0	A		C=0 WP does the work
		C=DAT1	1		C[A] = exponent
		D1=D1+	11
		?C=0	P
		RTNYES
		C=-C	A		Sign extend to C[A]
		C=-C	P
		RTN

*---------------------------------------------------------------------
* Bessel function approximation coefficients
*---------------------------------------------------------------------

* Coefficient format:
*		CON(1)	exponent	0 or 0-e
*		CON(9)	mantissa
*		CON(1)	sign

btable:		C=RSTK
		D1=C
		RTN

TabJ0x<3	GOSUB	btable:
		CON(1)	7-2
		NIBHEX	60000000120	 2.1E-4
		NIBHEX	70000444939	-3.9444E-3
		NIBHEX	80009744440	 4.44479E-2
		NIBHEX	90066836139	-3.163866E-1
		NIBHEX	00802656210	 1.2656208
		NIBHEX	00799994229	-2.2499997
		NIBHEX	00000000010	 1

TabJ1x<3	GOSUB	btable:
		CON(1)	7-2
		NIBHEX	50000090110	 1.109E-5
		NIBHEX	60000167139	-3.1761E-4
		NIBHEX	70009133440	 4.43319E-3
		NIBHEX	80098245939	-3.954289E-2
		NIBHEX	90375390120	 2.1093573E-1
		NIBHEX	90589942659	-5.6249985E-1
		NIBHEX	90000000050	 5E-1

TabY0x<3	GOSUB	btable:
		CON(1)	7-2
		NIBHEX	60000648429	-2.4846E-4
		NIBHEX	70006197240	 4.27916E-3
		NIBHEX	80041216249	-4.261214E-2
		NIBHEX	90711003520	 2.5300117E-1
		NIBHEX	90483053479	-7.4350384E-1
		NIBHEX	90663955060	 6.0559366E-1
		NIBHEX	90196647630	 3.6746691E-1

TabY1x<3	GOSUB	btable:
		CON(1)	7-2
		NIBHEX	70000378720	 2.7873E-3
		NIBHEX	80006790049	-4.00976E-2
		NIBHEX	90015932130	 3.123951E-1
		NIBHEX	00728461319	-1.3164827
		NIBHEX	00907286120	 2.1682709
		NIBHEX	90019021220	 2.212091E-1
		NIBHEX	90089166369	-6.366198E-1

TabI0x<3.75	GOSUB	btable:
		CON(1)	7-2
		NIBHEX	70000318540	 4.5813E-3
		NIBHEX	80008670630	 3.60768E-2
		NIBHEX	90023795620	 2.659732E-1
		NIBHEX	00294760210	 1.2067492
		NIBHEX	00424998030	 3.0899424
		NIBHEX	00922651530	 3.5156229
		NIBHEX	00000000010	 1

TabI0x>=3.75	GOSUB	btable:
		CON(1)	9-2
		NIBHEX	70007732930	 3.92377E-3
		NIBHEX	80033674619	-1.647633E-2
		NIBHEX	80073553620	 2.635537E-2
		NIBHEX	80060775029	-2.057706E-2
		NIBHEX	70001826190	 9.16281E-3
		NIBHEX	70005657519	-1.57565E-3
		NIBHEX	70009135220	 2.25319E-3
		NIBHEX	80029582310	 1.328592E-2
		NIBHEX	90822498930	 3.9894228E-1

TabI1x<3.75	GOSUB	btable:
		CON(1)	7-2
		NIBHEX	60000114230	 3.2411E-4
		NIBHEX	70002351030	 3.01532E-3
		NIBHEX	80033785620	 2.658773E-2
		NIBHEX	90439480510	 1.5084934E-1
		NIBHEX	90968894150	 5.1498869E-1
		NIBHEX	90495098780	 8.7890594E-1
		NIBHEX	90000000050	 5E-1

TabI1x>=3.75	GOSUB	btable:
		CON(1)	9-2
		NIBHEX	70009500249	-4.20059E-3
		NIBHEX	80045678710	 1.787654E-2
		NIBHEX	80021359829	-2.895312E-2
		NIBHEX	80076928220	 2.282967E-2
		NIBHEX	80055513019	-1.031555E-2
		NIBHEX	70001083610	 1.638E-3
		NIBHEX	70008102639	-3.62018E-3
		NIBHEX	80042088939	-3.988024E-2
		NIBHEX	90822498930	 3.9894228E-1

TabK0x<2	GOSUB	btable:
		CON(1)	7-2
		NIBHEX	40000000470	 7.4E-6
		NIBHEX	60000057010	 1.075E-4
		NIBHEX	70008962620	 2.62698E-3
		NIBHEX	80009588430	 3.48859E-2
		NIBHEX	90657960320	 2.3069756E-1
		NIBHEX	90024872240	 4.227842E-1
		NIBHEX	90665127759	-5.7721566E-1

TabK0x>=2	GOSUB	btable:
		CON(1)	7-2
		NIBHEX	60000802350	 5.3208E-4
		NIBHEX	70000451529	-2.5154E-3
		NIBHEX	70002787850	 5.87872E-3
		NIBHEX	80064426019	-1.062446E-2
		NIBHEX	80086598120	 2.189568E-2
		NIBHEX	80085323879	-7.832358E-2
		NIBHEX	04141335210	 1.25331414

TabK1x<2	GOSUB	btable:
		CON(1)	7-2
		NIBHEX	50000068649	-4.686E-5
		NIBHEX	70004040119	-1.10404E-3
		NIBHEX	80020491919	-1.919402E-2
		NIBHEX	90798651819	-1.8156897E-1
		NIBHEX	90975872769	-6.7278579E-1
		NIBHEX	90441344510	 1.5443144E-1
		NIBHEX	00000000010	 1

TabK1x>=2	GOSUB	btable:
		CON(1)	7-2
		NIBHEX	60000542869	-6.8245E-4
		NIBHEX	70004165230	 3.25614E-3
		NIBHEX	70003530879	-7.80353E-3
		NIBHEX	80086240510	 1.504268E-2
		NIBHEX	80002655639	-3.6552E-2
		NIBHEX	90916894320	 2.3498619E-1
		NIBHEX	04141335210	 1.25331414

TabJY00		GOSUB	btable:
		CON(1)	7-2
		NIBHEX	60000855310	 1.3558E-4
		NIBHEX	60000333929	-2.933E-4
		NIBHEX	60000521459	-5.4125E-4
		NIBHEX	70003752620	 2.62573E-3
		NIBHEX	50000045939	-3.954E-5
		NIBHEX	80079366149	-4.166397E-2
		NIBHEX	90618935879	-7.8539816E-1

TabJY01		GOSUB	btable:
		CON(1)	7-2
		NIBHEX	60000674410	 1.4476E-4
		NIBHEX	60000508279	-7.2805E-4
		NIBHEX	70007327310	 1.37237E-3
		NIBHEX	50000021599	-9.512E-5
		NIBHEX	70000472559	-5.5274E-3
		NIBHEX	30000000779	-7.7E-7
		NIBHEX	90654887970	 7.9788456E-1

TabJY10		GOSUB	btable:
		CON(1)	7-2
		NIBHEX	60000661929	-2.9166E-4
		NIBHEX	60000428970	 7.9824E-4
		NIBHEX	60000843470	 7.4348E-4
		NIBHEX	70009787369	-6.37879E-3
		NIBHEX	50000005650	 5.65E-5
		NIBHEX	90216994210	 1.2499612E-1
		NIBHEX	09449165329	-2.35618449

TabJY11		GOSUB	btable:
		CON(1)	7-2
		NIBHEX	60000330029	-2.0033E-4
		NIBHEX	70003563110	 1.13653E-3
		NIBHEX	70001159429	-2.49511E-3
		NIBHEX	60000501710	 1.7105E-4
		NIBHEX	80076695610	 1.659667E-2
		NIBHEX	40000006510	 1.56E-6
		NIBHEX	90654887970	 7.9788456E-1

* Hmm, same last coefficient as in JY01. Mistake?

ENDCODE
**********************************************************************



