	STITLE	ALG Z Conversions
**********************************************************************
*		ALG library: Integer conversions
**********************************************************************

**********************************************************************
* Infinite precision integers are HXS binaries whose body contains
* the absolute value of the number first, then separate sign nibble.
* Integer is positive if sign is 0, negative if F
* Thus:
*	 1	= HXS 2 10
*	-1	= HXS 2 1F
*	 0	= HXS 1 0
*
* Naming conventions:
*	Z	Infinite precision integer
*	N	Positive infinite precision integer
*	P	Polynomial ( list of Q, where Q is P or Z )
*	Q	Z or P
**********************************************************************

**********************************************************************
* EXPORT:
*
*  CK1Z		( ob --> Z / Error )
*  CK2Z		( ob1 ob2 --> Z1 Z2 / Error )
*  CK3Z		( ob1 ob2 ob3 --> Z1 Z2 Z3 / Error )
*  H>Z		( HXS --> Z / Error )
*  #>Z		( # --> Z )
*  Z>R		( Z --> % ) 
*  R>Z		( % --> Z / Error )
*  Z>S		( Z --> $ )
*  S>Z		( S --> Z / Error )
**********************************************************************


**********************************************************************
* Desc:		Object conversions with validity checks.
* Stack:	( ob1 .. obn --> Z1 .. Zn / Error )
**********************************************************************

NULLNAME CK1Z
::
  CK&DISPATCH0
	hxs	H>Z
	real	R>Z
	str	S>Z
;

NULLNAME CK2Z
::
  CK1Z SWAP CK1Z SWAP
;

NULLNAME CK3Z
::
  CK1Z ROT CK1Z ROT CK1Z ROT
;

**********************************************************************
* Desc:		Check HXS is a proper Z number and trim it
* Stack:	( HXS --> Z / Error )
**********************************************************************
NULLNAME H>Z
::
   CODE
   		GOSBVL	=SAVPTR
		C=DAT1	A		->Z
		CD1EX
		D1=D1+	5
		A=DAT1	A		Zlen
		CD1EX
		A=A-CON	A,6		Mantissa size
		GOC	ZCKErr		
		C=C+A	A		->Ztail-6
		CD1EX
		D1=D1+	5		->Zsign
		C=DAT1	S		Zsign
		?C=0	S
		GOYES	ZCKOk
		?A=0	A		If mantissa=0 then
		GOYES	ZCKErr		   sign must be positive
		C=-C-1	S		-Zsign
		?C#0	S
		GOYES	ZCKErr
ZCKOk		GOVLNG	=GETPTRLOOP

ZCKErr		LC(5)	=SETTYPEERR
		GOVLNG	=GETPTREVALC
  ENDCODE
  ZTrim
;

**********************************************************************
* Desc:		Convert # to Z
* Stack:	( # --> Z )
**********************************************************************
NULLNAME #>Z
CODE
		A=0	W		Clear high nibbles for sign
		GOSBVL	=POP#
		GOSBVL	=SAVPTR
		P=	5-1
#>Ztrim		?A#0	P
		GOYES	#>Zpsh
		P=P-1
		GONC	#>Ztrim
#>Zpsh		P=P+1			Include sign nibble (or nib1 for 0)
		GOVLNG	=PUSHhxsLoop
ENDCODE
**********************************************************************
* Desc:		Convert Z to real
* Stack:	( Z --> % )
**********************************************************************
NULLNAME Z>R
CODE
		GOSBVL	=PopASavptr
		D1=A			->Z
		D1=D1+	5
		A=DAT1	A		Zlen
		A=A-CON	A,7
		GONC	Z2R<>0
		LC(5)	=%0
Z2Rpushob	A=C	A
		GOVLNG	=GPPushALp

Z2R<>0		R1=A	A		R1[A] = nibbles-1

		CD1EX
		C=C+A	A
		D1=C
		D1=D1+	7-1		->Zsign

		C=DAT1	S		Convert sign to DEC
		?C=0	S
		GOYES	Z2Rsignok
		P=	16-1
		LCHEX	9
		P=	0
Z2Rsignok	R2=C	S		R2[S] = result sign

		A=0	W		Init result
		B=0	W

* Add mantissa nibble into the result

zrloop		D=0	W
		P=	15-1
		D1=D1-	1
		C=DAT1	P
		D=C	P		mantissa
		C=0	S		sign = +
		C=0	A		exponent = 0

		LCHEX	A		Convert to decimal
		?D<C	P
		GOYES	Z2Rdigok

		D=D-C	P		A->0  B->1 ..
		DSR	WP
		LCHEX	1
		D=C	P		10 11 ..
		C=C+1	A		exponent = 1

Z2Rdigok	P=	0
		SETDEC
		GOSBVL	=ADDF		result += digit	(USES D0!!)
		P=	0
		SETHEX

* Mantissa done - or *16

		C=R1	A		counter
		C=C-1	A
		GOC	zrsign
		R1=C	A

		C=0	W
		P=	14-1
		LCHEX	16
		D=C	W
		P=	0
		C=C+1	A		exponent = 1
		SETDEC
		GOSBVL	=MULTF
		P=	0
		SETHEX

* Check if overflow

		LCHEX	00499
		?A<=C	A
		GOYES	zrloop

		LC(5)	=%MAXREAL	Assume 9.999..E499
		A=R2	S		Sign
		?A=0	S
		GOYES	Z2Rmaxok
		LC(5)	=%-MAXREAL	Return -9.999..E499
Z2Rmaxok	GOTO	Z2Rpushob

zrsign		A=R2.F	S		Set sign
		GOSBVL	=PACK
		GOVLNG	=PUSH%LOOP



ENDCODE

**********************************************************************
* Desc:		Convert real to Z
* Stack:	( % --> Z )
**********************************************************************
NULLNAME R>Z

CODE
		GOSBVL	=POP1%SPLITA
		?B#0	W
		GOYES	R2Z<>0
		A=0	A
		GONC	R2Zexit		Exit with P=0
R2Z<>0		R0=A	S		Save sign
		GOSBVL	=CLRFRC
		GONC	R2Zerr		Nonzero fractional part
		?P=	15
		GOYES	R2Zerr		Infinite number
		?P=	1
		GOYES	R2Zerr		One digit is lost
R2Zshiftlp	BSR	W		Shift mantissa down to word start
		P=P-1
		GONC	R2Zshiftlp
		C=B	W
*		P=	0
		GOSBVL	=DCHXW		Convert to hex
* Common small Z push exit
		P=	15		(This could be smaller)
R2Ztrimlp	P=P-1			Strip loop
		?A=0	P
		GOYES	R2Ztrimlp
		P=P+1			Sign nibble
		C=R0	S		Saved sign
		?C=0	S
		GOYES	R2Zexit
		A=-A-1	P		Negative
R2Zexit		SETHEX
		GOVLNG	=PUSHhxsLoop
R2Zerr		P=	0
		LC(5)	#203		"Bad Argument Value"
		GOVLNG	=GPErrjmpC
ENDCODE

**********************************************************************
* Desc:		Convert Z into decimal in a string
* Stack:	( Z --> String )
**********************************************************************
NULLNAME Z>S

* Converstion using Horner method

ASSEMBLE
sPOS	EQU 0		Flag to indicate positive integer
RPL
::
 H>Z DupQIsZero? casedrop tok0
 CODE
		GOSBVL	=SAVPTR
		A=DAT1	A
		D0=A
		D0=D0+	5
		A=DAT0	A
		A=A-CON	A,7
		R2=A	A		mantissa nibbles -1
		ASRB.F	A
		ASRB.F	A
		ASRB.F	A
		R1=A	A		number of 8-nibble units -1
		A=A+1	A
		C=A	A
		CSL	A
		C=C+C	A		enough for certain
		RSTK=C
		GOSBVL	=MAKE$N		R0[A] = ->output
		CD0EX
		D1=C
		C=RSTK
		GOSBVL	=WIPEOUT
	
		GOSBVL	=D1=DSKTOP
		A=DAT1	A
		D0=A
		GOSBVL	=SKIPOB

		D0=D0-	1		Setup sign flag
		A=DAT0	S
		ST=1	sPOS		Assume positive
		?A=0	S
		GOYES	todecsignok
		ST=0	sPOS		Negative
todecsignok

* Loop:
*	Add nibble to result
*	If more nibbles expected then result *= 16
*	Loop until no nibbles expected

todeclp		D0=D0-	1
		A=0	W
		A=DAT0	1		nibble to add
		LC(1)	9
		?A<=C	P
		GOYES	todecdigok
		A=A+CON	B,6		A-F to 10-15
todecdigok	C=R0	A
		D1=C
		D1=D1+	10		->output
		SETDEC			Addition loop, finishes eventually
		C=DAT1	W		since sufficient space is guaranteed
		C=C+A	W
		DAT1=C	W
		GONC	todec+ok
todec+loop	D1=D1+	16
		C=DAT1	W
		C=C+1	W
		DAT1=C	W
		GOC	todec+loop
todec+ok	SETHEX

		C=R2	A
		C=C-1	A		nibbles--
		R2=C	A
		GOC	todecfin	No more - finish up

		C=R0	A
		D1=C
		D1=D1+	10		->output
		C=R1	A
		D=C	A		number of 8-nibble units

		C=R2	A		nibs left
		CSRB.F	A		Decrement since result isn't full yet
		CSRB.F	A		(eg high nibbles are known to be zero)
		CSRB.F	A
		D=D-C	A		cntr
		B=0	W		Init last carry
todec*lp	A=0	B		Init new carry
		SETDEC
		C=DAT1	W		Multiply word by 16 with carry in A[B]
		C=C+C	W
		GONC	todec*2
		A=A+1	B
todec*2		A=A+A	B
		C=C+C	W
		GONC	todec*4
		A=A+1	B
todec*4		A=A+A	B
		C=C+C	W
		GONC	todec*8
		A=A+1	B
todec*8		A=A+A	B
		C=C+C	W
		GONC	todec*16
		A=A+1	B
todec*16	C=C+B	W		Add old carry
		GONC	todec*cry
		A=A+1	B
todec*cry	B=A	B		Set new old carry
		DAT1=C	W
		D1=D1+	16	
		SETHEX
		D=D-1	A
		GONC	todec*lp	Loop until only 0-nibbles to come
		GOTO	todeclp
* Finish up:
*	Reverse nibbles in the result
*	Respace into characters
*	Add sign
*	Shrink & exit

todecfin	C=R0
		D1=C
		D1=D1+	10
		D0=C
		GOSBVL	=SKIPOB		->output end
todectrimlp	D0=D0-	1		Stripping loop
		C=DAT0	B
		?C=0	P
		GOYES	todectrimlp
		CD0EX
		D0=C
		C=C+1	A
		D=C	A		->result tail
todecrevlp	C=DAT0	B		Reversing loop
		A=DAT1	B
		DAT1=C	1
		DAT0=A	1
		D1=D1+	1
		D0=D0-	1
		AD0EX
		D0=A
		CD1EX
		D1=C
		?A>C	A
		GOYES	todecrevlp	Loop until pointers cross

		C=D	A		->result tail
		A=R0	A
		C=C-A	A		size of result + prologs
		C=C-CON	A,10		size of result
		CDEX	A
		D0=C			->result tail
		C=C+D	A		->result tail in chars
		?ST=1	sPOS
		GOYES	todecroomok
		C=C+1	A		add room for minus sign
		C=C+1	A
todecroomok	R1=C	A		->result tail (full)
		D1=C
		A=R0	A
		A=A+CON	A,10		->output
		LCASC	'0'		Init #30 for character digits
		CD0EX	
todecasclp	CD0EX
		D0=D0-	1
		D1=D1-	2
		C=DAT0	1
		DAT1=C	B
		CD0EX
		?C>A	A
		GOYES	todecasclp

		?ST=1	sPOS		And add sign if negative
		GOYES	todecexit
		D1=D1-	2
		LCASC	'-'
		DAT1=C	B
todecexit	C=R1	A
		D0=C			->result tail (full)
		GOSBVL	=Shrink$
		GOVLNG	=GPOverWrR0Lp
 ENDCODE
;
**********************************************************************
* Desc:		Convert decimal in a string into Z
* Stack:	( String --> Z )
**********************************************************************
NULLNAME S>Z

* Conversion using Horner method

CODE
		GOSBVL	=SAVPTR
		A=DAT1	A
		D0=A
		D0=D0+	5
		C=DAT0	A		$len
		C=C-CON	A,5		$nibbles
		?CBIT=1	0
		GOYES	toherr		Error if not even nibbles
		?C=0	A
		GOYES	toherr		Error if empty $
		CSRB.F	A
		R1=C	A		chars
		CSR	A
		C=C+1	A
		R2=C	A		words
		CSL	A
		C=C+1	A		Enough room for certain
		RSTK=C
		GOSBVL	=MAKE$N
		CD0EX
		D1=C
		C=RSTK
		GOSBVL	=WIPEOUT

		GOSBVL	=D1=DSKTOP
		A=DAT1	A
		D0=A
		D0=D0+	10		->1st char

		ST=1	sPOS		Assume +
		A=DAT0	B
		LCASC	'-'
		?A#C	B
		GOYES	tohexpos?
		ST=0	sPOS
		GONC	skpsgn
tohexpos?	LCASC	'+'		Allow "+" too
		?A#C	B
		GOYES	tohloop
skpsgn		D0=D0+	2
		C=R1
		C=C-1	A		chars-- due to sign
		R1=C
		?C#0	A
		GOYES	tohloop		Error if only sign
toherr		LC(5)	=SYNTAXERR
		GOVLNG	=GETPTREVALC

* Loop:
*	Add digit to result
*	If more digits expected result *= 10
*	Loop until no more digits

tohloop		A=0	W
		A=DAT0	B		digit
		D0=D0+	2

		LCASC	'0'		Convert to hex with error checks
		A=A-C	B
		GOC	toherr
		LC(2)	9
		?A>C	B
		GOYES	toherr

		C=R0
		D1=C
		D1=D1+	10		->output

		C=DAT1	W		Add to result - loop WILL terminate
		C=C+A	W
		DAT1=C	W
		GONC	tohex+ok
tohex+loop	D1=D1+	16
		C=DAT1	W
		C=C+1	W
		DAT1=C	W
		GOC	tohex+loop
tohex+ok
		C=R1	A		chars to go--
		C=C-1	A
		R1=C	A
		?C=0	A
		GOYES	tohfinish
* Result *= 10
* Note how 15 is used to simplify carry, this is safe since
* hex integer will be shorten more than that, always
		C=R0
		D1=C
		D1=D1+	10		->output
		C=R2	A
		D=C	A		words
		D=D-1	A
		B=0	W		init carry
tohex*loop	C=0	W
		C=DAT1	15
		C=C+C	W
		A=C	W		2*
		C=C+C	W
		C=C+C	W		8*
		C=C+A	W		10*
		C=C+B	W
		DAT1=C	15
		CSLC			Now get new carry from C[S]
		B=C	P
		D1=D1+	15
		D=D-1	A
		GONC	tohex*loop
		GOTO	tohloop

* Finish up:
*	Strip result, handle zero & sign
*	Shrink & exit

tohfinish	C=R0
		D0=C			->$
		LC(5)	=DOHSTR		Convert to Z
		DAT0=C	A
		GOSBVL	=SKIPOB
tohextrimlp	D0=D0-	1		Stripping loop
		C=DAT0	B
		?C=0	P
		GOYES	tohextrimlp
		D0=D0+	1
		CD0EX			C[A] = ->result tail
		D0=C
		A=R0
		C=C-A	A		result size (w/ prologs)
		C=C-CON	A,10		result size
		GONC	tohexput	No overflow - not zero result
		D0=A
		D0=D0+	11		Just use one nibble of the 0s
		GONC	tohexexit
tohexput	C=0	S		Set sign
		?ST=1	sPOS
		GOYES	tohexsgnput
		C=-C-1	S
tohexsgnput	DAT0=C	S
		D0=D0+	1
tohexexit	GOSBVL	=Shrink$
		GOVLNG	=GPOverWrR0Lp
ENDCODE
**********************************************************************

