	STITLE	ALG Matrix
**********************************************************************
*		ALG Library: Matrix Subroutines
**********************************************************************

**********************************************************************
* EXPORT:
*
*   MDim	( A --> n m )
*   MIdn	( #n --> In )
*   MTrn	( A --> A' )
*   MAdd	( A B --> C=A+B )
*   MSub	( A B --> C=A-B )
*   MNeg	( A --> -A )
*   MAddS	( A x --> A+xI )
*   MSubS	( A x --> A-xI )
*   SSubM	( x A --> xI-A )
*   MMulS	( A s --> s*A )
*   MDivS	( A s --> A/s )   
*   MMul	( A B --> A*B )
*   MPow%	( M % --> M^% )
*   MLU		( M --> M' %rank )
*   MInv	( A --> A^-1 )
*   MDiv	( B A --> C=A^-1 * B )
*   MDet	( A --> det[A] )
**********************************************************************

**********************************************************************
* Desc:		Check array for unquoted constants, and if there are
*		any enclose them in quotes.
* Stack:	( A --> A' )
**********************************************************************
NULLNAME MCKConst
::
  CODE
		GOSBVL	=SAVPTR
		A=DAT1	A
		D0=A			D0 = ->arry
		D=0	A		D[A] = depth
--		D0=D0+	5		Check the addresses of all elements
		A=DAT0	A
-		LC(5)	=DOLIST
		?A#C	A
		GOYES	+
		D=D+1	A
		GONC	--
+		LC(5)	=xPI		If address of constant then
		?A=C	A
		GOYES	++		return TRUE
		LC(5)	=xi
		?A=C	A
		GOYES	++
		LC(5)	=xCONSTANTe
		?A=C	A
		GOYES	++
		GOSBVL	=SKIPOB		Else skip to next element
		A=DAT0	A		SEMI is *not* skipped by SKIPOB!
		LC(5)	=SEMI
		?A#C	A
		GOYES	-
		D=D-1	A
		GONC	--		.. so it is ok not to check for "; ;"
		A=-A-1	A		clear carry
++		GOVLNG	=GPPushT/FLp
  ENDCODE
  NOT?SEMI                      ( Abort if elements are fine )
  '
  :: REQcase xPI ONESYMBN
     REQcase xi ONESYMBN
     REQcase xCONSTANTe ONESYMBN
  ;
  DoP
;

**********************************************************************
* Desc:		Check whether all rows of A are vectors with same lenght m
* Stack:	( A m --> A / Error )
**********************************************************************
NULLNAME MCKRows
::
  DUP#0=case SETTYPEERR			( Bad Arg Type error )
  SWAP FIVE				( m A offset )
  BEGIN
    NEXTCOMPOB
  WHILE					( m A offset' ob )
    DUPTYPELIST? 			( m A offset' ob flag )
    NcaseTYPEERR			( Bad Arg Type error )
    LENCOMP 4PICK #=			( m A offset' flag )
    NcaseTYPEERR			( Bad Arg Type error )
  REPEAT				( m A )
  SWAPDROP
;

**********************************************************************
* Desc:		Check whether A is proper square matrix
* Stack:		( A --> A / Error )
**********************************************************************
NULLNAME MCKSq
::
  MCKConst
  DUPLENCOMP		( A n )
  MCKRows
;

**********************************************************************
* Desc:		Check that A is a proper matrix.
*		Transform vectors into 1-column matrices.
* Stack:	( A --> A' / Error )
**********************************************************************
NULLNAME MCKMat
::
  MCKConst
  DUP CARCOMP				( A A1 )
  DUPTYPELIST?
  case :: LENCOMP MCKRows ;		

  DROPZERO		( A offset )
  BEGIN
    NEXTCOMPOB
  WHILE			( A offset' ai )
    ONE{}N		( A offset' ai' )
    UNROT		( ai' A offset )
  REPEAT		( a1'..an' A )
  LENCOMP {}N		( A' ) 
;

**********************************************************************
* Desc:		Check that A and B are proper matrices of the same size
* Stack:	( A B --> A' B' / Error )
**********************************************************************
NULLNAME MCKSame
::
  SWAP MCKMat		( B A' )
  DUP MDim		( B A' na ma )
  4ROLL MCKMat		( A' na ma B' )
  DUP MDim		( A' na ma B' nb mb )
  4ROLL #=		( A' na B' nb sm )
  SWAP4ROLL #=		( A' B' sm sn )
  AND ?SKIP ErrBadDim
;

**********************************************************************
* Desc:		Get matrix dimensions
* Stack:	( A --> n m )
**********************************************************************
NULLNAME MDim
::
  DUPLENCOMP		( A n )
  SWAP CARCOMP		( n a1 )
  LENCOMP		( n m )
;

**********************************************************************
* Desc:		If A is a 1-column matrix transforms it into a vector
* Stack:	( A --> A' )
**********************************************************************
NULLNAME MNorm
::
  DUP CARCOMP LENCOMP	( A m )
  #1<> ?SEMI 
  FIVE			( A offset )
  BEGIN
    NEXTCOMPOB
  WHILE			( A offset' ai )
    CARCOMP		( A offset' ai' )
    UNROT		( ai' A offset' )
  REPEAT		( a1'..an' A )
  LENCOMP {}N		( A' ) 
;

**********************************************************************
* Desc:		Creates a n x n identity matrix
* Stack:	( #n --> In )
**********************************************************************
NULLNAME MIdn
::
  1LAMBIND
    1GETLAM			( n )
    #1+_ONE_DO (DO)	
      1GETLAM			( n )
      #1+_ONE_DO (DO)	
      	JINDEX@ INDEX@ 		( ... i j )
	#=ITE %1 %0		( ... Iij )
      LOOP			( ... Ii1 .. Iin )
      1GETLAM {}N 		( ... Ii )
    LOOP			( I1 .. In )
  1GETABND {}N			( In )	
;

**********************************************************************
* Desc:		Transpose a matrix - no check
* Stack:	( A --> A' )
**********************************************************************
NULLNAME MTrn
::
  DUP MDim			( A n m )
  { NULLLAM NULLLAM NULLLAM }
  BIND
    1GETLAM			( m )
    #1+_ONE_DO (DO)
      2GETLAM			( n )
      #1+_ONE_DO (DO)
  	3GETLAM			( A )
	INDEX@ NTHCOMPDROP	( Aj )
	JINDEX@ NTHCOMPDROP	( aji )
      LOOP
      2GETLAM {}N		( A'i )
    LOOP
  1GETABND			( A'1 .. A'm m )
  {}N 				( A' )
;

**********************************************************************
* Desc:		Create a matrix by applying f to each pair of
*		corresponding elements in A and B
* Stack:	( A B f --> C=[f[aij,bij]] )
**********************************************************************
NULLNAME MApp2
::
  OVER MDim					( A B f n m )
  { NULLLAM NULLLAM NULLLAM NULLLAM NULLLAM }
  BIND
    2GETLAM					( n )
    #1+_ONE_DO (DO)
	1GETLAM					( m )
	#1+_ONE_DO (DO)
		5GETLAM JINDEX@ NTHCOMPDROP	( ai )
		INDEX@ NTHCOMPDROP		( aij )
		4GETLAM JINDEX@ NTHCOMPDROP	( aij bi )
		INDEX@ NTHCOMPDROP		( aij bij )
		3GETLAM EVAL ARSim		( cij )
	LOOP
	1GETLAM	{}N				( ci )
     LOOP
     2GETLAM {}N				( C )
   ABND
;

**********************************************************************
* Desc:		Add two matrices
* Stack:	( A B --> C=A+B )
**********************************************************************
NULLNAME MAdd
::
   MCKSame
   RESOROMP XAdd MApp2
   MNorm
;

**********************************************************************
* Desc:		Substract two matrices
* Stack:	( A B --> C=A-B )
**********************************************************************
NULLNAME MSub
::
   MCKSame
   RESOROMP XSub MApp2
   MNorm
;

**********************************************************************
* Desc:		Negate a matrix
* Stack:	( A --> -A )
**********************************************************************
NULLNAME MNeg
:: MCKConst RESOROMP XNeg DoP ;

**********************************************************************
* Desc:		Compute C by applying f to each elem of the diag of A
* Stack:	( A s f --> C )
**********************************************************************
NULLNAME MDoDiagS
::
  { NULLLAM NULLLAM }
  BIND
    INNERCOMP #1+			( a1 .. an n+1 )
    DUP ONE_DO (DO)
	DUPROLL				( a2 .. an n+1 ai )
	DUPINDEX@ NTHCOMPDROP		( a2 .. an n+1 ai aii )
	2GETLAM 1GETLAM EVAL ARSim	( a2 .. an n+1 ai cii )
	INDEX@ ROT PUTLIST		( a2 .. an n+1 ci )
  	SWAP				( a2 .. an ci n+1 )
    LOOP				( c1 .. cn n+1 )
    #1- {}N
  ABND
;

**********************************************************************
* Desc:		Add scalar to matrix
* Stack:	( A x --> A+xI )
**********************************************************************
NULLNAME MAddS
::
  SWAP MCKSq SWAP
  RESOROMP XAdd MDoDiagS
;

**********************************************************************
* Desc:		Substract scalar from matrix
* Stack:	( A x --> A-xI )
**********************************************************************
NULLNAME MSubS
::
  SWAP MCKSq SWAP
  RESOROMP XSub MDoDiagS
;

**********************************************************************
* Desc:		Substract scalar from matrix
* Stack:	( x A --> xI-A )
**********************************************************************
NULLNAME SSubM
::
  MCKSq MNeg			( x -A )
  SWAP RESOROMP XAdd MDoDiagS	( -A+xI )
;

**********************************************************************
* Desc:		Multiply matrix by a scalar
* Stack:	( A s --> s*A )
**********************************************************************
NULLNAME MMulS
:: ' :: XMul ARSim ; DoPS ;

**********************************************************************
* Desc:		Divide matrxi by a scalar
* Stack:	( A s --> A/s )
**********************************************************************
NULLNAME MDivS
:: ' :: XDiv ARSim ; DoPS ;

**********************************************************************
* Desc:		Multiply two matrices - no checks
* Stack:	( A B --> A*B )
* Future:
*	Possibly explode both arrays, CACHE elements to LAMs,
*	and then just calculate the GETLAM indexes in the loops.
**********************************************************************
NULLNAME MMul0
::
  DUP MDim SWAPDROP				( A B mb )
  ROTDUP MDim					( B mb A na ma )

  { NULLLAM NULLLAM NULLLAM NULLLAM NULLLAM }
  BIND
    2GETLAM					( na )
    #1+_ONE_DO (DO)
    	3GETLAM INDEX@ NTHCOMPDROP		( Ai )
	4GETLAM					( Ai mb )
	#1+_ONE_DO (DO)			
	   %0 1GETLAM				( Ai %0 ma )
	   #1+_ONE_DO (DO)			
		OVER INDEX@ NTHCOMPDROP		( Ai s aik )
		5GETLAM INDEX@ NTHCOMPDROP	( Ai s aik Bk )
		JINDEX@ NTHCOMPDROP		( Ai s aik bkj )
		XMul XAdd			( Ai s' )
	   LOOP
	   ARSim				( Ai cij )
	   SWAP					( ci1..cij Ai )
	LOOP
	DROP					( ci1..cim )
	4GETLAM {}N				( Ci )
     LOOP					
     2GETLAM {}N				( C )
   ABND
;

**********************************************************************
* Desc:		Multiply two matrices
* Stack:	( A B --> A*B )
**********************************************************************
NULLNAME MMul
::
   SWAP MCKMat			( B A )
   SWAP MCKMat			( A B )
   DUPLENCOMP			( A B nb )
   3PICK MDim	 		( A B nb na ma )
   SWAPDROP			( A B nb ma )
   #<>case ErrBadDim		( Invalid Dimensions )

   MMul0
   MNorm
;

**********************************************************************
* Desc:		Compute the # power of M
* Stack:	( M # --> M^# )
**********************************************************************
NULLNAME MPow#
::
  DUP#1= caseDROP			( return M )
  TWO #=casedrop :: DUP MMul0 ; 	( return M*M )

  OVER LENCOMP MIdn			( M # In )
  UNROT					( In M # )
  BEGIN
	DUP#0<>				( C M k flag )
  WHILE
	DUP ONE #AND #0<>		( C M k flag )
	IT  ::
		OVER 4ROLL		( M k M C )
		MMul0			( M k C' )
		UNROT			( C' M k )
	    ;
	SWAPDUP MMul0			( C' k M' )
	SWAP #2/			( C' M' k' )
  REPEAT				( C M' #0 )
  2DROP					( C )
;

**********************************************************************
* Desc:		Compute the % power of M
* Stack:	( M % --> M^% )
**********************************************************************
NULLNAME MPow%
::
   DUP %0< ITE 
     :: %CHS SWAP MInv ;	( |%| M^-1 )
   ( ELSE )
     :: SWAP MCKSq ;		( % M' )

   SWAP COERCE MPow#		( M^% )
;

**********************************************************************
* Desc:		LU decomposition
* Stack:	( M --> M' Ord %rank )
**********************************************************************
NULLNAME MLU
::
  MCKSq				( M )
  DUP MAvars 			( M vars )
  DUP1LAMBIND
    MA2MM			( MM )
    MMlurank	 		( MM' #rank )
    UNCOERCE			( MM' %rank )
  1GETABND 			( MM' %rank vars )
  MM2MA				( %rank M' )
  SWAP				( M' %rank )
;

**********************************************************************
* Desc:		Matrix Inverse
* Stack:	( M --> M' )
**********************************************************************
NULLNAME MInv
::
  MCKSq				( M )
  DUP MAvars 			( M vars )
  1LAMBIND			( M )
      DUPLENCOMP MIdn		( M Midn )
      MAconcat			( M' )
      1GETLAM MA2MM		( MM' )
      MMsolve			( MM'' )
  1GETABND 			( MM'' vars )
  MM2MA				( M' )
;

**********************************************************************
* Desc:		Multiply B by the inverse of A
* Stack:	( B A --> C = A^-1 * B )
**********************************************************************
NULLNAME MDiv
::
  MCKSq				( B A )
  SWAP MCKMat			( A B )
  DUPLENCOMP			( A B nb )
  3PICK LENCOMP			( A B nb na )
  #<>case ErrBadDim		( Invalid Dimensions )
  MAconcat			( A&B )
  DUP MAvars 
  DUP1LAMBIND
     MA2MM			( MM )
     MMsolve			( MM' )
  1GETABND 
  MM2MA				( M' )
  MNorm				( C )
;

**********************************************************************
* Desc:		Symbolic matrix determinant
* Stack:	( M --> symb )
**********************************************************************
NULLNAME MDet
::
  MCKSq				( M )
  DUP MAvars 			( M vars )
  DUP1LAMBIND			( M vars )
    MA2MM			( MM )
    MMdet			( nom den )
  1GETABND R2A			( symb )
;

**********************************************************************
* Name:		MAvars
* Desc:		Get variables from elements of symbolic matrix
* Stack:	( MA --> Lvars )
**********************************************************************
NULLNAME MAvars
::
  NULL{} 1LAMBIND			( Init variables in 1LAM )
  FIVE BEGIN
    NEXTCOMPOB
  WHILE
    FIVE BEGIN
      NEXTCOMPOB
    WHILE
      CK&DISPATCH0
        real DROP
        idnt :: 1GETSWAP AppendList 1PUTLAM ;
        lam  :: 1GETSWAP AppendList 1PUTLAM ;
        symb :: INNERCOMP PVarSub ;
    REPEAT  
    DROP
  REPEAT
  DROP
  1GETABND				( Lvars )
  RESOROMP IsV>V? SortList		( Lvars' )
;

**********************************************************************
* Name:		MAconcat
* Desc:		Concatenate two matrices row by row
* Stack:	( MA MB -> MC )  
**********************************************************************
NULLNAME MAconcat
::
  DUPLENCOMP
  #1+_ONE_DO (DO)
    OVER INDEX@ NTHCOMPDROP
    OVER INDEX@ NTHCOMPDROP
    &COMP UNROT
  LOOP
  DROP LENCOMP {}N
;

**********************************************************************
