	STITLE	ALG Q Arithm
**********************************************************************
*		ALG library: Polynomial/Integer arithmetic routines
**********************************************************************

**********************************************************************
* Infinite precision integers are HXS binaries whose body contains
* the absolute value of the number first, then separate sign nibble.
* Integer is positive if sign is 0, negative if F
* Thus:
*	 1	= HXS 2 10
*	-1	= HXS 2 1F
*	 0	= HXS 1 0
*
* Multivariate polynomials are represented as lists of integers so
* that an embedded list indicates a change in variable. The variables
* themselves are stored in a separate list. Note especially that the
* list of variables is not used in basic operations such as addition.
* For example:
*	vars:	{ X Y }
*	poly:	{ { HXS 2 10 HXS 2 2F } HXS 2 30 { HXS 2 40 HXS 2 50 } }
*	symb:	X^2*Y - 2*X^2 + 3*X + 4*Y + 5
*
* Naming conventions:
*	Z	Infinite precision integer
*	N	Positive infinite precision integer
*	P	Polynomial ( list of Q, where Q is P or Z )
*	Q	Z or P
**********************************************************************

**********************************************************************
* EXPORT:
*
*  QSub		( Q1 Q1 --> Q1-Q2 )
*  QAdd		( Q1 Q2 --> Q1+Q2 )
*  QNeg		( Q --> -Q )
*  QMul		( Q1 Q1 --> Q1*Q2 )
*
**********************************************************************

**********************************************************************
* Desc:		General infinite precision polynomial substraction
* Stack:	( Q1 Q1 --> Q' )
**********************************************************************
NULLNAME QSub
CODE

sQMULREG	EQU 6		Flag for marking register QMul
sQSUB		EQU 7		Flag for marking sub/add
sQSWAP		EQU 8		Flag for marking swapped D0/D1
sQSWAP*		EQU 9		Flag for marking swapped D0/D1 in QMUL

		ST=1	sQSUB
		GOTO	QAdd/Sub
ENDCODE

**********************************************************************
* Desc:		General infinite precision polynomial addition
* Stack:	( Q1 Q1 --> Q' )
**********************************************************************
NULLNAME QAdd

CODE

		ST=0	sQSUB

QAdd/Sub	GOSUBL	AttnAbort?
		GOSBVL	=SAVPTR

* Optimized entry code to check for following special cases:
*
*	Z0+Q2	--> Q2		Z0-Q2	--> -Q2	(QNeg)
*	Q1+Z0	--> Q1		Q1-Z0	--> Q1

		A=DAT1	A
		D0=A			->Q2
		LCZ0
		A=C	W
		A=DAT0	11
		?A=C	W
		GOYES	Q1+-Z0
		D1=D1+	5
		A=DAT1	A
		D0=A			->Q1
		A=DAT1	11
		?A=C	W
		GOYES	Z0+-Q2
		GOSUBL	QSKIP
		A=DAT1	A		->Q1
		CD0EX
		C=C-A	A		Q1size
		D=C	A
		D1=D1-	5
		A=DAT1	A		->Q2
		D0=A
		GOSUBL	QSKIP
		A=DAT1	A		->Q2
		CD0EX
		C=C-A	A		Q2size
		C=C+D	A		Q1+Q2size
		GOSBVL	=GETTEMP
		CD0EX
		R3=C	A		->output
		R4=C	A		->output
		GOSBVL	=D1=DSKTOP
		C=DAT1	A
		D1=D1+	5
		A=DAT1	A
		D0=A			->Q1
		D1=C			->Q2
		GOSUBL	QADD/SUB
		A=R4	A		->Q
		GOTO	QMULexit	Push result, shrink & exit

Q1+-Z0		GOTO	Q1*Z1
Z0+-Q2		GOSBVL	=GETPTR
		A=DAT1	A
		D1=D1+	5
		D=D+1	A
		DAT1=A	A
		?ST=1	sQSUB
		GOYES	qnegentry
		GOVLNG	=Loop
ENDCODE

**********************************************************************
* Desc:		General infinite precision polynomial negation
* Stack:	( Q --> Q' )
* Algorithm:	For each Z in Q
*			If Z0 then copy to output
*			Else copy to output & negate sign nibble
* Notes:
*		For speed Q is copied as a whole, then the sign nibbles
*		are negated. The result is normalized only if input is.
**********************************************************************
NULLNAME QNeg

CODE
qnegentry	GOSUBL	AttnAbort?
		GOTO	DoQNeg		Embedded in QMul as a special case
ENDCODE

**********************************************************************
* Desc:		General infinite precision polynomial multiplication
* Stack:	( Q1 Q2 --> Q' )
**********************************************************************
NULLNAME QMul

CODE

* Optimized entry code to check for following special cases:
*
*	Z1*Q2	--> Q2		Z0*Q2	--> Z0
*	Q1*Z1	--> Q1		Q1*Z0	--> Z0
*	Z-1*Q2	--> -Q2
*	Q1*Z-1	--> -Q1
*
*	And short Z1*Z2 multiplication with the condition that
*	input and output are known to fit into registers

		GOSUBL	AttnAbort?
		GOSBVL	=SAVPTR

		ST=0	sGARB		Flag garbage collection not done

		C=DAT1	A		->Q2
		D1=D1+	5
		A=DAT1	A		->Q1
		D0=A			->Q1
		D1=C			->Q2

		A=DAT0	11
		C=A	W
		LCZ0
		?A=C	W
		GOYES	Z0*Q2
		A=DAT1	11
		?A=C	W
		GOYES	Q1*Z0
		A=DAT0	12
		LCZ1
		?A=C	W
		GOYES	Z1*Q2
		A=DAT1	12
		?A=C	W
		GOYES	Q1*Z1
		P=	12-1
		C=-C-1	P		C[W] = Z-1
		P=	0
		?A=C	W
		GOYES	Q1*Z-1
		A=DAT0	12
		?A=C	W
		GOYES	Z-1*Q2
		GOTO	tstshortmul

* Return stack level 2 as result
Q1*Z1
Z0*Q2		GOSBVL	=GETPTR
		D1=D1+	5
		D=D+1	A
		GOVLNG	=Loop
* Return stack level 1 as result
Z1*Q2
Q1*Z0		GOSBVL	=GETPTR
		A=DAT1	A
		D1=D1+	5
		D=D+1	A
		DAT1=A	A
		GOVLNG	=Loop
* Return negated Q2 as result
Z-1*Q2		GOSBVL	=GETPTR
		A=DAT1	A
		D1=D1+	5
		D=D+1	A
		DAT1=A	A
		GONC	DoQNeg

* Return negated Q1 as result
Q1*Z-1		GOSBVL	=GETPTR
		D1=D1+	5
		D=D+1	A

* Special entry for QNeg

DoQNeg		GOSBVL	=SAVPTR
		A=DAT1	A
		D0=A			->Q
		GOSUBL	QSIZE		C[A] = Qsize
		GOSBVL	=GETTEMP	Allocate new copy
		RSTK=C			Qsize
		GOSBVL	=D1=DSKTOP
		A=DAT1	A		->Q
		AD0EX
		DAT1=A	A		OverWrite Q with Q' address
		D1=A			->Q'
		R0=A			->Q'
		C=RSTK			Qsize
		GOSBVL	=MOVEDOWN
		C=R0
		D0=C
		GOSUBL	QNEG
		GOVLNG	=GETPTRLOOP

* Z0 and Z1 tests failed. Now see if short multiplication is possible

tstshortmul	?A#C	A
		GOYES	doslow*		?*Q1 - use slow method
		A=DAT0	A
		?A#C	A
		GOYES	doslow*		Q1*Z2 - use slow method

		D0=D0+	5		->Z1len
		D1=D1+	5		->Z2len
		A=DAT0	A		Z1len
		C=DAT1	A		Z2len
		A=A+C	A		Z1len+Z2len
		LC(5)	10+16
		?A>C	A
		GOYES	doslow*		Too long result - use slow method

* (short) Z1*Z2 - use fast method

dofast*		GOSUBL	shortpopd1	A[W] = mantissa2	C[S] = sign2
		B=A	W		B[W] = mantissa2
		D=C	S		D[S] = sign2
		GOSUBL	shortpop	A[W] = mantissa1	C[S] = sign1
		D=D+C	S		Calculate result sign
		GONC	+
		D=0	S
+		GOSUBL	BMULB		A[W]=C[W] = A[W]*B[W]
		C=D	S		C[S] = sign
		GOLONG	PushZ2Loop	Push the result and exit

* Slow multiplication needed. Initialize memory, pointers & start

doslow*		GOSBVL	=ROOM		C[A] = free mem
		LA(5)	400		Minimum total required
		?C>A	A
		GOYES	doslowmem

slowmulretry	GOSBVL	=DOGARBAGE
*		ST=1	sGARB		DOGARBAGE sets this
		GOTO	doslow*

doslowmem	ASRB.F	A		200
		C=C-A	A		room-200
		GOSBVL	=GETTEMP	D0=->slot	D1=->slotend
		AD1EX
		R4=A.F	A		R4[A]=->stack
		AD0EX
		R3=A.F	A		->output

		GOSBVL	=D1=DSKTOP
		C=DAT1	A
		D1=D1+	5
		A=DAT1	A
		D0=A			->Q1
		D1=C			->Q2

		GOSUB	QMUL		A[A] = ->result	R3[A]=->result end

QMULexit	R0=A	A		R0[A] = ->Q
		C=R3	A		C[A] = ->Qtail
		GOSUBL	ShrinkQ
		D1=D1+	5		Drop Q2
		D=D+1	A
		GOVLNG	=OverWrALoop	Overwrite Q1 & exit

* Trash allocated memory first so it won't corrupt memory

QMULretry	GOSUBL	TrashTop
		GOTO	slowmulretry

**********************************************************************
* Desc:		Multiply Q1 and Q2 to output
* Entry:	D0  = ->Q1
*		D1  = ->Q2
*		R3[A] = ->output
*		R4[A] = ->stack
* Exit:		D0    = ->Q1tail
*		D1    = ->Q2tail
*		A[A]  = ->output	(start of result)
*		R3[A] = ->output end	(end of result)
*		R4[A] = ->stack		(same as input)
**********************************************************************

qmulerr		GOTO	QMULretry

* General multiplication entry, saves pointers & dispatches to subtypes
* Push[A] and Pop[A] are not used for speed.

QMUL		A=R4.F	A		->stack
		C=R3.F	A		->output
		A=A-C	A		free mem
		GOC	qmulerr		Shouldn't happen - but just in case
		LC(5)	3*5		Need 3 levels
		?A<C	A
		GOYES	qmulerr

		A=R4.F	A		->stack
		AD0EX
		D0=D0-	5		Save RSTK1
		C=RSTK
		DAT0=C	A
		D0=D0-	5		Save RSTK2
		C=RSTK
		DAT0=C	A
		D0=D0-	5		Save ->output (result address)
		C=R3.F	A
		DAT0=C	A
		AD0EX
		R4=A.F	A		->stack'
		GOSUB	QMULdispatch	Dispatch
		A=R4.F	A		->stack'
		AD0EX
		C=DAT0	A		Restore ->output (result address)
		B=C	A
		D0=D0+	5
		C=DAT0	A		Restore RSTK2
		RSTK=C
		D0=D0+	5
		C=DAT0	A		Restore RSTK1
		D0=D0+	5
		AD0EX
		R4=A.F	A		->stack'
		A=B	A		->output (result address)
		PC=C			*RETURN*

* Note that only 2 RSTK levels are saved, and one is replaced by
* the QMULdispatch return address. This implies that QMUL can be called
* recursively from one of the dispatchees or one of their subroutines,
* but not subroutines subroutine or deeper.

* Dispatch based on types

QMULdispatch	A=DAT1	A		prolog2
		B=A	A
		A=DAT0	A		prolog1
		LC(5)	=DOHSTR
		?A=C	A
		GOYES	qmulz?
		?B#C	A
		GOYES	MULQQ
		GOTO	MULQZ
qmulz?		?B=C	A
		GOYES	qmulzz
		GOTO	MULZQ
qmulzz		GOTO	MULZZ

**********************************************************************
* Desc:		Multiply Q1*Q2
* Algorithm:
*	1) Traverse Q1, saving elements and length
*	2) Traverse Q2, saving elements and length
*	3) Init double loop
*	4) Do the common double loop
*
*	FOR J = 0 TO (LEN1+LEN2-2)
*		Element[J] = 0
*		FOR I = MAX[0,J-LEN2+1] TO MIN[J,LEN1-1]
*			Element[J] += Q1[I] * Q2[J-I]
*		NEXT I
*	NEXT J
*
*	Saved to the stack:
*
*	X[1] X[2] ... X[LEN1]
*	Y[1] Y[2] ... Y[LEN2]
*	END1 END2
*	LEN1 LEN2
*	J JSTOP
*	ELEMENT
*	I ISTOP
*
*	Future speed improvement:
*		Z1+Z2 backaddition can be done in place.
*		==> Q1+Z2 backaddition can be done in place.
*
**********************************************************************
MULQQ		GOSUB	TraverseQ
		R0=A.F	A		LEN1
		R2=C.F	A		END1
		AD1EX
		D0=A
		GOSUB	TraverseQ
		R1=A.F	A		LEN2
		CR2EX.F	A		R2[A] = END2
		A=C	A
		GOSUB	Push[A]		END1
		A=R2.F	A
		GOSUB	Push[A]		END2
		A=R0.F	A
		GOSUB	Push[A]		LEN1
		A=R1.F	A
		GOSUB	Push[A]		LEN2
		A=0	A
		GOSUB	Push[A]		J
		A=R0.F	A		LEN1
		C=R1.F	A		LEN2
		A=A+C	A		LEN1 + LEN2
		A=A-1	A		LEN1 + LEN2 - 1
		A=A-1	A		LEN1 + LEN2 - 2
		GOSUB	Push[A]		JSTOP

		P=	5
		GOSUB	Ensure[P]
		A=R3.F	A		->output
		AD1EX
		LC(5)	=DOLIST
		DAT1=C	A
		D1=D1+	5
		AD1EX
		R3=A.F	A		->output'

* Do things in register for as long as possible, then step into
* the slower second loop which does the multiplications
* and back additions in memory. This often provides a drastic
* speedup since the overhead in stack saves is quite significant.

mulqqloop	A=R3.F	A		->element
		GOSUB	Push[A]
		LC(5)	5+5+16+1+10	Make sure there's room for
		GOSUB	Ensure[C]	result if it fits into register
		A=0	W		(+10 for I and ISTOP)
		R0=A	W		R0[W]R1[S] = signed result = 0 now
		R1=A.F	S
		ST=1	sQMULREG	Flag result in registers

		A=R4.F	A		->stack
		D1=A
		D1=D1+	10		->J (stack)
		C=DAT1	A
		D=C	A		J
		D1=D1+	10		->LEN1
		C=DAT1	A
		B=C	A		LEN1
		D1=D1-	5		->LEN2
		A=DAT1	A		LEN2

* I = MAX[0,J-LEN2+1]

		C=D	A		J
		C=C+1	A		J+1
		C=C-A	A		J-LEN2+1
		GONC	indexok
		C=0	A
indexok		A=C	A
		GOSUB	Push[A]

* ISTOP = MIN[J,LEN1-1]

		C=B	A		LEN1
		C=C-1	A		LEN1-1
		?C<=D	A
		GOYES	istopok
		C=D	A		J
istopok		A=C	A
		GOSUB	Push[A]

*	X[1] X[2] ... X[LEN1]
*	Y[1] Y[2] ... Y[LEN2]
*	END1 END2
*	LEN1 LEN2
*	J JSTOP
*	ELEMENT
*	I ISTOP
*
*	X[I]   = STACK[8+LEN2+LEN1-I]
*	Y[J-I] = STACK[8+LEN2-J+I)]

mulqqresloop	A=R4.F	A
		D0=A			->stack
		D0=D0+	5
		A=DAT0	A		I
		D0=D0+	15
		D0=D0+	5
		C=DAT0	A		LEN2
		C=C-A	A		LEN2-I
		D0=D0+	5
		A=DAT0	A		LEN1
		A=A+C	A		LEN1+LEN2-I
		A=A+CON	A,8		8+LEN1+LEN2I
		C=R4.F	A
		C=C+A	A
		A=A+A	A
		A=A+A	A
		C=C+A	A
		D0=C
		C=DAT0	A
		D0=C			->X[I]

		A=R4.F	A
		D1=A
		D1=D1+	5
		A=DAT1	A		I
		D1=D1+	15
		C=DAT1	A		J
		A=A-C	A		I-J
		D1=D1+	5
		C=DAT1	A		LEN2
		A=A+C	A		LEN2-J+I
		A=A+CON	A,8		8+LEN2-J+I
		C=R4.F	A
		C=C+A	A
		A=A+A	A
		A=A+A	A
		C=C+A	A
		D1=C
		C=DAT1	A
		D1=C			->Y[J-I]

* If not in register multiplication then use old memory code
* Else try to get new temporary result in registers.
* If fails then setup new mulqqresloop pass with old memory code.

		?ST=1	sQMULREG
		GOYES	mulqqresreg
		GOTO	mulqqresmem

* First save D0 & D1 so we don't have to jump all the way back to
* mulqqresloop if overflow occurs. This benefits especially
* multivariate multiplications.

mulqqresreg	AD0EX
		R1=A.F	A
		D0=A
		CD1EX
		R2=C.F	A
		D1=C

* Faster code for doing the math in registers

		GOSUBL	SHORTPOPCK?
		GOC	mulqqregfail	* Change to memory code *
		B=A	W		B[W]=mantissa1
		D=C	S		D[S]=sign1
		CD1EX
		D0=C
		GOSUBL	SHORTPOPCK?
		GOC	mulqqregfail	* Change to memory code *
		CBEX	W		AB=int1	CD=int2
		GOSUBL	SHORT*		AC=int1*int2
		GOC	mulqqregfail	* Change to memory code *
		B=C	S		AB=int1*int2
		C=R1.F	S
		D=C	S
		C=R0	W		CD=result
		GOSUBL	SHORT+
		GOC	mulqqregfail	* Change to memory code *
		R0=A	W
		R1=C.F	S		R0R1=result'
		GOTO	mulqqresnext

* Overflow! Output current result and try again

mulqqregfail	ST=0	sQMULREG	No more in registers
		C=R3.F	A
		D1=C			->output
		A=R0	W
		C=R1.F	S		AC = current result
		GOSUBL	SHORTOUT	Free memory was verified earlier
		CD1EX
		R3=C.F	A		->output'
		A=R1.F	A
		D0=A
		C=R2.F	A
		D1=C			Now we can just fall through
		
* Slower code for doing the math in memory

mulqqresmem	GOSUB	QMUL		A[A]=->result
		ST=0	sQMULREG	Must have this one due to recursion!
		D1=A			->result
		A=R4.F	A		->stack
		D0=A
		D0=D0+	10
		A=DAT0	A
		D0=A			->element

		C=R3.F	A		->output
		GOSBVL	=CSLW5
		C=R3.F	A
		R3=C			->output ->output
		C=C-A	A		size of Q1 and Q2
		GOSUB	Ensure[C]
		GOSUB	QADD		R3[A]=->output tail

		C=R4.F	A		->stack
		D1=C
		D1=D1+	10
		C=DAT1	A
		D1=C			->element

		A=R3
		C=A	A		->output tail
		GOSBVL	=ASRW5
		D0=A			->output
		C=C-A	A		output size
		GOSBVL	=MOVEDOWN
		CD1EX
		R3=C.F	A		->output'

mulqqresnext	A=R4.F	A
		D0=A			->stack
		C=DAT0	A		istop
		D0=D0+	5
		A=DAT0	A		I
		A=A+1	A		I++
		DAT0=A	A
		?A>C	A
		GOYES	mulqqresok
		GOTO	mulqqresloop	Loop until ISTOP

* Now if we managed to do it all in registers we have to output
* the result - otherwise the result is already in place

mulqqresok	?ST=0	sQMULREG
		GOYES	mulqqresokmem
		C=R3.F	A
		D1=C			->output
		A=R0	W
		C=R1.F	S		AC = result
		GOSUBL	SHORTOUT
		CD1EX
		R3=C.F	A		->output'

mulqqresokmem	D0=D0+	10
		AD0EX
		R4=A.F	A		->stack' (popped ELEMENT I ISTOP)
		D0=A
		C=DAT0	A		JSTOP
		D0=D0+	5
		A=DAT0	A		J
		A=A+1	A		J++
		DAT0=A	A
		?A>C	A
		GOYES	mulqqloopok
		GOTO	mulqqloop

* Finish Q*Q loop ; drop stack & output SEMI
		
mulqqloopok	D0=D0+	5
		C=DAT0	A		LEN2
		D0=D0+	5
		A=DAT0	A		LEN1
		C=C+A	A		LEN1+LEN2
		D0=D0+	5
		A=DAT0	A		END2
		D1=A
		D0=D0+	5
		A=DAT0	A		END1
		AD0EX
		A=A+C	A
		C=C+C	A
		C=C+C	A
		A=A+C	A
		A=A+CON	A,5
		R4=A.F	A		->stack'

		A=R3.F	A		->output
		AD1EX
		LC(5)	=SEMI		Memory exists due to stack pops
		DAT1=C	A
		D1=D1+	5
		AD1EX
		R3=A.F	A		->output'
		RTN

**********************************************************************
* Desc:		Traverse polynomial in D0, save element addresses
*		and return polynomial degree
* Entry:	D0    = ->P
*		R3[A] = ->output
*		R4[A] = ->stack
* Output:	A[A] = degree
*		C[A] = ->Ptail
**********************************************************************
TraverseQ	D0=D0+	5		Skip DOLIST
		D=0	A		complevel = 0
		B=0	A		degree = 0
travelQloop	A=DAT0	A
		LC(5)	=SEMI
		?A=C	A
		GOYES	travelS
		?D#0	A
		GOYES	travelQ10
		B=B+1	A		degree++
		AD0EX
		D0=A
		GOSUB	Push[A]
travelQ10	A=DAT0	A
		D0=D0+	5
		LC(5)	=DOLIST
		?A=C	A
		GOYES	travelL
		C=DAT0	A		Zlen
		AD0EX
		A=A+C	A
		AD0EX
		GONC	travelQloop
travelL		D=D+1	A		complevel++
		GONC	travelQloop
travelS		D0=D0+	5
		D=D-1	A
		GONC	travelQloop
		CD0EX			->Ptail
		A=B	A		degree
		RTNCC
**********************************************************************
* Desc:		Ensure memory exists between ->output and ->stack
* Input:	C[A] = need
* Alters:	C[A]
**********************************************************************

Ensure[P]	C=0	A
		CPEX	0
Ensure[C]	AR3EX.F	A
		C=A+C	A		->output end
		AR3EX.F	A
		AR4EX.F	A
		?C>A	A
		GOYES	ensure00	CS if ->output end > ->stack
ensure00	AR4EX.F	A
		RTNNC
		GOTO	QMULretry	Try again after GC

**********************************************************************
* Desc:		Multiplication stack push
* Entry:	A[A] = push address
*		R3[A] = ->output	(current output pointer)
*		R4[A] = ->stack		(address of topmost stack level)
* Modifies:	C[A] R4[A]
* Errors:	If new level would overwrite something below ->output
*		QMUL is restarted after forcing a GC
**********************************************************************

Push[A]		AR4EX.F	A		A[A]=->stack	R4[A]=input
		A=A-CON	A,5		->stack'
		C=R3.F	A		->output
		?A<C	A
		GOYES	StackErr	Not enough memory
		AD0EX
		C=R4.F	A		input
		DAT0=C	A
		AD0EX
		AR4EX.F	A		A[A]=input	R4[A]=->stack'
		RTNCC
StackErr	GOTO	QMULretry	Try again after GC

**********************************************************************
* Desc:		Multiplication stack pop
* Entry:	R3[A] = ->output	(current output pointer)
*		R4[A] = ->stack		(address of topmost stack level)
* Exit:		A[A] = addr
* Modifies:	R4[A]
**********************************************************************

Pop[A]		CR4EX.F	A		C[A]=->stack	R4[A]=C[A]
		CD0EX
		A=DAT0	A		A[A]=pop
		D0=D0+	5		->stack'
		CD0EX
		CR4EX.F	A		C[A]=C[A]	R4[A]=->stack'
		RTNCC

**********************************************************************
* Desc:		Multiply Z1*Q2
* Algorithm:
*	1) Check special cases for fast code
*	2) Multiply Q2 element by element to output
* Notes:
*	One could easily use the stack approach but it would just create
*	unnecessary overhead by requiring the stack saves.
**********************************************************************

gomulz0q	GOTO	mulz0q
gomulz1q	GOTO	mulz1q
gomulz-1q	GOTO	mulz-1q

MULZQ		ST=0	sQSWAP*		No swap done
mulzqentry	D0=D0+	5
		A=DAT0	7		Z1 length & first 2 nibbles
		D0=D0-	5
		C=A	W
		LCHEX	000006		Z0 body
		?A=C	W
		GOYES	gomulz0q	Z0 * Q ==> Z0
		LCHEX	0100007
		?A=C	W
		GOYES	gomulz1q	Z1 * Q ==> Q
		LCHEX	F100007
		?A=C	W
		GOYES	gomulz-1q	Z-1 * Q ==> -Q
		B=0	A		complevel
mulzqloop	A=DAT1	A
		LC(5)	=DOHSTR
		?A=C	A
		GOYES	qmulzq.zz
		P=	5
		GOSUB	Ensure[P]
		C=R3.F	A		->output
		CD1EX
		DAT1=A	A		Write DOLIST/SEMI
		D1=D1+	5
		CD1EX
		R3=C.F	A		->output'
		B=B+1	A		complevel++
		D1=D1+	5		Skip DOLIST/SEMI
		LC(5)	=DOLIST
		?A=C	A
		GOYES	mulzqloop
		B=B-1	A
		B=B-1	A		complevel--
		?B#0	A
		GOYES	mulzqloop
		D0=D0+	5
		C=DAT0	A		Z1len
		AD0EX
		A=A+C	A		->Z1tail
		AD0EX
mulzqexit	?ST=0	sQSWAP*		No swap done?
		GOYES	+		Yep - done
		AD0EX			No - restore order
		AD1EX
		AD0EX
+		RTNCC
qmulzq.zz	C=B	A		Save complevel
		RSTK=C
		CD0EX			Save ->Z1
		RSTK=C
		CD0EX
		GOSUB	MULZZ
		C=RSTK			Restore ->Z1
		D0=C
		C=RSTK			Restore complevel
		B=C	A
		GOTO	mulzqloop

*  Z0*Q2 ; Skip both and output Z0
mulz0q		GOSUBL	QSKIPD1
		P=	11
		GOSUB	Ensure[P]
		A=DAT0	11		Z0
		D0=D0+	11
		C=R3.F	A
		CD1EX
		DAT1=A	11
		D1=D1+	11
		CD1EX
		R3=C.F	A
		GOTO	mulzqexit

* Z1*Q2 ; Skip both and output Q2
mulz1q		D0=D0+	12		Skip Z1
		CD1EX
		B=C	A		->Q2
		CD1EX
		GOSUBL	QSKIPD1		->Q2tail
		CD1EX
		C=C-B	A		Q2size
		CBEX	A		B[A]=Q2size
		D1=C
		C=B	A
		GOSUB	Ensure[C]
		C=R3.F	A
		CD1EX
		CD0EX
		CBEX	A
		GOSBVL	=MOVEDOWN
		C=B	A
		CD0EX
		CD1EX
		R3=C.F	A
		GOTO	mulzqexit

* -Z1*Q2 ; Skip both and output -Q2
mulz-1q		C=R3.F	A		->output
		RSTK=C
		GOSUB	mulz1q		Output Q2
		C=RSTK
		CD0EX
		B=C	A		->Z1tail
		GOSUBL	QNEG
		C=B	A
		D0=C			->Z1tail
		GOTO	mulzqexit

**********************************************************************
* Desc:		Multiply Q1*Z2
*	1) Multiply Q1 element by element to output
**********************************************************************

MULQZ		ST=1	sQSWAP*		Flag swapped order
		AD0EX
		AD1EX
		AD0EX
		GOTO	mulzqentry	And do the reverse multiplication

**********************************************************************
* Desc:		Multiply Z1*Z2
* Algorithm:
*	1) Check sizes, ensure sufficient memory
*	2) If either Z is zero then output Z0 and return
*	3) If multiplication result fits into register then
*	   so do the integers - multiply in registers & return
*	4) Do slow multiplication in segments
*
* In:	D0	->Z1		Out:	D0	->Z1tail
*	D1	->Z2			D1	->Z2tail
*	R3[A]	->output		R3[A]	->output' (tail)
*	R4[A]	->stack			R4[A]	->stack
*
* Alters:	D0 D1 A[W] B[W] C[W] D[S,A] R0[A] R1[A] R2[A] R3[A]
*
* Stack levels:	2 (Data address storage in slow * loop)
*
* Warning:	Due to a bug in the HP interrupt system the interrupt
*		system will use *3* RSTK levels if the keyboard gets
*		filled and beep is enabled. The original MULZZ used
*		4 RSTK levels and was thus very prone to RSTK
*		overflows.
*		==> Made BMUL inline code
*		==> Made R2[A] to serve as an extra pointer
*		==> Only 2 RSTK levels used
**********************************************************************

mulzzerr	GOTO	QMULretry	Insufficient memory - retry

MULZZ		D0=D0+	5		->Z1len
		D1=D1+	5		->Z2len
		A=DAT0	A		Z1len
		C=DAT1	A
		D0=D0-	5
		D1=D1-	5
		A=A+C	A		Z1len+Z2len
		LC(5)	10+50		+ 2 prologs & safety
		C=C+A	A		needed memory (max)

		A=R3.F	A
		B=A	A		->output
		A=R4.F	A		->stack
		A=A-B	A		free memory
		?A<C	A
		GOYES	mulzzerr

* Special entry to avoid memory checks (R4 becomes free)

MULZZentry	D0=D0+	5
		D1=D1+	5
		A=DAT0	A		Z1len
		C=DAT1	A		Z2len
		A=A-CON	A,6		nibs1 (sign not included)
		C=C-CON	A,6		nibs2 (sign not included)

* Now check if either Z is zero

		?A#0	A
		GOYES	mulz1#0

mulzz0skp	C=DAT0	A		Skip Z1
		AD0EX
		A=A+C	A
		AD0EX
		C=DAT1	A		Skip Z2
		AD1EX
		A=A+C	A
		AD1EX
mulzz0		A=R3.F	A		->output
		AD1EX
		LCZ0
		DAT1=C	11
		D1=D1+	11
		AD1EX
		R3=A.F	A		->output'
		RTNCC

mulz1#0		?C=0	A
		GOYES	mulzz0skp

* Neither Z was zero - check for short multiplication

		R0=A.F	A		nibs1
		R1=C.F	A		nibs2
		A=A+C	A		nibs1 + nibs2
		A=0	P
		?A#0	A
		GOYES	MULZZslow	14+1 is allowed, 15+1 is not

** Fast register multiplication

MULZZfast	GOSUBL	shortpopd1	A[W]=mantissa2	C[S]=sign2
		B=A	W		B[W]=mantissa2
		D=C	S				D[S]=sign2
		GOSUBL	shortpop	A[W]=mantissa1	C[S]=sign1
		D=D+C	S		calculate result sign
		GONC	+
		D=0	S
+		GOSUBL	BMULB		A[W]=C[W] = A*B
		C=D	S		C[S] = sign
		CD1EX
		CR3EX.F	A		->output
		D1=C
		GOSUBL	SHORTOUT
		CD1EX
		CR3EX.F	A		->output'
		CD1EX
		RTN

** Slow Z multiplication for long integers in 8-nibble chunks

MULZZslow

* Determine sign of result

		A=DAT0	A		Z1len
		CD0EX
		A=A+C	A
		A=A-1	A
		D0=A
		A=DAT0	S		Z1sign
		D0=C			->Z1len

		A=DAT1	A		Z2len
		CD1EX
		A=A+C	A
		A=A-1	A
		D1=A
		C=DAT1	S		Z2sign
		D1=C			->Z1len

		C=C+A	S		[00 0F F0 FF] --> [0 F F E]
		GONC	mulzzsgnok
		C=0	S		[00 0F F0 FF] --> [0 F F 0]
mulzzsgnok	D=C	S		D[S] = Zsign

* Initialize output

		A=R0			nibs1
		C=R1			nibs2
		A=A+C	A
		LC(5)	10+2+1+20	prolog+len+extra2+sign +safety
		C=C+A	A		need
		A=R3.F	A		->output
		AD1EX
		B=A	A		->Z2len
		C=0	S		Output a non-zero stop mark just
		C=C+1	S		in case we have an abnormal zero
		DAT1=C	S		in multiplication
		D1=D1+	1
		GOSBVL	=WIPEOUT
		A=B	A
		D1=A			->Z2len

* Now determine loop counters for doing recursive multiplication
* with max 8-nibble units.
*
* For each 8 nibbles in the first integer
*	Multiply each 8 nibbles in the second one and
*	add the result to the proper place in output
*	Uses an 8 nibble carry

		D0=D0+	5		->Z1mantissa
		D1=D1+	5		->Z2mantissa
		CD1EX
		R2=C.F	A		->Z2mantissa
		CD1EX
		C=R3.F	A		->output
		C=C+CON	A,10		->Zbody
		RSTK=C			->output start

* Multiply 8 nibbles from D0 with remainder of D1 to output

ZMULloop1	C=RSTK
		RSTK=C			->output start
		RSTK=C			->output slider
		C=R1.F	A
		D=C	A		nibs2
		B=0	W		No carry

ZMULloop2	P=	8-1
		D=D-CON	A,8		Try using 8 nibs of Z2
		GONC	zmul10
		C=D	B		Not that much left, use what's left
		C=C+CON	B,7
		P=C	0
		D=0	A		No more nibbles left
zmul10		A=0	W
		A=DAT1	WP		A=Vj (block j of Z2)
		CD1EX
		C+P+1			->next block of Z2
		CD1EX

		C=R0.F	A		nibs1
		P=	8-1
		C=C-CON	A,8		Try using 8 nibs of Z1
		GONC	zmul20
		C=C+CON	B,7		Not that much left, use what's left
		P=C	0
zmul20		ABEX	W		A[8-15]=carry	B[W]=Vj
		C=0	W
		C=DAT0	WP		C=Ui (block i of Z1)
		P=	0

* Compute A=Ui*Vj+carry.

		ASR	W
		ASR	W
		ASR	W
		ASR	W
		ASR	W
		ASR	W
		ASR	W
		ASR	W		A[0-7]=carry
zmul30		SB=0
		CSRB
		?SB=0
		GOYES	zmul40
		A=A+B	W
zmul40		B=B+B	W
		?C#0	W
		GOYES	zmul30

* Add the multiplication result to target

		C=RSTK
		CD1EX
		RSTK=C			->next block of Z2
		C=0	W
		C=DAT1	8		C=Wi+j  (block i+j of result)
		C=C+A	W		C=Ui+Vj+carry+Wi+j  (<2^16)
		?D=0	A
		GOYES	ZMULlpnext	Done with inner loop
* write block result and repeat inner loop
		DAT1=C	8		write new Wi+j
		B=C	W		B[8-15]=carry
		D1=D1+	8		Next target block
		C=RSTK
		CD1EX
		RSTK=C			->output slider
		GOTO	ZMULloop2
* end inner loop and update D0 pointer & output start location
ZMULlpnext	DAT1=C	W		write new Wi+j and final carry
		D1=D1+	16		->output tail
		C=RSTK			->drop ->Z2 tail
		C=RSTK			->old output start
		A=C	A
		A=A+CON	A,8		->new output start
		C=R2.F	A		->mantissa2
		CD1EX			D1 = ->mantissa2
		ACEX	A		A  = ->output tail
		RSTK=C			->new output tail
		C=R0.F	A
		C=C-CON	A,8		nibs1--
		GOC	zmuldone
		?C=0	A
		GOYES	zmuldone	done if there were <= 8 nibs
		R0=C.F	A
		D0=D0+	8
		GOTO	ZMULloop1

zmuldone	D1=A			->output tail
		C=R0.F	A		nibs1
		P=C	0
		CD0EX
		C+P+1			Skips sign too
		CD0EX
		P=	0
		C=RSTK			drop ->output start
		C=R2.F	A		->mantissa2
		CD1EX
		D1=D1-	5
		A=DAT1	A		Z2len
		CD1EX
		C=C+A	A		->Z2tail
		D=C	A		->Z2tail

* Now scan backwards until non-zero nibble

zmulstriplp	D1=D1-	1
		A=DAT1	S
		?A=0	S
		GOYES	zmulstriplp
		D1=D1+	1
		C=D	S		Output sign
		DAT1=C	S
		D1=D1+	1
* Now check what we got
		AD1EX
		D1=A			->output'
		C=R3.F	A		->output
		A=A-C	A		size
		LC(5)	10
		?A<=C	A
		GOYES	zmuliszero	Output zero for too small results
		C=R3.F	A		(Meaning we had an unnormalized zero)
		CD1EX
		R3=C.F	A		->output'
		LC(5)	=DOHSTR
		DAT1=C	A
		D1=D1+	5
		A=A-CON	A,5
		DAT1=A	A
		C=D	A
		D1=C			->Z2tail
		RTNCC
zmuliszero	C=D	A
		D1=C			->Z2tail
		GOTO	mulzz0		Output Z0



**********************************************************************
* Desc:		Add Q1+Q2 to output
* Entry:	D0    = ->Q1
*		D1    = ->Q2
*		R3[A] = ->output (with sufficient free memory)
* Exit:		R3[A] = ->output'
* Algorithm:	For each Z1,Z2 pair
*			output Z1+Z2
*		For each Z1,P2 pair
*			complevel2++
*			copy DOLIST & all but last element of P2 to output
*		For each P1,Z2 pair
*			complevel1++
*			copy DOLIST & all but last element of P1 to output
*		For each P1,P2 pair
*			complevel1++, complevel2++
*			copy DOLIST & leading elements from the longer
*			polynomial to output.
*		For each SEMI SEMI pair
*			complevel1--m complevel2--, output SEMI
*		For each ob SEMI pair
*			complevel2--, output SEMI
*		For each SEMI ob pair
*			complevel1--, output SEMI
**********************************************************************
QSUB		ST=1	sQSUB
		GOTO	QADD/SUB

QADD		ST=0	sQSUB

QADD/SUB	C=0	A
		R0=C.F	A		comp1 = 0
		R1=C.F	A		comp2 = 0
		ST=0	sQSWAP		Start in normal order

* Main Q1+Q2 addition loop

Qaddloop	A=DAT1	A
		B=A	A		Q2type
		A=DAT0	A		Q1type
		LC(5)	=SEMI
		?A=C	A
		GOYES	Qaddsemi?
		?B=C	A
		GOYES	Qadd?semi
		LC(4)	=DOHSTR
		?A=C	A
		GOYES	Qaddz?
		?B=C	A
		GOYES	Qaddpz
		GOTO	Qaddpp

Qaddsemi?	D0=D0+	5		Skip SEMI
		A=R0.F	A
		A=A-1	A		comp1--
		R0=A.F	A
		?B#C	A		SEMIs in both polys?
		GOYES	Qaddsemi	No - just output SEMI
Qadd?semi	D1=D1+	5		Skip SEMI
		A=R1.F	A
		A=A-1	A		comp2--
		R1=A.F	A
Qaddsemi	A=R3.F	A		->output
		AD1EX
		DAT1=C	A		Write SEMI
		D1=D1+	5
		AD1EX
		R3=A.F	A
		GONC	Qaddnext	Continue loop

Qaddzz		GOSUB	ZADD		Add Z1 & Z2 to output

Qaddnext	A=R0.F	A		Done if either complevel = 0
		C=R1.F	A
		?A=0	A
		GOYES	Qaddok
		?C=0	A
		GOYES	Qaddok
		GOTO	Qaddloop
Qaddok		GOTO	Qaddfinish	Finish up result

Qaddz?		?B=C	A
		GOYES	Qaddzz

* Add P+Z by copying all but last element of P to output, then
* reloop immediately with no complevel checks!

Qaddzp		GOSUB	QaddSwap
Qaddpz		C=R0.F	A
		C=C+1	A		comp1++
		R0=C.F	A
		GOSUBL	QLEN		Count elements in the list in D0
		B=B-1	A
		GOSUB	QaddCopy	Copy all but last to output
		GOTO	Qaddloop	Loop immediately!

* Add P+P by copying the lading elements of the longer P to output,
* then reloop to add the equal number of elements (whatever they are)

Qaddpp		C=R0.F	A
		C=C+1	A		comp1++
		R0=C.F	A
		C=R1.F	A
		C=C+1	A		comp2++
		R1=C.F	A
		GOSUBL	QLEN		Count P1
		R2=A.F	A		N1
		GOSUB	QaddSwap
		GOSUBL	QLEN		A = N2
		C=R2.F	A		N1
		?A>=C	A		N2 >= N1 ?
		GOYES	Qaddpp10
		GOSUB	QaddSwap
		ACEX	A
Qaddpp10	A=A-C	A		Difference in lenght
		B=A	A
		GOSUB	QaddCopy	Copy the lead
		D1=D1+	5		Skip DOLIST in the shorter P too
		GOTO	Qaddloop

* Here atleast one of the complevels is zero, but we must output the
* trailing SEMIs from the other P too

Qaddfinish	A=A+C	A		comp1 + comp2 = needed SEMIs
		CD0EX
		CR3EX.F	A
		CD0EX
		LC(5)	=SEMI
Qaddsemilp	A=A-1	A
		GOC	Qadddone
		DAT0=C	A
		D0=D0+	5
		GONC	Qaddsemilp
Qadddone	CD0EX
		CR3EX.F	A
		CD0EX
		RTNCC

****************************************
* Swap addition polynomial order
****************************************
QaddSwap	CD0EX
		CD1EX			Swap P1&P2
		CD0EX
		CR0EX.F	A
		CR1EX.F	A		Swap comp1&comp2
		CR0EX.F	A
		?ST=1	sQSWAP		Flag new order
		GOYES	Qaddswap10
		ST=1	sQSWAP
		RTN
Qaddswap10	ST=0	sQSWAP
		RTN

****************************************
* Copy first N elements of a polynomial
* to output (including DOLIST).
* If substract is flagged and P is 
* the substracted P then also
* negate all the copied integers.
* Entry:	D0 = ->P
*		B[A] = N
*		R3[A] = ->output
* Exit:		D0 = ->P'
*		R3[A] = ->output'
* Alters:	A[W] C[W] B[A] D[W] R3[A]
* Stack:	2
****************************************
QaddCopy	?ST=0	sQSUB
		GOYES	QaddCopy+
		?ST=0	sQSWAP
		GOYES	QaddCopy+

* P is being substracted, copy and negate. First copy all elements

		C=R3.F	A
		GOSBVL	=CSLW5
		D=C	W			D[A1] = ->output save
		GOSUB	QaddCopy+
* Then traverse and negate all copied elements
		C=D	W
		GOSBVL	=CSRW5			->output
		CD1EX
		D=C	A			->P2
Qaddcpneglp	A=DAT1	A
		D1=D1+	5
		LC(5)	=DOHSTR
		?A=C	A
		GOYES	Qaddcpnegz
Qaddcpnegnext	AD1EX
		D1=A
		C=R3.F	A			->output'
		?A<C	A
		GOYES	Qaddcpneglp
		C=D	A
		D1=C
		RTNCC
Qaddcpnegz	A=DAT1	A
		CD1EX
		C=C+A	A
		CD1EX
		A=A-CON	A,7			CS if Z0
		GOC	Qaddcpnegnext
		D1=D1-	1
		A=DAT1	S
		A=-A-1	S
		DAT1=A	S
		D1=D1+	1
		GONC	Qaddcpnegnext

* Straight copy of the elements

QaddCopy+	GOSUBL	QSKIPN			D0=->Q[N]	C[A]=->Q
		AD0EX
		D0=C				->P
		C=A-C	A			size to copy
		CD1EX
		CR3EX.F	A
		CD1EX
		GOSBVL	=MOVEDOWN
		CD1EX
		CR3EX.F	A
		CD1EX
		RTNCC

**********************************************************************
* Add Z1+Z2 to output
* In:		D0  ->Z1
*		D1  ->Z2
*		R3  ->output
*		sQSUB		Indicates substraction
*		sQSWAP		Indicates arguments ar reversed
* Out:		D0  ->Z1tail	\ possibly
*		D1  ->Z2tail	/ swapped
*		R3  ->output'
* Uses:		A[W] C[W] B[S] D[S]
* Stack:	2 (WIPEOUT + RSTK save)
**********************************************************************
ZADD

* See if addition can be done in registers

		D0=D0+	5		->Z1len
		D1=D1+	5		->Z2len
		A=DAT0	A		Z1len
		C=DAT1	A		Z2len
		A=A-CON	A,5
		C=C-CON	A,5
		A=0	P
		C=0	P
		?A#0	A
		GOYES	goZaddSlow
		?C=0	A
		GOYES	ZaddFast
goZaddSlow	GOTO	ZaddSlow

* Fast addition

ZaddFast	GOSUB	shortpopd1	A[W] = mantissa2	C[S]=sign2
		B=A	W		B[W] = mantissa2
		D=C	S					D[S]=sign2
		GOSUB	shortpop	A[W] = mantissa1	C[S]=sign1
		CBEX	W		C[W] = mantissa2	B[S]=sign1

		?ST=0	sQSUB		No need to change signs if addition
		GOYES	+
		D=-D-1	S		Negate Z2 for addition call
		?ST=0	sQSWAP		Signs ok if not reversed args
		GOYES	+
		B=-B-1	S		Else negate both
		D=-D-1	S
+		GOSUB	SHORT+		A[W]=result	C[S]=sign

		CD1EX	
		CR3EX.F	A		->output
		CD1EX
		GOSUB	SHORTOUT
		CD1EX
		CR3EX.F	A		->output'
		CD1EX
		P=	0
		RTNCC

* Slow addition. First fetch the signs

ZaddSlow	A=DAT0	A		Z1len
		CD0EX
		A=A+C	A
		D0=A
		D0=D0-	1
		A=DAT0	S
		B=A	S		Z1sign
		D0=C			->Z1len

		A=DAT1	A		Z2len
		CD1EX
		A=A+C	A
		D1=A
		D1=D1-	1
		C=DAT1	S
		D=C	S		Z2sign
		D1=C			->Z2len

* Modify for SUB if necessary

		?ST=0	sQSUB
		GOYES	Zaddslow10
		D=-D-1	S	Add -D1 to D0
		?ST=0	sQSWAP
		GOYES	Zaddslow10
		B=-B-1	S	Vice versa if arguments are swapped
		D=-D-1	S
Zaddslow10

* First get the shorter one to Z1
		A=DAT0	A		Z1len
		C=DAT1	A		Z2len
		D0=D0-	5
		D1=D1-	5
		?A<=C	A
		GOYES	Zaddslow20	Get shorter to Z1
		GOSUB	QaddSwap
		CBEX	S		Swap signs too
		CDEX	S
		CBEX	S
* Copy Z1 (shorter now) to output
Zaddslow20	C=R3.F	A		->output
		CD1EX
		B=C	A		->Z2
		D0=D0+	5
		C=DAT0	A		Z1len
		D0=D0-	5
		C=C+CON	A,5		Z1size
		D=C	A		Z1size
		GOSBVL	=MOVEDOWN
* Add zero nibbles to get equal sizes
		C=B	A		->Z2
		CD0EX
		RSTK=C			->Z1tail
		D0=D0+	5
		C=DAT0	A		Z2len
		C=C+CON	A,5		Z2size
		C=C-D	A		Z2size-Z1size = missing zero nibbles
		D1=D1-	1		Start over the old sign
		C=C+1	A		1 extra for clearing the sign
		GOSBVL	=WIPEOUT	D1 = ->Z'sign
* Now add the equal size hexes
		C=R3.F	A
		D1=C
		D1=D1+	10		->Z'mantissa
		C=DAT0	A		Z2size
		C=C-CON	A,7		Z2nibs (cannot be 0 in slow addition!)
		D0=D0+	5		->Z2mantissa
* Dispatch on signs
		?B#0	S
		GOYES	ZaddslowN?
		?D=0	S
		GOYES	ZaddslowPP

* Positive argument was shorter. Thus assume result will be negative
* and calculate D0-D1 (longer-shorter). If no overflow occurs then the
* result really is negative.

ZaddslowPN

* Negative argument was shorter. Thus assume result will be positive
* and calculate D0-D1 (longer-shorter). If no overflow occurs then the
* result really is positive

ZaddslowNP	GOTO	Zaddslow2

ZaddslowN?	?D=0	S
		GOYES	ZaddslowNP

* Same signs, add the mantissas. Return sign in D[S]
ZaddslowPP
ZaddslowNN	B=0	W		Carry
		P=C	0
		CSR	A
		D=C	A
		D=D-1	A
		GOC	ZaddslppWP
ZaddslLP1	A=DAT0	W
		A=A+B	W
		GOC	Zaddslpp20
		B=0	B		Clear carry
Zaddslpp20	C=DAT1	W
		A=A+C	W
		GONC	Zaddslpp30
		B=B+1	B
Zaddslpp30	DAT1=A	W
		D0=D0+	16
		D1=D1+	16
		D=D-1	A
		GONC	ZaddslLP1
ZaddslppWP	A=DAT0	WP
		A=A+B	WP
		GOC	Zaddslpp40
		B=0	B
Zaddslpp40	C=DAT1	WP
		A=A+C	WP
		GONC	Zaddslpp50
		B=B+1	B
Zaddslpp50	DAT1=A	WP
		CD0EX
		C+P+1
		CD0EX
		D0=D0+	1		Skip sign too
		CD1EX
		C+P+1
		CD1EX
		P=	0
		C=B	B
		DAT1=C	B
		GOTO	Zaddslstrip	Check the result

* Substract D1 from D0 to output, if no overflow use sign in D[S]
* Else negate & use negate of sign in D[S]

Zaddslow2	B=0	W		Carry
		P=C	0
		CSR	A
		D=C	A
		D=D-1	A
		GOC	ZaddslnpWP
ZaddslLP2	A=DAT0	W
		A=A-B	W
		GOC	Zaddslnp20
		B=0	B		Clear carry
Zaddslnp20	C=DAT1	W
		A=A-C	W
		GONC	Zaddslnp30
		B=B+1	B
Zaddslnp30	DAT1=A	W
		D0=D0+	16
		D1=D1+	16
		D=D-1	A
		GONC	ZaddslLP2
ZaddslnpWP	A=DAT0	WP
		A=A-B	WP
		GOC	Zaddslnp40
		B=0	B
Zaddslnp40	C=DAT1	WP
		A=A-C	WP
		GONC	Zaddslnp50
		B=B+1	B
Zaddslnp50	DAT1=A	WP
		CD0EX
		C+P+1
		CD0EX
		D0=D0+	1		Skip sign too
		CD1EX
		C+P+1
		CD1EX
		P=	0
		C=0	B
		C=C-B	B
		DAT1=C	B
		GONC	Zaddslstrip	Got expected result, check it
* Got unexpected result, negate mantissa & sign before exit
		D=-D-1	S
		D1=D1+	1		->tail
		A=R3.F	A
		A=A+CON	A,10		->mantissa
		CD1EX			->tail
		D1=A			->mantissa
		C=C-A	A		nibbles to negate
		C=C-1	A
		P=C	0
		CSR	A
		C=C-1	A
		GOC	Zaddslneg10
Zaddslneglp	A=DAT1	W
		A=-A	W
		GOC	Zaddslneg20
		DAT1=A	W
		D1=D1+	16
		C=C-1	A
		GONC	Zaddslneglp
Zaddslneg10	A=DAT1	WP
		A=-A	WP
		GOTO	Zaddslneg30
Zaddslneglp2	A=DAT1	W
		A=-A-1	W
Zaddslneg20	DAT1=A	W
		D1=D1+	16
		C=C-1	A
		GONC	Zaddslneglp2
		A=DAT1	WP
		A=-A-1	WP
Zaddslneg30	DAT1=A	WP
		CD1EX
		C+P+1
		D1=C
		P=	0
		D1=D1-	1
*		GOTO	Zaddslstrip
		
* Strip result of slow addition, then fix lenght & append sign from D[S]
Zaddslstrip	A=DAT1	S
		D1=D1-	1
		?A=0	S
		GOYES	Zaddslstrip
		D1=D1+	2		->Z'sign
		C=D	S
		DAT1=C	S
		D1=D1+	1		->Z'tail
		AD1EX
		D1=A
		C=R3.F	A		->output
		A=A-C	A		size (with sign)
		LC(5)	11
		?A<=C	A
		GOYES	Zaddslzero	Too short, must be zero!
		C=R3.F	A
		CD1EX
		D1=D1+	5
		A=A-CON	A,5
		DAT1=A	A		Z'len
Zaddsldone	R3=C.F	A		->output'
		C=RSTK
		CD0EX			->Z1tail
		D1=C			->Z2tail
		RTNCC
Zaddslzero	C=R3.F	A
		D1=C
		LCZ0
		DAT1=C	11
		D1=D1+	11
		CD1EX
		GONC	Zaddsldone

**********************************************************************
* Desc:		Negate polynomial
* Entry:	D0 = ->Q
* Output:	D0 = ->Qtail
* Alters:	A[A] C[A] D[A]
**********************************************************************
QNEG		D=0	A		complevel = 0

Qnegloop	A=DAT0	A
		LC(5)	=DOHSTR
		?A=C	A
		GOYES	Qnegz
		LC(2)	=DOLIST
		?A=C	A
		GOYES	Qnegp
		D=D-1	A
		D=D-1	A
Qnegp		D=D+1	A		complevel++
		D0=D0+	5
Qnegnext	?D#0	A
		GOYES	Qnegloop	Lop until complevel = 0
		RTNCC
Qnegz		D0=D0+	5		Skip Z even if it is Z0
		A=DAT0	A
		CD0EX
		C=C+A	A
		CD0EX
		LC(5)	6
		?A<=C	A
		GOYES	Qnegnext	Continue if was Z0
		D0=D0-	1
		A=DAT0	B
		A=-A-1	P		0 <--> F
		DAT0=A	B
		D0=D0+	1
		GONC	Qnegnext

**********************************************************************
ENDCODE

