	STITLE	ALG Root Finder
**********************************************************************
*		ALG library: Polynomial Root Finder Routines
**********************************************************************

**********************************************************************
* EXPORT:
*
*   QRoot	( Q #root --> Qint Qroot )
*   ZSQRT	( Z --> Z' flag )
*
**********************************************************************


**********************************************************************
* Desc:		Extract Nth power factors from polynomial
* Stack.	( Q #root --> Qint Qroot )
**********************************************************************
NULLNAME QRoot
::
  DUP#0=csedrp SETSIZEERR	( * Cannot take 0th root * )
  DUP#1= casedrop HXS 2 10	( * Trivial 1st root * )

  SWAP PSqff 			( #r Lf )

  ( Multiply the square free factors according to their power modulo root )

  HXS 2 10 DUPROT FIVE		( #r Qint Qroot Lf offset )
  BEGIN
    NEXTCOMPOB
  WHILE				( #r Qint Qroot Lf offset {fi #ki} )
    INCOMPDROP			( #r Qint Qroot Lf offset fi #ki )
    7PICK #/			( #r Qint Qroot Lf offset fi #rem #quo )
    3PICKSWAP PPow#		( #r Qint Qroot Lf offset fi #rem fi^#quo )
    7ROLL QMul 6UNROLL		( #r Qint' Qroot Lf offset fi #rem )
    PPow#			( #r Qint' Qroot Lf offset fi^#rem )
    4ROLL QMul UNROT		( #r Qint' Qroot' Lf offset )
  REPEAT
  DROP ROTDROP			( Qint Qroot )
;

**********************************************************************
* Desc:		Calculate integer part of square root
* Stack:	( Z --> Z' flag )
*		where flag = TRUE if square root is exact
**********************************************************************
NULLNAME ZSQRT
::

* First try fast algorithm in registers only

  CODE
		GOSBVL	=PopASavptr	A[A] = ->Z
		D0=A
		D0=D0+	5
		A=DAT0	A		A[A] = Zlen
		LC(5)	5+15+14+1	Max length of mantissa < 30 nibbles
		?A<C	A
		GOYES	fastzsqrt

		GOSBVL	=GETPTR		Unpop Z
		D1=D1-	5
		D=D-1	A
		GOVLNG	=PushFLoop	Push FALSE for failure

fastzsqrt	A=A-CON	A,7		mantissa length - 1
		GONC	fastsqrtx

fastsqrt0	A=0	W		sqrt(0) = 0
		GOSBVL	=PUSHhxs
		GOTO	truetrue	exact

fastsqrtx	C=A	A		mantissa length - 1
		D0=D0+	5		->mantissa
		CPEX	0
		?C#0	A
		GOYES	fastsqrt2W	Z takes 2 words - handle it
		A=0	W
		A=DAT0	WP		A[W] = mantissa
		GOSUB	ISQRT		A[W] = sqrt	C[W] = remainder
		GOTO	fastsqrtpush	Now push the result

fastsqrt2W	C=DAT0	W		low 16 nibbles
		D0=D0+	16
		A=0	W
		A=DAT0	WP
		GOSUB	ISQRT2		A[W] = sqrt	C[W] = remainder

fastsqrtpush	R4=C	W		R4[W] = remainder
		GOSUBL	PushN		Push A[W] as Z
		A=R4	W		remainder
		LC(5)	=FalseTrue	assume inexact
		?A#0	W
		GOYES	+
truetrue	LC(5)	=TrueTrue
+		A=C	A
		PC=(A)

*---------------------------------------------------------------------
* Desc:		Double square root calculating subroutine
* Entry:	A[W] = high_nibbles
*		C[W] = low nibbles
* Exit:		A[W] = SQRT
*		C[W] = remainder
*---------------------------------------------------------------------

ISQRT2		R0=C			R0[W] = low nibbles
		GOSUB	ISQRT
		B=A	W
		A=R0	W
		D=C	W
		LC(2)	32-1
		CDEX	W
		GOTO	ISQRTLP

*---------------------------------------------------------------------
* Desc:		Square root calculating subroutine
* Entry:	A[W] = X
* Exit:		A[W] = SQRT(X)
*		C[W] = remainder
*---------------------------------------------------------------------

ISQRT		P=	0
		C=0	W		remainder = 0
		?A=0	W		sqrt(0) = 0
		RTNYES
		LC(2)	32-1
		D=C	A		D[B]=loop counter
		B=0	W		B[W]=result
		C=0	A		C[W]=remainder
-		?A#0	S		Find range faster
		GOYES	+
		ASL	W
		D=D-CON	B,2
		GONC	-
+
ISQRTLP		C=C+C	W		Generate bit 1
		A=A+A	W
		GONC	+
		C=C+1	P
+		C=C+C	W		Generate bit 2
		A=A+A	W
		GONC	+
		C=C+1	P
+		B=B+B	W		result *= 2
		B=B+B	W
		B=B+1	P
		?C<B	W		y < result*2+1
		GOYES	+
		C=C-B	W		y -= result*2+1
		B=B+1	P
+		BSRB			fix back to correct order
		D=D-1	B
		GONC	ISQRTLP
		A=B	W		A[W] = result
		RTN
*---------------------------------------------------------------------
  ENDCODE

* Quit if done

  ?SEMI

* Make a good guess on start value by taking the square root
* of the leading 32 nibbles. This provides a great speedup
* for integers of that magnitude.

  CODE
		GOSBVL	=SAVPTR
		A=DAT1	A
		D0=A
		D0=D0+	5
		C=DAT0	A		Zlen
		AD0EX
		D0=A
		A=A+C	A		->Ztail
		D1=A
		D1=D1-	14
		?CBIT=1	0
		GOYES	+		Adjust to byte boundary
		D1=D1-	1
+		A=0	W
		A=DAT1	14
		D1=D1-	16
		C=DAT1	W
		GOSUB	ISQRT2		A[W] = sqrt
		R1=A	W

		C=DAT0	A		Zlen
		C=C-CON	A,6
		SB=0
		CSRB.F	A
		?SB=0
		GOYES	+
		C=C+1	A		Adjust to nibble boundary
+		C=C+1	A
		RSTK=C
		GOSBVL	=MAKE$N
		C=R0
		D1=C			->$
		LC(5)	=DOHSTR
		DAT1=C	A
		D1=D1+	10
		C=RSTK
		C=C-CON	A,15+1
		A=0	W
		A=A-1	W		A[W]=FFFFFFFFFFFFFFFF
		GOSBVL	=STUFF		Make sure estimate is too big
		A=R1	W
		DAT1=A	W		High bits and the sign (=0)
		GOVLNG	=GPPushR0Lp
  ENDCODE

* Now use Newton's algorithm with the given guess

  DUP					( Z Zg Zg )
  BEGIN
	SWAPDROP			( Z Zg )
	2DUP QDiv			( Z Zg Z/Zg )
	OVER QAdd			( Z Zg Zg+Z/Zg )
	HXS 2 20 QDiv			( Z Zg [Zg+Z/Zg]/2 )
	2DUPSWAP ZNLT?			( Z [Zg+Z/Zg]/2 Zg flag )
  NOT_UNTIL (UNTIL)
  DROP					( Z Zsqrt )
  DUPDUP QMul				( Z Zsqrt Zsqrt^2 )
  ROT EQUAL				( Zsqrt flag )
;

**********************************************************************
