	STITLE	ALG String Matrix
**********************************************************************
*		ALG Library: String Matrix Subroutines
**********************************************************************

**********************************************************************
* DEFINITION:
*
* String matrix (MS) is a matrix of short rational integer elements
* stored in a string for fast assembly access as follows:
*
*	ASSEMBLE
*		CON(5)	=DOCSTR		String prolog
*		REL(5)	EndOfMS		Size of string
*					Body of actual array follows
*		CON(5)	Y		Y-dimension
*		CON(5)	X		X-dimension
*		CON(16)	el11		Element 1,1
*		CON(16)	el12		Element 1,2
*		..
*		CON(16)	elYX		Element Y,X
*	EndOfMS
*	RPL
*
*	The elements themselves are formed by a 15-nibble mantissa + sign:
*
*		CON(15)	integer
*		CON(1)	sign		0/F for +/-
*
* It must be noted that any subroutine operating on a MS does all
* math in place. Therefore the original MM must be saved in case
* the MS-function fails due to an intermediate overflow.
*
**********************************************************************


**********************************************************************
* EXPORT:
*
*	MM2MS		( MM --> MM FALSE / MM MS TRUE )
*	MS2MM		( MS --> MM )
*	MScrout		( MS --> MS' FALSE / MS' #rank TRUE )
*	MSsolvetri	( MS --> MS' flag )
**********************************************************************


**********************************************************************
* Name:		MM2MS
* Desc:		Convert meta matrix to string matrix if possible
* Stack:	( MM --> MM FALSE / MM MS TRUE )
**********************************************************************
NULLNAME MM2MS
CODE
		GOSBVL	=SAVPTR
		D1=D1+	5		Skip NULL$ ID
		GOSBVL	=POP2#
		R2=A	A		R2[A]=Y
		R3=C	A		R3[A]=X
		GOSBVL	=MUL#
		B=B+B	A		B[A]=2*X*Y
		A=B	A
		R4=A	A		R4[A]=2X*Y

* Now check all MM elements are short rational integers

MM2MSloop1	B=B-1	A
		GOC	MM2MSok1	All passed the check
		A=DAT1	A
		D1=D1+	5
		D0=A			->element (nominator or denominator)
		A=DAT0	A
		LC(5)	=DOHSTR
		?A#C	A
		GOYES	MM2MSfail	Not Z - fail
		D0=D0+	5
		A=DAT0	A
		LC(5)	5+16		Allow 16 nibbles..
		?A<=C	A		.. would make sense to use ~8
		GOYES	MM2MSloop1
MM2MSfail	GOVLNG	=GPPushFLoop	( --> MM FALSE )

* Now allocate the space for the MS

MM2MSok1	C=R4	A		2*X*Y
		CSL	A		* 16 for each rational
		C=C+CON	A,10		+ 10 for the dimensions
		GOSBVL	=MAKE$N		R0[A] = ->$
		A=R2	A
		DAT0=A	A		Store Y
		D0=D0+	5
		A=R3	A
		DAT0=A	A		Store X
		D0=D0+	5
		AD1EX			Skip to first MM element
		C=R4	A		2*X*Y
		D=C	A		D[A]=index = 2*X*Y
		A=A+C	A
		C=C+C	A
		C=C+C	A
		A=A+C	A
		D1=A
		D1=D1+	10		Skip extra due to MM dims

* Now copy the elements from MM to MS

MM2MSloop2	D=D-1	A
		GOC	MM2MSok2	Copied all
		A=DAT1	A
		D1=D1-	5
		AD1EX
		B=A	A		B[A]=->stack pos
		A=0	W		Init mantissa
		D1=D1+	5
		C=DAT1	A		Zlen
		D1=D1+	5
		C=C-CON	A,7
		GOC	+		Handle 0-special case
		P=C	0
		A=DAT1	WP		Read mantissa
		CD1EX
		C+P+1
		CD1EX
		P=	0
+		A=DAT1	S		Read sign
		DAT0=A	W
		A=B	A
		D1=A			->stack pos
		D0=D0+	16
		GONC	MM2MSloop2

* Conversion done - exit

MM2MSok2	A=R0	A		->$
		GOSBVL	=GPPushA
		GOVLNG	=PushTLoop	( --> MM MS TRUE )
ENDCODE

**********************************************************************
* Name:		MS2MM
* Desc:		Convert string matrix to meta matrix
* Stack:	( MS --> MM )
**********************************************************************
NULLNAME MS2MM
CODE
		GOSBVL	=SAVPTR
		A=DAT1	A
		D0=A			->MS
		D0=D0+	10
		A=DAT0	A
		R2=A	A		R2[A]=Y
		D0=D0+	5
		C=DAT0	A
		R3=C	A		R3[A]=X
		D0=D0+	5
		GOSBVL	=MUL#
		B=B+B	A		B[A]=2*X*Y
		C=B	A
		RSTK=C			RSTK=index = 2*X*Y
		LC(5)	10+10
		R4=C	A		current offset into MS
MS2MMloop	C=RSTK
		C=C-1	A		index--
		GOC	MS2MMdone	Done pushing elements
		RSTK=C
		A=DAT0	W		A[W]=element
		C=A	S		C[S]=sign
		A=0	S		A[W]=mantissa
* Following could be replaced by a call to PushZ
		?A=0	W
		GOYES	++		Handle 0 properly
		P=	14
-		?A#0	P		Find most significant nibble
		GOYES	+
		CSR	W
		P=P-1
		GONC	-
+		P=P+1
		A=C	P		Put sign to high nibble
++		GOSBVL	=PUSHhxs	(Uses R0,R1)
* Fix stack ordering
		C=DAT1	A		->Z
		D1=D1+	5
		A=DAT1	A		->MS
		DAT1=C	A
		D1=D1-	5
		DAT1=A	A
		GOSBVL	=SAVPTR
		C=R4	A		offset counter
		C=C+CON	A,16		offset++
		R4=C	A
		A=A+C	A		->next element
		D0=A
		GONC	MS2MMloop	Loop until pushed all

* Now push dimensions and NULL$ ID
MS2MMdone	A=R2	A
		R0=A	A		R0[A]=Y
		A=R3	A
		R1=A	A		R1[A]=X
		GOSBVL	=PUSH2#		Push Y,X
		A=DAT1	10
		D1=D1+	5		Overwrite MS,Y,X with Y,X,X
		DAT1=A	10
		D1=D1-	5
		LC(5)	=NULL$		Overwrite Y,X,X with Y,X,NULL$
		DAT1=C	A
		GOVLNG	=Loop
ENDCODE

**********************************************************************
* Name:		MScrout
* Desc:		LU decomposition of string matrix
* Stack:	( MS --> MS' FALSE )		if overflow
*		( MS --> MS' #rank TRUE )	if success
*		( MS --> #rank ) + ERRJMP	if not full rank
**********************************************************************
NULLNAME MScrout
CODE
		ST=0	sTRASH		No top-tempob delete if overflow
		GOSBVL	=SAVPTR
		A=DAT1	A
		A=A+CON	A,10
		R4=A	A		R4[A] = ->MS.Y
		D0=A
		A=DAT0	A
		A=A-1	A
		?A#0	A
		GOYES	MScroutN	Handle > 1x1 matrix

* Handle 1x1 matrix

		D0=D0+	10		->first element
		A=DAT0	W
		A=0	S		Make sure sign is ignored
		?A=0	W
		GOYES	MScrouterr	Element is zero - error
		A=0	A
		A=A+1	A		rank=1
		GOTO	MScroutdone

MScrouterr	R0=A	A		Push #rank
		GOSBVL	=PUSH#
		A=DAT1	A		Overwrite MS with #rank
		D1=D1+	5
		D=D+1	A
		DAT1=A	A
MSinfres	LC(5)	#305		"Infinite result"
		GOVLNG	=ErrjmpC

* Handle > 1x1 matrix
*
* Keep R3[W] as a stack of loop indexes
* Each new loop shifts R3[W] to the left, each exitloop to the right
* Lowest index is always called i, next j, next k so that the algorithm
* is easy to compare with the SRPL one which uses INDEX@, JINDEX@ etc

MScroutN	A=0	A
		R3=A	A		LOOP: [0,Y-1)

MScroutlp1	A=R3	A		A[A]=i
		C=A	A		C[A]=i
		GOSUB	MSget		D0 = ->el(i,i)
		A=DAT0	W
		A=0	S		Make sure sign is ignored
		?A#0	W
		GOYES	MSpivotok

* Find nonzero pivot - accept the first one found for simplicity

		C=R3	A		i
		C=C+1	A
		R2=C	A		R2[A]=j=i+1 = test pivot row
MSpivotloop	A=R2	A		j
		C=R3	A		i
		GOSUB	MSget		D0 = ->el(j,i)
		A=DAT0	W
		A=0	S		Make sure sign is ignored
		?A#0	W
		GOYES	MSgotpivot	Found pivot
		A=R2	A
		A=A+1	A		j=j+1
		R2=A	A
		C=R4	A
		D0=C
		C=DAT0	A		C[A]=Y
		?A<C	A
		GOYES	MSpivotloop

* Failed to find a valid pivot - error

		A=R3	A		rank = i
		GOTO	MScrouterr	Handle the error condition

MSgotpivot	A=R2	A		A[A]=j
		C=R3	A		C[A]=i
		GOSUB	MSrowswap	Swap the rows
		GOTO	MScroutlp1	And jump back to get the pivot again

* Now pivoting has been done

MSpivotok	AD0EX
		D1=A			D1 = ->pivot

		A=R3	W
		GOSBVL	=ASLW5		A[A1]=i
		A=R3.F	A
		A=A+1	A		A[A]=i+1
		R3=A	W		LOOP: [i+1,Y)

MScroutlp2	C=R3	W
		A=C	A		A[A]=i
		GOSUB	csrw5		C[A]=j
		GOSUB	MSget		D0 = ->el(i,j)

		A=DAT1	W		Divide element by pivot
		D0=D0+	16
		C=DAT0	W
		D0=D0-	16
		GOSUB	ZSMUL		A[W]=denominator
		R1=A	W
		D1=D1+	16
		A=DAT1	W
		D1=D1-	16
		C=DAT0	W
		GOSUB	ZSMUL		A[W]=nominator
		C=R1	W
		GOSUB	ZSCANNO		A[W]/C[W] = new element
		DAT0=A	W
		D0=D0+	16
		DAT0=C	W
		R0=A	W		R0[W]/R1[W]= element/pivot
		R1=C	W
		
* Now if the result is zero we don't need to adjust the other elements

		?A#0	W
		GOYES	+
		A=R4	A		Prepare loop3 exit registers
		D0=A
		C=R3	W
		GOTO	MScroutxlp3	Jump to loop3 exit

* Adjust other elements

+		A=R3	W
		GOSUB	asrw5		A[A]=j
		A=A+1	A
		C=R3	W
		GOSBVL	=CSLW5
		C=A	A
		R3=C	W		LOOP: [j+1,X)

MScroutlp3	C=R3	W		C[A]=i
		A=C	W
		GOSUB	asrw5
		GOSUB	asrw5		A[A]=k
		GOSUB	MSget		D0= ->el(k,i)
		A=DAT0	W		Multiply with element/pivot
		C=R0	W
		GOSUB	ZSMUL
		R2=A	W		R2[W]=nominator
		D0=D0+	16
		A=DAT0	W
		C=R1	W
		GOSUB	ZSMUL
		C=A	A
		D=C	W		D[W]=denominator

		A=R3	W
		C=A	A		C[A]=i
		GOSUB	asrw5		A[A]=j
		GOSUB	MSget		D0 = ->el(j,i)

* Substract the previously calculated value from this element
* Rational substraction is annoying due to the many registers it needs..

		D0=D0+	16
		C=DAT0	W
		?C#D	W
		GOYES	MScroutrat	Handle rational substraction

		D0=D0-	16		Same denominator - sub nominators only
		A=DAT0	W
		C=R2	W
		GOSUB	ZSSUB
		GOTO	++		Simplify regardless

MScroutrat	A=C	W		Calculate new denominator
		C=D	W
		GOSUB	ZSMUL
		C=DAT0	W
		DAT0=A	W		And store it
		A=R2	W
		GOSUB	ZSMUL		1st part of nominator sum
		R2=A	W
		D0=D0-	16
		A=DAT0	W
		C=D	W
		GOSUB	ZSMUL		2nd part of nominator sum
		C=R2	W
		GOSUB	ZSSUB		And substract them
++		D0=D0+	16
		C=DAT0	W
		GOSUB	ZSCANNO		Simplify the result
		DAT0=C	W		And store it
		D0=D0-	16
		DAT0=A	W

		A=R4	A		Repeat loop until stopvalue X
		D0=A
		D0=D0+	5
		A=DAT0	A		A[A] = X
		C=R3	A
		C=C+1	A		i=i+1
		R3=C	A
		?C>=A	A
		GOYES	+		Exit the loop
		GOTO	MScroutlp3
+		GOSUB	csrw5		2 loops remain
		D0=D0-	5
MScroutxlp3	C=C+1	A		i=i+1
		R3=C	W
		A=DAT0	A		A[A]=Y
		?C>=A	A
		GOYES	+		Stop value Y reached
		GOTO	MScroutlp2
+		GOSUB	csrw5		1 loop remains
		C=C+1	A
		R3=C	A
		A=A-1	A		A[A]=Y-1
		?C>=A	A
		GOYES	+		Stop value Y-1 reached
		GOTO	MScroutlp1
* Final checks against nonfull rank
+		C=A	A		A[A]=C[A]=Y-1
		GOSUB	MSget		D0 = ->el(Y-1,Y-1)
		C=DAT0	W
		A=R4	A
		D0=A
		A=DAT0	A		A[A]=Y
		C=0	S		Make sure sign is ignored
		?C#0	W
		GOYES	MScroutdone	Success - exit with true
		A=A-1	A		A[A]=Y-1
		GOTO	MScrouterr	Handle nonfull rank

MScroutdone	R0=A	A		R0[A] = rank
		GOVLNG	=Push#TLoop	( --> MS #rank TRUE )

asrw5		GOVLNG	=ASRW5		Code savings
csrw5		GOVLNG	=CSRW5

ENDCODE

**********************************************************************
* Name:		MSsolvetri
* Desc:		Solve triangular system
* Stack:	( MS --> MS' FALSE )		if overflow
*		( MS --> MS' TRUE )		if success
**********************************************************************
NULLNAME MSsolvetri
CODE
		ST=0	sTRASH		No top-tempob delete if overflow
		GOSBVL	=SAVPTR
		A=DAT1	A
		A=A+CON	A,10
		R4=A	A
		D1=A			D1 = ->MS.Y
		A=0	A
		R3=A	A		LOOP: [0,Y)
MSsolvelp1	A=R3	W
		GOSBVL	=ASLW5
		R3=A	W		LOOP: [0,X-Y)
MSsolvelp2	A=DAT1	A		A[A]=Y
		C=R3	W
		B=C	A		B[A]=i
		GOSUB	csrw5		C[A]=j
		D=C	A		D[A]=j
		A=A-C	A
		A=A-1	A		A[A]=Y-j-1
		C=DAT1	A
		C=C+B	A		C[A]=Y+i
		GOSUB	MSget		D0 = ->el(Y-j-1,Y+i)
		A=DAT0	W
		D0=D0+	16
		C=DAT0	W
		R0=A	W		R0/R1 = element to modify
		R1=C	W

		?D#0	A
		GOYES	+
		GOTO	MSsolvefix	No adjustments at the bottom row
+		C=R3	W
		GOSBVL	=CSLW5
		C=C+1	A
		R3=C	W		LOOP: [1,j+1)

MSsolvelp3	A=DAT1	A		A[A]=Y
		C=R3	W
		GOSUB	csrw5
		GOSUB	csrw5		C[A]=k
		A=A-C	A
		A=A-1	A		A[A]=Y-k-1
		C=R3	A
		C=A+C	A		C[A]=Y-k-1+i
		D=C	A		D[A]=Y-k-1+i
		GOSUB	MSget		D0 = ->el(Y-k-1,Y-k-1+i)
		C=R3	W
		GOSUB	csrw5		C[A]=j
		CDEX	A
		A=C	A		A[A]=Y-k-1+i
		C=DAT1	A		C[A]=Y
		D=D+C	A		D[A]=Y+j
		CD0EX
		D1=C			D1 = ->el(Y-k-1,Y-k-1+i)
		C=D	A		C[A]=Y+j
		GOSUB	MSget		D0 = ->el(Y-k-1+i,Y+j)

		A=DAT0	W		Multiply nominators
		C=DAT1	W
		GOSUB	ZSMUL

		?A=0	W		HHUMM, check this later
		GOYES	MSsolveok3	but I think it's right

		A=-A-1	S
		R2=A	W		-nominator
		D0=D0+	16
		D1=D1+	16
		A=DAT0	W		Multiply denominators
		C=DAT1	W
		GOSUB	ZSMUL

* Now add the result to the element being modified

		C=R1	W
		?A#C	W
		GOYES	+		Different denominator - add rat
		A=R0	W
		C=R2	W
		GOSUB	ZSADD		Just add nominators
		GOTO	++		And simplify
+		D=C	W
		R1=A	W
		GOSUB	ZSMUL		Multiply denominators
		AR1EX	W
		C=R0	W
		GOSUB	ZSMUL		1st part of nominator sum
		R0=A	W
		A=R2	W
		C=D	W
		GOSUB	ZSMUL		2nd part of nominator sum
		C=R0	W
		GOSUB	ZSADD		New nominator
++		C=R1	W		Uncomment to prevent too early
		GOSUB	ZSCANNO		overflows due to RSub
		R0=A	W
		R1=C	W

MSsolveok3	A=R4	A
		D1=A			D1 = ->MS.Y
		C=R3	A
		C=C+1	A		i=i+1
		R3=C	W
		A=C	W
		GOSUB	asrw5
		GOSUB	asrw5
		?C>A	A
		GOYES	+
		GOTO	MSsolvelp3
+		GOSUB	csrw5		2 loops remaining
		R3=C	W

* Now divide by diagonal element and store result

MSsolvefix	C=R3	W		
		GOSUB	csrw5		j
		A=DAT1	A		A[A] = Y
		A=A-C	A
		A=A-1	A		A[A] = Y-j-1
		C=A	A		C[A] = Y-j-1
		D=C	A		D[A] = Y-j-1
		GOSUB	MSget		D0 = ->el(Y-j-1,Y-j-1)
		A=DAT0	W
		C=R1	W
		GOSUB	ZSMUL
		AR0EX
		D0=D0+	16
		C=DAT0	W
		GOSUB	ZSMUL
		R1=A	W

* Store result

		A=DAT1	A		A[A] = Y
		C=R3	A
		A=A+C	A		A[A] = Y+i
		C=D	A		C[A] = Y-j-1
		ACEX	A		A[A] = Y-j-1	C[A] = Y+i
		GOSUB	MSget		D0 = ->el(Y-j-1,Y+i)
		A=R1	W
		C=R0	W
		GOSUB	ZSCANNO		Simplify result before storing
		DAT0=A	W
		D0=D0+	16
		DAT0=C	W

		C=R3	W
		C=C+1	A		i=i+1
		R3=C	W
		A=DAT1	A
		B=A	A		B[A]=Y
		D1=D1+	5
		A=DAT1	A		A[A]=X
		D1=D1-	5
		A=A-B	A		A[A]=X-Y
		?C>=A	A
		GOYES	+
		GOTO	MSsolvelp2

+		GOSUB	csrw5		1 loop remaining
		C=C+1	A		i=i+1
		R3=C	W
		A=DAT1	A		A[A]=Y
		?C>=A	A
		GOYES	+
		GOTO	MSsolvelp1
+		
* Done solving - return solved right hand side only

		A=DAT1	A
		R0=A	A		R0[A]=Y
		D1=D1+	5
		A=DAT1	A
		R1=A	A		R1[A]=X
		D1=D1+	5		D1 = ->output
		A=0	A
		R3=A	A		LOOP: [0,Y)
MSsolvesublp	A=R3	A		A[A]=i
		C=R0	A		C[A]=Y
		GOSUB	MSget		D0 = ->el(i,Y)
		A=R0	A
		C=R1	A
		C=C-A	A		C[A]=X-Y
		C=C+C	A		Copy 2(X-Y) elements
-		C=C-1	A
		GOC	+		Done copying this row
		A=DAT0	W
		DAT1=A	W
		D0=D0+	16
		D1=D1+	16
		GONC	-
+		C=R3	A
		C=C+1	A		i=i+1
		R3=C	A
		A=R0	A		A[A]=Y
		?C<A	A
		GOYES	MSsolvesublp

* Now change the dimensions after MSget is no longer needed

		C=R4	A
		D0=C			D0 = ->MS.Y
		D0=D0+	5
		C=R1	A
		C=C-A	A		C[A] = X-Y
		DAT0=C	A		New X dimension
		D0=D0-	15		->$
		CD0EX
		R0=C	A		R0[A]=->$
		D0=C
		D0=D0+	5		D0 = ->$len
		AD1EX			A[A] = ->end of output
		D1=A
		A=A-C	A		A[A] = $size
		C=DAT0	A		C[A] = $oldlen
		A=A-CON	A,5		A[A] = $len
		DAT0=A	A		Fix $len
		A=A+CON	A,5+6		A[A] = $size+6
		DAT1=A	A		Fix link
		AD0EX
		A=A+C	A		A[A] = ->end of old
		AD0EX
		D0=D0+	5		Skip past the old link
		C=C-A	A
		C=C+CON	A,11		C[A] = amount to strip
		AD0EX
		GOSBVL	=MOVERSD	Strip
		GOVLNG	=GPPushTLoop	( --> MS TRUE )
ENDCODE



**********************************************************************
*		MS assembly utilities
**********************************************************************
ASSEMBLE

*---------------------------------------------------------------------
* Name:		MSget
* Desc:		Get MS element
* Entry:	A[A]=y	C[A]=x	R4[A]=->MSbody
* Exit:		A[A]=D0 = ->element
* Alters:	A[A] B[A] C[A] D0
*---------------------------------------------------------------------
MSget		RSTK=C			RSTK=x
		C=R4.F	A
		D0=C
		D0=D0+	5
		C=DAT0	A		C[A]=X
		D0=D0+	5		->first element
		GOSBVL	=MUL#		B[A]=y*X
		C=RSTK			C[A]=x
		C=C+B	A		C[A]=y*X+x
		CSL	A
		C=C+C	A		*32 to get skip in nibbles
		AD0EX
		A=A+C	A
		D0=A			->element y,x
		RTNCC

*---------------------------------------------------------------------
* Name:		MSrowswap
* Desc:		Swap rows in MS
* Entry:	A[A]=y1	C[A]=y2	R4[A]=->MSbody
* Exit:
* Alters:	A[W] B[A] C[W] D[A] D0 D1
*---------------------------------------------------------------------
MSrowswap	D=C	A		D[A]=y2
		C=0	A
		GOSUB	MSget		A[A]=D0= ->element y1,0
		D1=A
		C=D	A
		A=C	A
		C=0	A
		GOSUB	MSget		D0= ->element y2,0
		C=R4	A
		CD0EX
		D0=D0+	5
		A=DAT0	A		A[A]=X
		D0=C
		B=A	A
		B=B+B	A		B[A]=index = 2*X
MMrowswaploop	B=B-1	A		index--
		RTNC
		A=DAT0	W
		C=DAT1	W
		DAT1=A	W
		DAT0=C	W
		D0=D0+	16
		D1=D1+	16
		GONC	MMrowswaploop

*---------------------------------------------------------------------
* Name:		ZSSUB
* Desc:		Signed short substraction
* Entry:	A[W]=z1	C[W]=z2
* Exit:		A[W]=z1-z2
* Alters:	A[W] C[W] R3[S]
*---------------------------------------------------------------------
ZSSUB		C=-C-1	S		z2 = -z2

*---------------------------------------------------------------------
* Name:		ZSADD
* Desc:		Signed short addition
* Entry:	A[W]=z1	C[W]=z2
* Exit:		A[W]=z1+z2
* Alters:	A[W] C[W] R3[S]
*---------------------------------------------------------------------
ZSADD		R3=A.F	S		Assume z1 sign will be correct
		?A#C	S
		GOYES	+		Handle substraction

		A=0	S		Handle addition
		C=0	S
		A=A+C	W		Always CC since signs are clear
		GONC	ZSCHECK		Check result for overflow

+		A=0	S
		C=0	S
		?A>=C	W		Check which mantissa is bigger
		GOYES	+		A[W] was - R3[S] is correct
		AR3EX.F	S
		A=-A-1	S		Negate answer
		AR3EX.F	S
		ACEX	W		Swap max to A[W]
+		A=A-C	W

ZSCHECK		?A=0	W		Don't restore sign if zero result
		RTNYES
		?A#0	S
		GOYES	ZSFAIL		Overflow!!
		A=R3.F	S
		RTN

ZSFAIL		P=	0		Safety
		GOSUBL	?TrashTop	Trash top-tempob if requested
		GOVLNG	=GPPushFLoop	( --> MS FALSE )

*---------------------------------------------------------------------
* Name:		ZSMUL
* Desc:		Signed short multiplication
* Entry:	A[W]=z1	C[W]=z2
* Exit:		A[W]=z1*z2
* Alters:	A[W] B[W] C[W] R3[S]
*---------------------------------------------------------------------
ZSMUL		A=A+C	S		Multiply the signs first
		GONC	+
		A=0	S		negneg = pos
+		R3=A.F	S		Save sign
ZSMULentry	A=0	S
		C=0	S
		B=0	W		init result=0
		?C<=A	W		Make sure C[W] is min for speed
		GOYES	zsmulloop	(just consider 0*X)
		ACEX	W
zsmulloop	SB=0			Regular multiplication loop
		CSRB
		?SB=0
		GOYES	+
		B=B+A	W
		GOC	ZSFAIL
+		A=A+A	W
		GOC	ZSFAIL
		?C#0	B		Speed up trick
		GOYES	zsmulloop
		?C#0	W
		GOYES	zsmulloop
		A=B	W		A[W]=result mantissa
		GONC	ZSCHECK		Check the result

*---------------------------------------------------------------------
* Name:		ZSDIV
* Desc:		Signed short division
* Entry:	A[W]=z1	C[W]=z2
* Exit:		A[W]=quo
*		B[W]=rem
* Alters:	A[W] B[W] C[W]
*---------------------------------------------------------------------
ZSDIV		A=A+C	S		Multiply the signs first
		GONC	+
		A=0	S		negneg=pos
+		R3=A.F	S		Save sign
		A=0	S
		C=0	S
		GOSUBL	BDIV		A[W]=quo	B[W]=rem
		A=R3.F	S		Set quotient sign
		RTNCC

*---------------------------------------------------------------------
* Name:		ZSCANNO
* Desc:		Simplify signed short rational
* Entry:	A[W]=z1		C[W]=z2
* Exit:		A[W]=z1'	C[W]=z2'
* Alters:	A[W] B[W] C[W] D[W] R3[S]
*---------------------------------------------------------------------

ZSCANNO		A=A+C	S		Multiply the signs
		GONC	+
		A=0	S		negneg=pos
+		R3=A.F	S		R3[S]=sign
		A=0	S
		C=0	S
		GOSUBL	SHORTCANNO	Unsigned simplification
		?A=0	W
		RTNYES			Don't bother with sign if zero
		A=R3.F	S
		RTN

**********************************************************************
RPL

