	STITLE	ALG Sym2R
**********************************************************************
*		ALG Library: Symbolic => Rational routines
**********************************************************************

**********************************************************************
* EXPORT:
*
*   SortList	( L pred --> L' )
*   PVars{}	( Symb L --> L' )
*   PVars	( Symb --> L )
*
*   A2R		( Symb v --> Pn Pd )
*
*   AA2RR{}	( Ap Aq vars --> v Pn Pd Qn Qd )
*   AA2PP{}	( Ap Aq vars --> v P Zp Q Zq )
*   AA2PP1{}	( Ap Aq vars --> v P Q )
*   AA2RR	( Ap Aq --> v Pn Pd Qn Qd )
*   AA2PP	( Ap Aq --> v P Zp Q Zq )
*
*   AAGcd	( Ap Aq --> Agcd )
*   AALcm	( Ap Aq --> Alcm )
*   RNumDen	( A -> An Ad )
**********************************************************************

**********************************************************************
* Desc:		Convert from meta to integer
* Stack:	( M --> Z )
**********************************************************************
NULLNAME M2Z
::
  M-1stcasechs :: M2Z QNeg ;
  DROP R>Z
;

**********************************************************************
* Desc:		Sort list according to predicate (a>b)
* Stack:	( L pred --> L' )
**********************************************************************
NULLNAME SortList
::
  OVER LENCOMP TWO #< caseDROP		( Ignore {} and {ob} )
  1LAMBIND				( Bind predicate )

  ( Get 1st object into it's place )
  ONESWAP FIVE NEXTCOMPOB DROP 4UNROLL	( Meta {} #off )

  ( Avoid a WHILE loop this way )
  NEXTCOMPOB DROP			( Meta {} #off ob2 )
  BEGIN
	4PICK				( Meta {} #off ob #maxpos )
	BEGIN
		2DUP #6+ PICK		( Meta {} #off ob #maxpos ob ob' )
		1GETLAM EVAL		( Meta {} #off ob #maxpos flag )
		ITE
		:: #1- DUP#0= ;
		TRUE
	UNTIL				( Meta {} #off ob #pos )
	#4+ UNROLL ROT#1+UNROT		( Meta' {} #off )
	NEXTCOMPOB
  NOT_UNTIL (UNTIL)
  DROP {}N				( {}' )
  ABND
;

**********************************************************************
* Desc:		True if v1 is lexicographically after v2
* Stack:	( v1 v2 --> flag )
**********************************************************************
NULLNAME IsV>V?
CODE
		GOSBVL	=PopASavptr
		C=DAT1	A
		D0=C				->stk2
		D1=A				->stk1
		GOSUB	cmpobs
		GOVLNG	=GPOverWrT/FL		( --> flag )

* Do a > test on objects

cmpobs		A=DAT0	A		First do a prolog test so that
		C=DAT1	A		id comes before symb
		?A=C	A
		GOYES	cmpbody
		?A<C	A
		RTNYES
		RTN

cmpbody		LC(5)	=DOIDNT		Dispatch based on type
		?A=C	A
		GOYES	cmpids
		LC(2)	=DOLAM
		?A=C	A
		GOYES	cmpids
		GOTO	cmpmisc

cmpids		D0=D0+	5		Compare body of id/lam
		D1=D1+	5
		A=DAT0	B		chars2
		C=DAT1	B		chars1
		D0=D0+	2
		D1=D1+	2
		ST=0	0		Flag to indicate the longer one
		?A<=C	B
		GOYES	cmpidnow
		ST=1	0		stk2 is longer
		A=C	B
cmpidnow	B=A	B		length counter
cmpidlp		B=B-1	B
		GOC	cmpidexit	equal body - test length
		A=DAT0	B
		C=DAT1	B
		D0=D0+	2
		D1=D1+	2
		?A=C	B
		GOYES	cmpidlp		Loop until end or nonequal char
		?A>=C	B
		RTNYES			CS if >
		RTNCC
cmpidexit	?ST=1	0
		RTNYES			CS if > (length)
		RTNCC

* cmpmisc is designed for symbs, but works for all types
* it simply compares all nibbles in the body in reverse order

cmpmisc		AD0EX
		B=A	A		->ob2
		D0=A
		GOSBVL	=SKIPOB
		CD1EX
		D=C	A		->ob1
		CD0EX
		D1=C			->ob2end
		C=C-B	A
		B=C	A		size2
		GOSBVL	=SKIPOB
		CD0EX			->ob1end
		D0=C
		C=C-D	A		size1

		?B<=C	A		B[A] = min(size1,size2)
		GOYES	cmpmisclp
		B=C	A
cmpmisclp	B=B-1	A		make sure test finishes
		GOC	cmpmisc<
		D0=D0-	1
		D1=D1-	1
		A=DAT0	B
		C=DAT1	B
		?A=C	P
		GOYES	cmpmisclp	Loop until nonequal
		?C>=A	P
		RTNYES			CS: >
cmpmisc<	RTNCC			CC: <
ENDCODE

**********************************************************************
* Name:		PVars{}
* Desc:		Collect variables from symbolic with given
*		initial ordering.
* Stack:	( symb Lvars --> Lvars' )
**********************************************************************
NULLNAME PVars{}
::
  DUPTYPELIST? ?SKIP ONE{}N			( symb Lvars )
  DUPLENCOMP UNROT				( #n symb Lvars )
  1LAMBIND					( #n symb )
  syminner PVarSub				( #n )
  1GETABND					( #n Lvars' )
  INNERCOMP get1				( [vars] #n )
  SWAPOVER#- DUPUNROT #2+UNROLL			( [vars1] [vars2] )
  {}N RESOROMP IsV>V? SortList INNERCOMP	( [vars1] [vars2]' )
  top& {}N					( Lvars' )
;

**********************************************************************
* Desc:		Get the variables w.r.t rational in Symb and sort them
* Stack:	( Symb --> L )
**********************************************************************
NULLNAME PVars
::
  syminner			( M )
  NULL{} 1LAMBIND		( M )
    PVarSub			( - )
  1GETABND			( L )
  RESOROMP IsV>V? SortList	( L' )
;

**********************************************************************
* Desc:		Put in 1LAM the variables w.r.t rational in M
* Stack:	( M --> )
**********************************************************************
NULLNAME PVarSub
CODE
PVarSubLoop	GOSUB	A2Rdispatch
		GOTO	PVarSymb	M1 is symbolic
		GOTO	PVarReal	M1 is %
		GOTO	PVarName	M1 is id/lam
		GOTO	PVarSeco	M1 is seco
		GOTO	PVar+		M1 is +
		GOTO	PVar-		M1 is -
		GOTO	PVar*		M1 is *
		GOTO	PVar/		M1 is /
		GOTO	PVarNEG		M1 is NEG
		GOTO	PVarINV		M1 is INV
		GOTO	PVarSQ		M1 is SQ
		GOTO	PVar^		M1 is ^ with real exponent
		GOTO	PVarMisc	None of the above
ENDCODE
*---------------------------------------------------------------------
ASSEMBLE
PVarReal	GOVLNG	=DropLoop
RPL
*---------------------------------------------------------------------
ASSEMBLE
PVarSymb	GOSUB	EvalSub:
RPL
:: #1-SWAP INNERCOMP top&
   CODE
	GOTO	PVarSubLoop
   ENDCODE
;
*---------------------------------------------------------------------
ASSEMBLE
PVarName	GOSUB	EvalSub:
RPL
:: 1GETSWAP AppendList 1PUTLAM ;
*---------------------------------------------------------------------
ASSEMBLE
PVarSeco	GOSUB	EvalSub:
RPL
:: ONESYMBN 1GETSWAP AppendList 1PUTLAM ;
*---------------------------------------------------------------------
ASSEMBLE
PVar+
PVar-
PVar*
PVar/		GOSUB	EvalSub:
RPL
:: pullDROP splitonearg
   CODE
	GOTO	PVarSubLoop
   ENDCODE
   CODE
	GOTO	PVarSubLoop
   ENDCODE
;
*---------------------------------------------------------------------
ASSEMBLE
PVarNEG
PVarINV
PVarSQ		GOSBVL	=POP#
		D1=D1+	5		Drop function
		D=D-1	A
		GOSBVL	=SAVPTR
		A=A-1	A		N-1
		R0=A
		GOSBVL	=PUSH#
		GOTO	PVarSubLoop
RPL
*---------------------------------------------------------------------
ASSEMBLE
PVar^		GOSUB	EvalSub:
RPL
:: pullDROP splitonearg NDROP
   CODE
	GOTO	PVarSubLoop
   ENDCODE
;
*---------------------------------------------------------------------
ASSEMBLE
PVarMisc	GOSUB	EvalSub:
RPL
:: SYMBN 1GETSWAP AppendList 1PUTLAM ;
RPL

**********************************************************************
* A2R dispatcher expects a GOTO table in RSTK, and dispatches
* in the following order:
*	M1 is symbolic
*	M1 is real
*	M1 is idnt/lam
*	M1 is seco
*	M1 is +
*	M1 is -
*	M1 is *
*	M1 is /
*	M1 is NEG
*	M1 is INV
*	M1 is SQ
*	M1 is ^(real integer)
*	M1 is none of the above
*
*	Errors if M1 is = or M1 type is bad
**********************************************************************
ASSEMBLE

EvalSub:	C=RSTK
		A=C	A
		PC=(A)

A2Rdispatch	C=RSTK			Swap D0 and RSTK
		CD0EX
		RSTK=C

		D1=D1+	5		Dispatch to symbolic?
		C=DAT1	A
		D1=D1-	5
		CD1EX
		A=DAT1	A
		D1=C
		LC(5)	=DOSYMB
		?A=C	A
		GOYES	A2Rthis
		D0=D0+	4	No

		A=DAT1	A
		LC(5)	=ONE
		?A#C	A
		GOYES	A2Rfunc
		D1=D1+	5		Drop ONE
		D=D+1	A
		C=DAT1	A
		CD1EX
		A=DAT1	A		type of object
		CD1EX

		LC(5)	=DOREAL		#02933
		?A=C	A
		GOYES	A2Rthis
		D0=D0+	4
		LC(3)	=DOIDNT		#02E48
		?A=C	A
		GOYES	A2Rthis
		LC(2)	=DOLAM		#02E6D
		?A=C	A
		GOYES	A2Rthis
		D0=D0+	4
		LC(3)	=DOCOL		#02D9D
		?A=C	A
		GOYES	A2Rthis
* Error in dispatch
A2Rerror	C=RSTK
		D0=C
		GOVLNG	(=SETTYPEERR)+5
* Dispatch to address D0
A2Rthis		C=RSTK			Restore D0
		CD0EX
		PC=C			And dispatch to the GOTO statement

A2Rfunc		D0=D0+	3*4		Not real/name/seco
		D1=D1+	5
		A=DAT1	A
		D1=D1-	5
		LC(5)	=x+		#1AB67
		?A=C	A
		GOYES	A2Rthis
		D0=D0+	4
		LC(3)	=x-		#1AD09
		?A=C	A
		GOYES	A2Rthis
		D0=D0+	4
		LC(2)	=x*		#1ADEE
		?A=C	A
		GOYES	A2Rthis
		D0=D0+	4
		LC(3)	=x/		#1AF05
		?A=C	A
		GOYES	A2Rthis
		D0=D0+	4
		LC(3)	=xNEG		#1A995
		?A=C	A
		GOYES	A2Rthis
		D0=D0+	4
		LC(4)	=xINV		#1B278
		?A=C	A
		GOYES	A2Rthis
		D0=D0+	4
		LC(3)	=xSQ		#1B426
		?A=C	A
		GOYES	A2Rthis
		D0=D0+	4
* The final 3 in detail
		LC(3)	=x^		#1B02D
		?A=C	A
		GOYES	A2Rtry^
		D0=D0+	4
		LC(4)	=x=		#1A8D8
		?A=C	A
		GOYES	goA2Rerror
		GONC	goA2Rthis
* Try integer exponent
A2Rtry^		CD1EX
		RSTK=C			Save D1
		D1=C

		D1=D1+	5		Skip all NEG commands
		LC(5)	=xNEG
A2Rtry^lp	D1=D1+	5
		A=DAT1	A
		?A=C	A
		GOYES	A2Rtry^lp

		D1=A			Test if ^ is real
		R0=A	A		Save for ->exponent
		A=DAT1	A		ob type
		C=RSTK
		D1=C			Restore D1
		LC(5)	=DOREAL
		?A=C	A
		GOYES	A2Rtry^intg
		D0=D0+	4		Skip to the final dispatchee
goA2Rthis	GOTO	A2Rthis
* Require that the real exponent is integer
A2Rtry^intg	C=R0	A		->real
		CD1EX
		D1=D1+	5
		A=DAT1	W		number
		D1=C
		GOSUB	Integer?
		GONC	goA2Rthis
goA2Rerror	GOTO	A2Rerror

* Test if real in A[W] is an integer
* Out:	CS: Not integer
*	CC: Integer

Integer?	LC(3)	11
		?A>C	X		More than 12 digits or < 1
		RTNYES
		?A=0	W
		GOYES	intgitis	0 is an integer
		P=	4-1		Lowest nibble of mantissa
intg?lp		?A#0	P
		GOYES	intgtst
		C=C-1	X		significant digits--
		P=P+1
		GONC	intg?lp
intgtst		P=	0
		?A<C	X		exponent < significant digits-1
		RTNYES			Yes - not integer
intgitis	RTNCC

RPL

**********************************************************************
* Desc:		Convert Symb with variables v into rational
* Stack:	( Symb v --> Pn Pd )
**********************************************************************
NULLNAME A2R
::
  OVER TYPEREAL? casedrop
  :: R>Z HXS 2 10 ;

  1LAMBIND
    syminner			( M )
    A2Rsub			( Pn Pd )
  ABND
  SWAP PTrim 	
  SWAP PTrim
  DupQIsZero? case ErrInfRes	( infinite result error )
; 

**********************************************************************
* Desc:		Convert meta into rational
* Stack:	( M --> Pn Pd )
**********************************************************************
NULLNAME A2Rsub
CODE
A2RLoop		GOSUB	A2Rdispatch
		GOTO	A2RSymb		M1 is symbolic
		GOTO	A2RReal		M1 is %
		GOTO	A2RName		M1 is id/lam
		GOTO	A2RSeco		M1 is seco
		GOTO	A2R+		M1 is +
		GOTO	A2R-		M1 is -
		GOTO	A2R*		M1 is *
		GOTO	A2R/		M1 is /
		GOTO	A2RNEG		M1 is NEG
		GOTO	A2RINV		M1 is INV
		GOTO	A2RSQ		M1 is SQ
		GOTO	A2R^		M1 is ^ with real exponent
		GOTO	A2RMisc		None of the above
ENDCODE
*---------------------------------------------------------------------
ASSEMBLE
A2RSymb		GOSUB	EvalSub:
RPL
:: #1-SWAP INNERCOMP top&
   CODE
	GOTO	A2RLoop
   ENDCODE
;
*---------------------------------------------------------------------
ASSEMBLE
A2RReal		GOSUB	EvalSub:
RPL
:: R>Z HXS 2 10 ;
*---------------------------------------------------------------------
ASSEMBLE
A2RName		GOSUB	EvalSub:
RPL
:: 1GETLAM ListPos ONE MkSplP HXS 2 10 ;
*---------------------------------------------------------------------
ASSEMBLE
A2RSeco		GOSUB	EvalSub:
RPL
:: ONESYMBN 1GETLAM ListPos ONE MkSplP HXS 2 10 ;
*---------------------------------------------------------------------
ASSEMBLE
A2R^		GOSUB	EvalSub:
RPL
::
  pullDROP splitonearg			( Mbase Mpow )
  M2Z					( Mbase Zpow )
  OVER#2+UNROLL				( Zpow Mbase )

  ( See if base is one of the variables )

  metaDUP CKSymbN			( Zpow Mbase Sbase )
  1GETLAM ListPos DUP#0=csedrp		( Mbase m )
  :: A2Rsub ROT RPow ;			( Pn Pd )

  ( It is - construct poly )

  OVER#2+UNROLL NDROP			( Zpow m )
  SWAP DupZIsNeg? DUP4UNROLL		( Sflag m Zpow Sflag )
  IT QNeg HXS># MkSplP HXS 2 10		( Sflag Pn Pd )
  ROT IT SWAP				( Pn Pd )
;
*---------------------------------------------------------------------
ASSEMBLE
A2RMisc		GOSUB	EvalSub:
RPL
:: SYMBN 1GETLAM ListPos ONE MkSplP HXS 2 10 ;
*---------------------------------------------------------------------
ASSEMBLE
A2R+		GOSUB	EvalSub:
RPL
::
  CODE
	GOTO	A2Rsubs
  ENDCODE
  RAdd
;
*---------------------------------------------------------------------
ASSEMBLE
A2R-		GOSUB	EvalSub:
RPL
::
  CODE
	GOTO	A2Rsubs
  ENDCODE
  RSub
;
*---------------------------------------------------------------------
ASSEMBLE
A2R*		GOSUB	EvalSub:
RPL
::
  CODE
	GOTO	A2Rsubs
  ENDCODE
  RMul
;
*---------------------------------------------------------------------
ASSEMBLE
A2R/		GOSUB	EvalSub:
RPL
::
  CODE
	GOTO	A2Rsubs
  ENDCODE
  SWAP RMul
;
*---------------------------------------------------------------------
ASSEMBLE
A2RNEG		GOSUB	EvalSub:
RPL
:: pullDROP
   CODE
	GOTO	A2RLoop
   ENDCODE
   SWAP QNeg SWAP
;
*---------------------------------------------------------------------
ASSEMBLE
A2RINV		GOSUB	EvalSub:
RPL
:: pullDROP
   CODE
	GOTO	A2RLoop
   ENDCODE
   SWAP
;
*---------------------------------------------------------------------
ASSEMBLE
A2RSQ		GOSUB	EvalSub:
RPL
:: pullDROP
   CODE
	GOTO	A2RLoop
   ENDCODE
   2DUP RMul
;
*---------------------------------------------------------------------

* Small utility for + - * / handling
* ( Meta1&Meta2&+ --> Rn1 Rd1 Rn2 Rd2 )

ASSEMBLE
A2Rsubs		GOSUB	EvalSub:
RPL
::
  pullDROP splitonearg		( --> Meta1 Meta2 )
  CODE
	GOTO	A2RLoop		Convert Meta2
  ENDCODE
  CODE
	C=B	A		Push 2 data stack levels to
	CD0EX			rpl return stack
	A=DAT1	10
	D1=D1+	10
	DAT0=A	10
	D0=D0+	10
	CD0EX
	B=C	A
	GOTO	A2RLoop		Convert Meta1
  ENDCODE
  CODE
	C=B	A		Pop 2 rpl return stack levels to
	CD0EX			data stack
	D0=D0-	10
	D1=D1-	10
	A=DAT0	10
	DAT1=A	10
	CD0EX
	B=C	A
	GOVLNG	=Loop
  ENDCODE
;

**********************************************************************
* Desc:		Convert 2 algebraics into polynomials with given vars
* Stack:	( Ap Aq vars --> v Pn Pd Qn Qd )
**********************************************************************
NULLNAME AA2RR{}
::
  3PICK3PICK XAdd			( Ap Aq vars Ap+Aq )
  SWAP PVars{}				( Ap Aq v )
  DUP4UNROLL A2R			( v Ap Qn Qd )
  ROT 4PICK A2R				( v Qn Qd Pn Pd )
  RCanno 2SWAP RCanno			( v Pn Pd Qn Qd )
;

**********************************************************************
* Desc:		Convert 2 algebraics into polynomials with given vars
* Stack:	( Ap Aq vars --> v P Zp Q Zq
**********************************************************************
NULLNAME AA2PP{}
::
  AA2RR{}				( v Pn Pd Qn Qd )
  DUPTYPEHSTR? NcaseTYPEERR		( * Check Qd is integer * )
  3PICK TYPEHSTR? NcaseTYPEERR		( * Check Pd is integer * )
;

**********************************************************************
* Desc:		Convert 2 algebraics into polynomials with given vars
* Stack:	( Ap Aq vars --> v P Q )
**********************************************************************
NULLNAME AA2PP1{}
::
  AA2RR{}				( v Pn Pd Qn Qd )
  ZIsOne? NcaseTYPEERR			( v Pn Pd Qn )
  SWAP ZIsOne? NcaseTYPEERR		( v Pn Qn )
;

**********************************************************************
* Desc:		Convert 2 algebraics into polynomials with vars v
* Stack:	( Ap Aq --> v Pn Pd Qn Qd )
**********************************************************************
NULLNAME AA2RR
::
  2DUP XAdd PVars			( Ap Aq v )
  DUP4UNROLL A2R			( v Ap Qn Qd )
  ROT 4PICK A2R				( v Qn Qd Pn Pd )
  RCanno 2SWAP RCanno			( v Pn Pd Qn Qd )
;

**********************************************************************
* Desc:		Convert 2 algebraics into polynomials with vars v
* Stack:	( Ap Aq --> v P Zp Q Zq )
**********************************************************************
NULLNAME AA2PP
::
  AA2RR					( v Pn Pd Qn Qd )
  DUPTYPEHSTR? NcaseTYPEERR		( * Check Qd is integer * )
  3PICK TYPEHSTR? NcaseTYPEERR		( * Check Pd is integer * )
;

**********************************************************************
* Desc:		Check if the 2 algebraic inputs are modular,
*		and if so then check them and return the modular
*		forms.
* Stack:	( sym1 sym2 --> sym1 sym2 FALSE )
*		( sym1 sym2 --> vars P1 P2 Zn TRUE )
**********************************************************************
NULLNAME TypeMod2?
::
  TypeMod? ROT TypeMod?			( sym2 mod2 sym1 mod1 )
  ROTOVER				( sym2 sym1 mod1 mod2 mod1 )
  EQUAL NcaseTYPEERR			( * Modulos must be equal * )
  ROTSWAP				( sym1 sym2 mod1 )
  DUPTYPEHSTR? NOTcsdrpfls		( --> sym1 sym2 FALSE )

  UNROT NULL{} AA2RR{}			( Zn v Pn Pd Qn Qd )
  DUPTYPEHSTR? NcaseTYPEERR		( * sym2 must have integer denom * )
  6PICK QDivMod DROP			( Zn v Pn Pd Q )
  UNROT					( Zn v Q Pn Pd )
  DUPTYPEHSTR? NcaseTYPEERR		( * sym1 must have integer denom * )
  5PICK QDivMod DROP			( Zn v Q P )
  SWAP4ROLL TRUE			( --> v P Q Zn TRUE )
;

**********************************************************************
* Desc:		Check if the algebraic input is modular, and if
*		so then check it and return the modular form.
* Stack:	( sym --> sym FALSE )
*		( sym --> vars P Zn TRUE )
**********************************************************************
NULLNAME TypeMod1?
::
  TypeMod?				( sym mod )
  DUPTYPEHSTR? NOTcsdrpfls		( --> sym FALSE )
  SWAP RSim0 UNROT			( mod v Pn Pd )
  DUPTYPEHSTR? NcaseTYPEERR		( * sym must have integer denom * )
  4PICK QDivMod DROPROT TRUE		( --> v P Zn TRUE )
;

*---------------------------------------------------------------------
* Desc:		Help utility which checks if input object
*		is a symbolic with integer modulo as topmost
*		function.
* Stack:	( sym --> sym FALSE / sym' Zn )
*---------------------------------------------------------------------
NULLNAME TypeMod?
::
  DUPTYPESYMB? NOTcaseFALSE			( --> sym FALSE )
  DUP DUPLENCOMP NTHCOMPDROP			( sym ob[n] )
  EQ: xMOD NOTcaseFALSE				( --> sym FALSE )
  INNERCOMP #2- UNROTDROP			( Meta ob[n-1] )
  CKREAL					( * mod must be real * )
  DUP %0>= NcaseTYPEERR				( * mod must be >0 * )
  R>Z						( * mod must be integer * )
  OVER#2+UNROLL CKSymbN SWAP			( sym' Zn )
;

**********************************************************************
* Desc:		Compute the GCD of two algebraics polynomials
* Stack:	( Ap Aq --> Agcd )
**********************************************************************
NULLNAME AAGcd
::
  AA2PP					( v P Zp Q Zq )
  RGcd					( v G Zg )
  ROT R2A				( Ag )
;

**********************************************************************
* Desc:		Compute the LCM of two algebraics polynomials
* Stack:	( Ap Aq --> Alcm )
**********************************************************************
NULLNAME AALcm
::
  AA2PP					( v P Zp Q Zq )
  RLcm					( v G Zg )
  ROT R2A				( Ag )
;

**********************************************************************
* Desc:		Get the numerator and denominator of algebraic
* Stack:	( A -> An Ad )
**********************************************************************
NULLNAME RNumDen	
::
   RSim0			( Pn Pd vars )
   DUPUNROT			( Pn vars Pd vars )
   HXS 2 10 SWAP		( Pn vars Pd Z1 vars )
   R2A				( Pn vars Ad )
   UNROT 			( Ad Pn vars )
   HXS 2 10 SWAP		( Ad Pn Z1 vars )
   R2A				( Ad An )
   SWAP				( An Ad )
;

**********************************************************************
